/* ==================================================== */
/* @File bali.h						*/
/* @history:						*/
/* @+	Apr. 97 <Gloup> first draft			*/
/* ==================================================== */

#ifndef _H_bali
#define _H_bali

/* -------------------------------------------- */
/* constantes programme				*/
/* -------------------------------------------- */

#define Offset_Msk	0x1fff  /* masque 13 bits d'offset		*/
#define Direct_Msk	0x6000  /* masque 2 bits direction		*/
#define Direct_Shf	13	/* shift <<  bits de direction 		*/
#define Jumper_Bit	0x8000  /* bit jumper				*/

#define Oper_Roo	0x0	    /* cellule root			*/
#define Oper_Ins	0x1	    /* insertion			*/
#define Oper_Del	0x2	    /* deletion				*/
#define Oper_Sub	0x3	    /* substitution			*/

#define Pretty_Indel	'-'	    /* pretty char for indels		*/
#define Pretty_Name_Len  10	    /* max length name for ali. output	*/
#define Pretty_Out_Len   50	    /* line length on ali. output	*/

#define ALPHABESTSIZE	 29	    /* [A-Z]+'*'+' '+'\000'		*/					
#define STDALPHABET	"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define NSTDALPHABET	26

#define ALPHABESTBINSIZE	 32	    /* 5 bits/symbol		*/


/* -------------------------------------------- */
/* struct de donnees				*/
/* -------------------------------------------- */

typedef struct {
        int wopen;	/* opening gap extension. Default = SCORENORM	*/
        int wext;       /* extension gap. Default = SCORENORM / 2	*/
        int fglobal;    /* if 1 global (NWS), if 0 Bestfit		*/
        int best_print; /* print alignment				*/
        int printmatrix; /* print alignment				*/
} BestParam;

typedef struct {
    char    alphabrut[ALPHABESTSIZE];
    char    alphabet[NSTDALPHABET+1];	/* +1 for '\000'	*/
    int     nalphabet;
    int     nalphabrut;
    int	    smin, smax;
			/* matrix "brut" score	*/
    int	    brutscore[ALPHABESTSIZE][ALPHABESTSIZE];
			/* normalized distance score	*/
    int	    alphascore[ALPHABESTBINSIZE*ALPHABESTBINSIZE];    
    int	    weight[ALPHABESTBINSIZE*ALPHABESTBINSIZE];    
			/* normalized similarity score	*/
    float   simscore[ALPHABESTBINSIZE*ALPHABESTBINSIZE];    
} ScorMat,*ScorMatPtr;

/* -------------------------------------------- */
/* macros					*/
/* -------------------------------------------- */

#define HCODE_INDEX(i1, i2) (((i1) << 5) | (i2))
#define HCODE_CHAR(c1, c2)  ((c1 - 'A') << 5) | (c2 - 'A')


#define MIN2(a, b) 	    ((a) < (b) ? (a) : (b))	
#define MAX2(a, b) 	    ((a) > (b) ? (a) : (b))	
#define MIN3(a, b, c) 	    ((a) < (b) ? MIN2((a), (c)) : MIN2((b), (c)))

#ifndef NEW
#define NEW(type)		(type *) malloc(sizeof(type))
#define NEWN(type, n)		(type *) malloc((n) * sizeof(type))
#define REALLOC(type, ptr, n) 	(type *) realloc(ptr, (n) * sizeof(type))
#define FREE(ptr)		free(ptr)
#endif

/* -------------------------------------------- */
/* Prototypes					*/
/* -------------------------------------------- */

				/* bali_align.c */
void	FreeTableBuffers    (void);
long	NwsAlign	    (char *seq1, char *seq2, int lseq1, int lseq2, 
			     int *w_subs, int w_indel, int w_all, 
			     char **buffer, int global);

				/* opscan_help.c */
void	PrintBestfitHelp (void);

				/* bali_mat.c */
int	ReadScoreMatrix(char *filename, ScorMat *mat);
int	IdentityMatrix(ScorMat *mat);
void	PrintMatrix(ScorMat *mat, FILE *fout);

#endif
