/* ==================================================== */
/* @File bali_align.c					*/
/* @history:						*/
/* @+	Apr. 97 <Gloup> first draft			*/
/* ==================================================== */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "bali.h"
#include "util.h"

#define N_BUFFERS 6

#define DUMP_TAB    0	/* debug */

/* -------------------------------------------- */
/* structures locales				*/
/* -------------------------------------------- */

typedef struct {	/* --- vecteur de saut -------- */
    long    dist;	/* stockage distance minimum	*/
    int	    offset;	/* offset saut			*/
} Jump, *Jump_ptr;

/* -------------------------------------------- */
/* variables locales				*/
/* -------------------------------------------- */

static void *sBuffersPtr[N_BUFFERS] = {NULL, NULL, NULL, NULL, NULL, NULL};
static long  sBuffersLen[N_BUFFERS] = {0, 0, 0, 0, 0, 0};

/* -------------------------------------------- */
/* debug					*/
/* -------------------------------------------- */

#if DUMP_TAB
static void sDumpRow(long *row, int width)
{
    int i;
    
    for (i = 0; i < width; i++, row++)
	    printf("%4ld ", *row);
	printf("\n");
}
#endif

/* -------------------------------------------- */
/* (re) allocate memory for table buffers	*/
/* -------------------------------------------- */

static void *sReallocTableBuffer(int which, long siz)
{
    void *ptr;
    
    ptr = sBuffersPtr[which];

    if (siz > sBuffersLen[which])	{	/* needs increase	*/

	ptr = (ptr ? realloc(ptr, siz) : malloc(siz));

	if (ptr) {
	    sBuffersPtr[which] = ptr;
	    sBuffersLen[which] = siz;
	}
    }

    return ptr;
}

/* -------------------------------------------- */
/* free all table buffers			*/
/* -------------------------------------------- */

void FreeTableBuffers(void)
{
    int i;
    
    for (i = 0; i < N_BUFFERS ; i++) {

	if (sBuffersPtr[i])
	    free(sBuffersPtr[i]);

	sBuffersPtr[i] = NULL;
	sBuffersLen[i] = 0;
    }    
}

/* -------------------------------------------- */
/* sequence indexing (in seq buffer)	 	*/
/* -------------------------------------------- */

static char *sIndexSequence(char *seq)
{
    char *s;

    for (s= seq ; *s ; s++)
	*s -= 'A'; 

    *s = (char)-1;

    return seq;
}

/* -------------------------------------------- */
/* sequence deindexing (from seq buffer)	*/
/* -------------------------------------------- */

static char *sDeIndexSequence(char *seq)
{
    char *s = seq;

    while ((signed char) *s >= 0)
	*s++ += 'A';

    *s = '\000';

    return seq;
}

/* ------------------------------------------------------------ */
/* fonction d'alignement sans backtracking			*/
/* Version kindel = k0 + (l-1)*k1				*/
/* Note: les sequences doivent deja etre indexees		*/
/* ------------------------------------------------------------ */

#define CHECK_PTR(ptr) if (! (ptr)) return MemoryErreur("[sNwsAlign]", -1)

static long sNwsAlign(char *seq1, char *seq2, int lseq1, int lseq2, 
	              int *w_subs, int w_indel, int w_all, int global)
{
    Jump_ptr	p_col_jump,	/* pointeur saut en colonne	*/
		p_row_jump, 	/* pointeur saut en ligne	*/
		col_jump,	/* saut en colonne		*/
		row_jump;	/* saut en ligne		*/
				
    long	*p_c, 	  	/* - pointeurs cellules		*/
		*p_sub,	 	/* cf plus bas			*/
		*p_ins,		/*				*/
		*p_del,		/*				*/
		*p_scor_up,	/* pointeur upper score line	*/
		*p_scor_dw,	/* pointeur lower score line	*/
		*p_scor_tempo,	/* dummy score line for swap	*/
		*scor_up,	/* upper score line		*/		
		*scor_dw; 	/* lower score line		*/
				
    int		i, j, 		/* indices			*/
		invert,		/* flag inversion		*/
		shifti,		/* 5 bit shift of ith letter	*/
		n1, n2, 	/* dimensions tableau		*/
		dist_ins,	/* distances temporaires	*/
		dist_del,	/*				*/
		dist_sub,	/*				*/
		dist_min; 	/*				*/

    char  	*s1, *s2, 	/* sequences a aligner		*/
		*ps1, *ps2;	/*				*/	
		
			/* ------------------------------------ */
			/* setup parametres			*/
			/* ------------------------------------ */
			
    invert =     (  global  && (lseq2 > lseq1))
	      || ((!global) && (lseq2 < lseq1));
	

    n1 = (invert ? lseq2 : lseq1) + 1;
    n2 = (invert ? lseq1 : lseq2) + 1;

    s1 = (invert ? seq2 : seq1);
    s2 = (invert ? seq1 : seq2);

				/* ------------------------------------ */
				/* allocation des tableaux de calcul	*/
				/* ------------------------------------ */
				
    col_jump = (Jump *) sReallocTableBuffer(1, n2 * (long)sizeof(Jump));

    CHECK_PTR(col_jump);

    row_jump = (Jump *) sReallocTableBuffer(2, n1 * (long)sizeof(Jump));

    CHECK_PTR(row_jump);

    scor_up = (long *) sReallocTableBuffer(3, n2 * (long)sizeof(long));

    CHECK_PTR(scor_up);

    scor_dw = (long *) sReallocTableBuffer(4, n2 * (long)sizeof(long));

    CHECK_PTR(scor_dw);

				/* ------------------------------------ */
				/* indexation des sequences		*/
				/* ------------------------------------ */
	
    /* sequences are already indexed <jompo>
    (void) sIndexSequence(s1); 
    (void) sIndexSequence(s2);
    */

				/* ------------------------------------	*/
				/* Algorithme d'alignement		*/
				/* version iterative			*/
				/* ------------------------------------	*/
				/* Schema des pointeurs			*/
				/* p_c   = pointeur courant		*/
				/* p_ins = pointeur insertion		*/
				/* p_del = pointeur deletion		*/
				/* p_sub = pointeur substitution	*/
				/* 					*/
				/*					*/
				/*  |    <--------nn2------>		*/
				/* -|-------|-------|-------|---  ^	*/
				/*  | p_sub | p_del |       |	  |	*/
				/*  |       |       |       |	  |	*/
				/* -|-------|-------|-------|-    2     */
				/*  | p_ins |  p_c  |       |	  |	*/
				/*  |       |       |       |	  |	*/
				/* -|-------|-------|-------|---  v	*/
				/*  |       |       |       |		*/

				/* ------------------------------------ */
				/* initialisation des pointeurs		*/
				/* ------------------------------------ */
    p_row_jump = row_jump;
    p_scor_up  = scor_up;
    p_scor_dw  = scor_dw;
    p_c = NULL;			/* for no warning ! :-)			*/
	 
				/* ------------------------------------ */
				/* debut d'exploration - loop on s1	*/
				/* ------------------------------------ */
	
    for (i = 0,  ps1 = s1 - 1; i < n1 ; i++, ps1++, p_row_jump++) {
	    
					/* ---------------------------- */
					/* swap score lines		*/
					/* ---------------------------- */
	p_scor_tempo = p_scor_up;		
	p_scor_up    = p_scor_dw;		
	p_scor_dw    = p_scor_tempo;		
					/* ---------------------------- */
					/* mise en place pointeurs	*/
					/* ---------------------------- */
	p_sub = p_del = p_scor_up;
	p_sub--;
	p_c = p_ins = p_scor_dw;
	p_ins--; 
	    
	p_col_jump = col_jump;
	
	shifti = (i ? *ps1 << 5 : 0);
					/* ---------------------------- */
					/* loop on s2			*/
					/* ---------------------------- */
	    
	for (j = 0 , ps2 = s2 - 1 ; j < n2 ; j++, ps2++, p_col_jump++) {

					/* ---------------------------- */
	    if (i && j) {		/* i et j non nuls		*/
					/* ---------------------------- */
						/* calcul des distances	*/
						/* et update des mins	*/

						/* -- insertion --	*/
		
		dist_ins = *p_ins + w_indel;
		p_row_jump->dist += w_all;
			
		if (dist_ins <= p_row_jump->dist) {		/* new gap opening  */
		    p_row_jump->dist   = dist_ins;		/*   new distance   */
		    p_row_jump->offset = j - 1;			/*   new offset	    */
		}
		else {						/* gap extension    */
		    dist_ins  = p_row_jump->dist;		/* actual distance  */
		}
		
						/* -- deletion --	*/
						
		dist_del = *p_del + w_indel;
		p_col_jump->dist += (global || (i != 1)) ? w_all : w_indel;
		
		if (dist_del <= p_col_jump->dist) {		/* new gap opening  */
		    p_col_jump->dist   = dist_del;		/*   new distance   */
		    p_col_jump->offset = i - 1;			/*   new offset	    */
		}
		else {						/* gap extension    */
		    dist_del  = p_col_jump->dist;		/* actual distance  */
		}	
						/* -- substitution --	*/
						
		dist_sub = *p_sub + w_subs[shifti | *ps2];

						/* distance resultante	*/
						
		*p_c = dist_min = MIN3(dist_ins, dist_del, dist_sub);

	    }
					/* ---------------------------- */
	    else if (i) {		/* j nul			*/
					/* ---------------------------- */

		p_row_jump->dist = *p_c = (i == 1 ? w_indel : *p_del + w_all);
		      	
		p_row_jump->offset = 0x0; 
	    }

					/* ---------------------------- */
  	    else if (j) {		/* i nul 			*/
					/* ---------------------------- */

		if (global) 
		    p_col_jump->dist = *p_c = (j == 1 ? w_indel : *p_ins + w_all);
		else
		    p_col_jump->dist = *p_c = 0;
		      
		p_col_jump->offset = 0x0;
	    }

					/* ---------------------------- */
	    else {			/* i et j nuls			*/
					/* ---------------------------- */

		p_row_jump->dist = p_col_jump->dist = *p_c = 0;

		p_row_jump->offset = p_col_jump->offset = 0x0;
	   }

				/* ---------------------------- */
	    p_c++;		/* incrementation 		*/
	    p_ins++;		/* de tous les			*/
	    p_del++;		/* pointeurs			*/
	    p_sub++;		/*				*/
				/* ---------------------------- */
	
	} /* fin de boucle sur j (colonnes) */

#if DUMP_TAB
	(void) sDumpRow(p_scor_dw, n2);
#endif

    } /* fin de boucle sur i (lignes) 	*/
	   		
				/* ------------------------------------ */
				/* fin d'exploration			*/
				/* ------------------------------------ */

				/* ------------------------------------ */
				/* de-indexation des sequences		*/
				/* ------------------------------------ */

    /* sequences stay indexed<jompo>
    (void) sDeIndexSequence(s1); 
    (void) sDeIndexSequence(s2);
     */
				/* ------------------------------------ */
				/* resultats				*/
				/* ------------------------------------ */

    p_c--;			
    dist_min = *p_c;			/* last distance -> cas global	*/

    if (! global) {			/* cas non global : min ligne	*/
	p_c = p_scor_dw;	
	for (i = 0 ; i < n2 ; i++, p_c++)
	    if (*p_c < dist_min)
		dist_min = *p_c;
    }

    return dist_min;
}
#undef CHECK_PTR

/* -------------------------------------------- */
/*  Fonction de backtracking pour l'ecriture	*/
/*  Version non recursive			*/
/*  l'alignement est ecrit dans le buffer	*/
/*  'buffer' de la maniere suivante :		*/
/*  indices pairs   -> premiere sequence	*/
/*  indices impairs -> seconde sequence		*/
/*  un '\000' marque la fin de chacune des	*/
/*  sequences. Ceci permet ensuite d'imprimer	*/
/*  les resultats sur 2 lignes.			*/
/* -------------------------------------------- */

#define TABLO(ii,jj) (tablo[((ii)*col) + (jj)])

static void sBacktrackPretty(char *buffer, short *tablo, int row, int col,
			     char *s1, char *s2)
{	
    int	    i, j, operation, offset, jumping, same;
    int     cellule;
    char    *pb, *p1, *p2, c1, c2;

    c1 = c2 = '\000';	/* for no compiler warning !	*/

    s1 += (row - 2);
    s2 += (col - 2);
    pb  = buffer;
	
    i = row - 1;
    j = col - 1;

	    /* ------------------------------------ */
	    /* remplissage du buffer		    */
	    /* ------------------------------------ */
	
    do {
		
	cellule = (int) TABLO(i, j);
	      
	operation = (cellule & Direct_Msk) >> Direct_Shf;
	offset    = (cellule & Offset_Msk);
	jumping   = ((cellule & Jumper_Bit) != 0);
	   
	switch (operation) {

	    case Oper_Roo :
		c1 = c2 = '\000';
		break;
		
	    case Oper_Sub :
		c1 = *s1--;
		c2 = *s2--;
		same = (c1 == c2);
		c1 = (same ? toupper(c1) : tolower(c1));
		c2 = (same ? toupper(c2) : tolower(c2));
		i--;
		j--;
		break;
			
	    case Oper_Ins :	
		c1 = Pretty_Indel;
		c2 = tolower(*s2--);
		j--;
		if (jumping && (offset < j))
		    TABLO(i, j) = (short) cellule;
		break;
			
	    case Oper_Del :	
		c1 = tolower(*s1--);
		c2 = Pretty_Indel;
		i--;
		if (jumping && (offset < i))
		    TABLO(i, j) = (short) cellule;
		break;
	}
	
	*pb++ = c1;
	*pb++ = c2;
	   
    } while (operation != Oper_Roo);

		    /* ------------------------------------ */
		    /* Renversement du buffer (2 par 2)	    */
		    /* ------------------------------------ */
    p1 = buffer;
    p2 = pb - 4;
	
    while (p1 < p2) {

	c1    = *p1;
	*p1++ = *p2;
	*p2++ = c1;
		
	c1  = *p1;
	*p1 = *p2;
	*p2 = c1;
		
	p1++;
	p2 -= 3; 
    }
	
} /* end of void sBacktrackPretty */

#undef Tablo

/* ------------------------------------------------------------ */
/* swappe sequences in output buffer				*/
/* ------------------------------------------------------------ */
static char *sSwap2Buffer(char *buffer)
{
    char *pb, c;
    
    for (pb = buffer ; *pb ; pb += 2) {
	c     = pb[0];
	pb[0] = pb[1];	
	pb[1] = c;
    }

    return buffer;
}

/* ------------------------------------------------------------ */
/* fonction d'alignement avec backtracking			*/
/* Version kindel = k0 + (l-1)*k1				*/
/* attention le tableau de direction est stocke en short avec   */
/* jump code sur 13 bits => plante si le jump > 8192		*/
/* Note: les sequences doivent deja etre indexees		*/
/* ------------------------------------------------------------ */
#define CHECK_PTR(ptr) if (! (ptr)) return MemoryErreur("[sNwsAlignB]", -1)

static long sNwsAlignB(char *seq1, char *seq2, int lseq1, int lseq2, 
		       int *w_subs, int w_indel, int w_all, 
		       char **buffer, int global)
{
    Jump_ptr	p_col_jump,	/* pointeur saut en colonne	*/
		p_row_jump, 	/* pointeur saut en ligne	*/
		col_jump,	/* saut en colonne		*/
		row_jump;	/* saut en ligne		*/
				
    long	*p_c, 	  	/* - pointeurs cellules		*/
		*p_sub,	 	/* cf plus bas			*/
		*p_ins,		/*				*/
		*p_del,		/*				*/
		*p_scor_up,	/* pointeur upper score line	*/
		*p_scor_dw,	/* pointeur lower score line	*/
		*p_scor_tempo,	/* dummy score line for swap	*/
		*scor_up,	/* upper score line		*/		
		*scor_dw; 	/* lower score line		*/
				
    int		i, j, 		/* indices			*/
		invert,		/* flag inversion		*/
		shifti,		/* 5 bit shift of ith letter	*/
		n1, n2, 	/* dimensions tableau		*/
		dist_ins,	/* distances temporaires	*/
		dist_del,	/*				*/
		dist_sub,	/*				*/
		dist_min; 	/*				*/

    short	*tablo_dir, 	/* direction and jump table	*/
		*p_tablo;	/*				*/

    char  	*s1, *s2, 	/* sequences a aligner		*/
		*ps1, *ps2,	/*				*/
		*pretty_buffer; /* buffer ecriture		*/
		
    p_c = NULL;			/* for no compiler warning :-)	*/

			/* ------------------------------------ */
			/* setup parametres			*/
			/* ------------------------------------ */
			
    invert =     (  global  && (lseq2 > lseq1))
	      || ((!global) && (lseq2 < lseq1));
	

    n1 = (invert ? lseq2 : lseq1) + 1;
    n2 = (invert ? lseq1 : lseq2) + 1;

    s1 = (invert ? seq2 : seq1);
    s2 = (invert ? seq1 : seq2);

				/* ------------------------------------ */
				/* allocation des tableaux de calcul	*/
				/* ------------------------------------ */

    tablo_dir = (short *) sReallocTableBuffer(0, n1 * n2 * (long)sizeof(short));

    CHECK_PTR(tablo_dir);
				
    col_jump = (Jump *) sReallocTableBuffer(1, n2 * (long)sizeof(Jump));

    CHECK_PTR(col_jump);

    row_jump = (Jump *) sReallocTableBuffer(2, n1 * (long)sizeof(Jump));

    CHECK_PTR(row_jump);

    scor_up = (long *) sReallocTableBuffer(3, n2 * (long)sizeof(long));

    CHECK_PTR(scor_up);

    scor_dw = (long *) sReallocTableBuffer(4, n2 * (long)sizeof(long));

    CHECK_PTR(scor_dw);

    pretty_buffer = (char *)  sReallocTableBuffer(5, (2 * (n1 + n2 + 1)) * (long)sizeof(char));

    CHECK_PTR(pretty_buffer);

				/* ------------------------------------ */
				/* indexation des sequences		*/
				/* ------------------------------------ */
		
    /* sequences are already indexed <jompo>
    (void) sIndexSequence(s1); 
    (void) sIndexSequence(s2);
     */

				/* ------------------------------------	*/
				/* Algorithme d'alignement		*/
				/* version iterative			*/
				/* ------------------------------------	*/
				/* Schema des pointeurs			*/
				/* p_c   = pointeur courant		*/
				/* p_ins = pointeur insertion		*/
				/* p_del = pointeur deletion		*/
				/* p_sub = pointeur substitution	*/
				/* 					*/
				/*					*/
				/*  |    <--------nn2------>		*/
				/* -|-------|-------|-------|---  ^	*/
				/*  | p_sub | p_del |       |	  |	*/
				/*  |       |       |       |	  |	*/
				/* -|-------|-------|-------|-    2     */
				/*  | p_ins |  p_c  |       |	  |	*/
				/*  |       |       |       |	  |	*/
				/* -|-------|-------|-------|---  v	*/
				/*  |       |       |       |		*/

				/* ------------------------------------ */
				/* initialisation des pointeurs		*/
				/* ------------------------------------ */
    p_row_jump = row_jump;
    p_scor_up  = scor_up;
    p_scor_dw  = scor_dw;
    
    p_tablo = tablo_dir;
	 
				/* ------------------------------------ */
				/* debut d'exploration - loop on s1	*/
				/* ------------------------------------ */
	
    for (i = 0,  ps1 = s1 - 1; i < n1 ; i++, ps1++, p_row_jump++) {
	    
					/* ---------------------------- */
					/* swap score lines		*/
					/* ---------------------------- */
	p_scor_tempo = p_scor_up;		
	p_scor_up    = p_scor_dw;		
	p_scor_dw    = p_scor_tempo;		
					/* ---------------------------- */
					/* mise en place pointeurs	*/
					/* ---------------------------- */
	p_sub = p_del = p_scor_up;
	p_sub--;
	p_c = p_ins = p_scor_dw;
	p_ins--; 
	    
	p_col_jump = col_jump;
	
	shifti = (i ? *ps1 << 5 : 0);
					/* ---------------------------- */
					/* loop on s2			*/
					/* ---------------------------- */
	    
	for (j = 0 , ps2 = s2 - 1 ; j < n2 ; j++, ps2++, p_col_jump++) {

					/* ---------------------------- */
	    if (i && j) {		/* i et j non nuls		*/
					/* ---------------------------- */
						/* calcul des distances	*/
						/* et update des mins	*/
						
		*p_tablo = Oper_Roo;		/* Oper_Roo = 0x0	*/
		
						/* -- insertion --	*/
		
		dist_ins = *p_ins + w_indel;
		p_row_jump->dist += w_all;
			
		if (dist_ins <= p_row_jump->dist) {		/* new gap opening  */
		    p_row_jump->dist   = dist_ins;		/*   new distance   */
		    p_row_jump->offset = (j-1) & Offset_Msk;	/*   new offset	    */
		}
		else {						/* gap extension    */
		    dist_ins  = p_row_jump->dist;		/* actual distance  */
		    *p_tablo |= Oper_Ins ;			/* jump	bit on	    */
		}
		
						/* -- deletion --	*/
						
		dist_del = *p_del + w_indel;
		p_col_jump->dist += (global || (i != 1)) ? w_all : w_indel;
		
		if (dist_del <= p_col_jump->dist) {		/* new gap opening  */
		    p_col_jump->dist   = dist_del;		/*   new distance   */
		    p_col_jump->offset = (i-1) & Offset_Msk;	/*   new offset	    */
		}
		else {						/* gap extension    */
		    dist_del  = p_col_jump->dist;		/* actual distance  */
		    *p_tablo |= Oper_Del;			/* jump	bit on	    */
		}	
						/* -- substitution --	*/
						
		dist_sub = *p_sub + w_subs[shifti | *ps2];

						/* distance resultante	*/
						
		*p_c = dist_min = MIN3(dist_ins, dist_del, dist_sub);

						/* determination path	*/
						/* en cas d'egalite la	*/
						/* priorite est :	*/
						/* sub > ins > del	*/
						
						/* -- substitution --	*/

		if (dist_min == dist_sub)
		    *p_tablo = Oper_Sub << Direct_Shf;

						/* -- insertion --	*/
						
		else if (dist_min == dist_ins) {
		
		    if (*p_tablo & Oper_Ins) 		/* -- jump ! --	*/
			*p_tablo = (Oper_Ins << Direct_Shf) | Jumper_Bit | p_row_jump->offset;
		    else
			*p_tablo = (Oper_Ins << Direct_Shf);

		}   
						/* -- deletion --	*/
		else {
		
		    if (*p_tablo & Oper_Del) 		/* -- jump ! --	*/
			*p_tablo = (Oper_Del << Direct_Shf) | Jumper_Bit | p_col_jump->offset;
		    else
			*p_tablo = (Oper_Del << Direct_Shf);

		}
	    }

					/* ---------------------------- */
	    else if (i) {		/* j nul			*/
					/* ---------------------------- */

		p_row_jump->dist = *p_c = (i == 1 ? w_indel : *p_del + w_all);
		      	
		*p_tablo = Oper_Del << Direct_Shf;  
	
		p_row_jump->offset = 0x0; 
		
	    }

					/* ---------------------------- */
  	    else if (j) {		/* i nul 			*/
					/* ---------------------------- */

		if (global) 
		    p_col_jump->dist = *p_c = (j == 1 ? w_indel : *p_ins + w_all);
		else
		    p_col_jump->dist = *p_c = 0;
		      
		*p_tablo = Oper_Ins << Direct_Shf;   

		p_col_jump->offset = 0x0;
	    }

					/* ---------------------------- */
	    else {			/* i et j nuls			*/
					/* ---------------------------- */

		p_row_jump->dist = p_col_jump->dist = *p_c = 0;

		*p_tablo = Oper_Roo << Direct_Shf;
		
		p_row_jump->offset = p_col_jump->offset = 0x0;
	   	
	   }

				/* ---------------------------- */
	    p_c++;		/* incrementation 		*/
	    p_ins++;		/* de tous les			*/
	    p_del++;		/* pointeurs			*/
	    p_sub++;		/*				*/
				/* ---------------------------- */
	
	    p_tablo++;

	} /* fin de boucle sur j (colonnes) */

#if DUMP_TAB
	(void) sDumpRow(p_scor_dw, n2);
#endif

    } /* fin de boucle sur i (lignes) 	*/
	   		
				/* ------------------------------------ */
				/* fin d'exploration			*/
				/* ------------------------------------ */

				/* ------------------------------------ */
				/* backtracking	    			*/
				/* ------------------------------------ */
				
    p_c--;				/* point to the last cell	*/
    p_tablo--;				/* point to the last cell	*/

    dist_min = *p_c;

    if (! global) {			/* cas non global  recherche min */
    
	p_ins = p_c = p_scor_dw;	

	for (i = 0 ; i < n2 ; i++, p_c++) {
	    if (*p_c <= dist_min) {
		p_ins    = p_c;
		dist_min = *p_ins;
	    }
	}
	
	p_c--;
	for ( ; p_ins < p_c; p_ins++, p_tablo--)
	    *(p_tablo) = Oper_Ins << Direct_Shf;
			
    }

				/* ------------------------------------ */
				/* de-indexation des sequences		*/
				/* ------------------------------------ */
	
    /* sequences must be indexed at end of function	<jompo>
       so we shall index them later in the function
    */
    (void) sDeIndexSequence(s1); 
    (void) sDeIndexSequence(s2);

    (void) sBacktrackPretty(pretty_buffer, tablo_dir, n1, n2, s1, s2);
    
    /* sequences must be indexed at end of function      <jompo>	*/
    (void) sIndexSequence(s1); 
    (void) sIndexSequence(s2);

    *buffer = (invert ? sSwap2Buffer(pretty_buffer) : pretty_buffer);
    
    return dist_min;
}

/* ------------------------------------------------------------ */
/* interface alignement						*/
/* ------------------------------------------------------------ */

long NwsAlign(char *seq1, char *seq2, int lseq1, int lseq2, 
	      int *w_subs, int w_indel, int w_all, 
	      char **buffer, int global)
{
    return (buffer ? sNwsAlignB(seq1, seq2, lseq1, lseq2, w_subs,
				w_indel, w_all, buffer, global)
		   : sNwsAlign (seq1, seq2, lseq1, lseq2, w_subs,
				w_indel, w_all, global));
}
