# ---------------------------------------------------------------
# @file: Makefile
# @desc: makefile for opscan
#
# @history:
# @+	    <jompo> : Apr 00 : Created
#           <jompo> : Jul 01 : last modified
# ---------------------------------------------------------------
#

SHELL = /bin/sh

SRC   = fasta_io.c	\
	util.c	\
	fastp_hash.c	\
	fastp_core.c	\
	operon.c	\
	bali_align.c	\
	bali_mat.c	\
	opscan_help.c	\
	class.c	\
	stat.c	\
	combi.c	\
	opscan.c


OBJ   = $(SRC:.c=.o)

PROGS  = opscan

#
# Machine dependant flags
# defaults
#

#Any:     MACHDEF =
#Sun:     MACHDEF = -DSUN
#Solaris: MACHDEF = -DSUN -DSOLARIS
#Sgi:     MACHDEF = -DSGI

MACHDEF = -DSGI 

#Any:   MALLOC =
#Sun:   MALLOC = 
#Sgi:   MALLOC = -lmalloc
#SgiCV: MALLOC = -lmalloc_cv

MALLOC = -lmalloc 

#Any: RANLIB =
#Sun: RANLIB = ranlib
#Sgi: RANLIB = touch

RANLIB = touch 

#Any: CC =
#Sun: CC = acc
#Sgi: CC = cc -fullwarn
#Sgi with gcc: CC = gcc -Wall


CC = gcc -Wall

#Any: TFLAGS =
#Sun: TFLAGS =
#Sgi: TFLAGS =

TFLAGS =

#
# Machine independant flags
#
#

INCDIR	 =	.
BINDIR   =	../bin/

#OPTIM = -g
OPTIM = -O

LINTFLAGS = 	$(MACHDEF) -u -m -x

CFLAGS 	 = 	$(OPTIM) $(MACHDEF) $(TFLAGS)

LFLAGS	 =

LIBS	 =	-lm $(MALLOC)

#
# Rules
#

.KEEP_STATE :

all  :	$(PROGS) install

opscan:	$(OBJ)
	$(CC) $(CFLAGS) -o $@ $(OBJ) $(LFLAGS) $(LIBS)

	
lint:
	-for f in $(SRC) ; 		    \
		do echo ">> lint $$f <<";   \
		lint $$f  $(LINTFLAGS) ;    \
	done

install:
	-for f in $(PROGS) ; do \mv `pwd`/$$f  $(BINDIR) ;  done

clean :
	-\rm -f *.o cvstatic* *% *.bak so_loc*

superclean: clean
	-\rm -f $(PROGS)
	-cd $(BINDIR); \rm -f $(PROGS)

rebuilt: superclean all
	@echo "rebuilt done"

test:
	$(BINDIR)/opscan -M ../matrices/BLOSUM62 -C -c -L 5.0 -o 1.2 -e 0.8 -n 4 -E -F -O \
	 ../tests/test_operon.db ../tests/urur.10.prt > ../tests/test.out;
	if `diff ../tests/test.out ../tests/test.out.orig > /dev/null`; then \
	echo "test OK"; else echo "test not OK, verify..."; fi


Makefile: $(FRC)
	-\mv -f $@ ${@}.old
	echo '# Ce fichier est genere a partir de basic.mk' > $@
	cat basic.mk >> $@
	echo '# Automatically-generated dependencies list:' >> $@
	cc -M $(CFLAGS) $(SRC) | grep -v "/usr/include"   >> $@

