/* ==================================================== */
/* @File class.c                                        */
/* @history:                                            */
/* @+   Jul. 00 <jompo> first draft                     */
/* ==================================================== */

#include <stdio.h>
#include <stdlib.h>		/* for malloc	*/
#include <string.h>		/* for memcpy	*/

#include "util.h"
#include "operon.h"
#include "class.h"
#include "tags.h"
#include "opscan_proto.h"

#ifndef ABS
#define ABS(x)              ((x) >= 0  ? (x) : -(x))
#endif

/* ----------------------------------------------------	*/
/* prototypes of static functions			*/
/* ----------------------------------------------------	*/
static int sComputeR2ClassesOnly(OperonSeqPtr *ope, int nb_ope, int nseuil,
			int circular, int minn, int maxn,
			int cur_clas, int verbose);
static int sComputeR1ClassesOnly(GenomeSeqPtr *gen, int nb_gen, int nseuil,
			int circular, int minn, int maxn,
			int cur_clas, int verbose);
static int sComputeR2ClassesOnR1(OperonSeqPtr *ope, int nb_ope, int nseuil,
			int ope_circular, int gen_circular,
			int mingn, int maxgn,
			int minon, int maxon, int verbose);
static int sComputeR1ClassesOnR2(GenomeSeqPtr *gen, int nb_gen, int nseuil,
			int gen_circular, int ope_circular,
			int minon, int maxon,
			int mingn, int maxgn,
			int verbose);
void sSetNOCLASS_for_gen_with_no_homol(GenomeSeqPtr *gen, int nb_gen);
void sSetOpeSingleton(OperonSeqPtr *ope, int nb_ope);
void sSetGenSingleton(GenomeSeqPtr *gen, int nb_gen);
void sSortOpeClassesAndSingleton(OperonSeqPtr *ope, int nb_ope);


/* ----------------------------------------------------	*/
/* compar for qsort number of best similar gene of ope1	*/
/* with number of best similar gene of ope2 and return:	*/
/* +1 if number of best similar gene of ope1 is >	*/
/*        number of best similar gene of ope2		*/
/* -1 if number of best similar gene of ope1 is <	*/
/*        number of best similar gene of ope2		*/
/*  0 if number of best similar gene of ope1 is ==	*/
/*        number of best similar gene of ope2		*/
/* Note: if no homolog (NOT_G), the ope is inf, so it	*/
/* will be at the beginning of the sorted array		*/
/* ----------------------------------------------------	*/
static int sComparGenNum(const void *ope1, const void *ope2)
{

  if (((*((OperonSeqPtr *)ope1))->homolog.gen_ptr == NOT_G) &&
      ((*((OperonSeqPtr *)ope2))->homolog.gen_ptr == NOT_G))
         return 0;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr == NOT_G)
     return -1;
  else if ((*((OperonSeqPtr *)ope2))->homolog.gen_ptr == NOT_G)
     return +1;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr->rel_number >
	(*((OperonSeqPtr *)ope2))->homolog.gen_ptr->rel_number)
     return 1;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr->rel_number <
	 (*((OperonSeqPtr *)ope2))->homolog.gen_ptr->rel_number)
     return -1;
  else 
     return 0;

}


/* ----------------------------------------------------	*/
/* compar for qsort class of best similar gene of ope1	*/
/* with class of best similar gene of ope2 and return:	*/
/* +1 if class of best similar gene of ope1 is >	*/
/*        class of best similar gene of ope2		*/
/* -1 if class of best similar gene of ope1 is <	*/
/*        class of best similar gene of ope2		*/
/*  0 if class of best similar gene of ope1 is ==	*/
/*        class of best similar gene of ope2		*/
/* Note: if no homolog (NOT_G), the ope is inf, so it	*/
/* will be at the begining of the sorted array		*/
/* Note2: the ope with homolog class == NO_CLASS	*/
/*  will be after the homolog  == NOT_G			*/
/*  and the ope with homolog class == SINGLETON		*/
/*  will be after ope with homolog class == NO_CLASS	*/
/* ----------------------------------------------------	*/
static int sComparOpeGenClass(const void *ope1, const void *ope2)
{

  if (((*((OperonSeqPtr *)ope1))->homolog.gen_ptr == NOT_G) &&
      ((*((OperonSeqPtr *)ope2))->homolog.gen_ptr == NOT_G))
         return 0;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr == NOT_G)
     return -1;
  else if ((*((OperonSeqPtr *)ope2))->homolog.gen_ptr == NOT_G)
     return +1;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr->class >
	(*((OperonSeqPtr *)ope2))->homolog.gen_ptr->class)
     return 1;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr->class <
	 (*((OperonSeqPtr *)ope2))->homolog.gen_ptr->class)
     return -1;
  else 
     return 0;

}


/* ----------------------------------------------------	*/
/* compar for qsort class of best similar ope of gen1	*/
/* with class of best similar ope of gen2 and return:	*/
/* +1 if class of best similar ope of gen1 is >		*/
/*        class of best similar ope of gen2		*/
/* -1 if class of best similar ope of gen1 is <		*/
/*        class of best similar ope of gen2		*/
/*  0 if class of best similar ope of gen1 is ==	*/
/*        class of best similar ope of gen2		*/
/* Note: if no homolog (NOT_G), the gen is inf, so it	*/
/* will be at the begining of the sorted array		*/
/* Note2: the gen with homol_ptr->class == NO_CLASS	*/
/*  will be after the homol_ptr == NOT_G		*/
/*  and the gen with homol_ptr->class == SINGLETON	*/
/*  will be after gen with homol_ptr->class == NO_CLASS	*/
/* ----------------------------------------------------	*/
static int sComparGenOpeClass(const void *gen1, const void *gen2)
{

  if (((*((GenomeSeqPtr *)gen1))->homol_ptr == NOT_O) &&
      ((*((GenomeSeqPtr *)gen2))->homol_ptr == NOT_O))
         return 0;
  else if ((*((GenomeSeqPtr *)gen1))->homol_ptr == NOT_O)
     return -1;
  else if ((*((GenomeSeqPtr *)gen2))->homol_ptr == NOT_O)
     return +1;
  else if ((*((GenomeSeqPtr *)gen1))->homol_ptr->class >
	(*((GenomeSeqPtr *)gen2))->homol_ptr->class)
     return 1;
  else if ((*((GenomeSeqPtr *)gen1))->homol_ptr->class <
	 (*((GenomeSeqPtr *)gen2))->homol_ptr->class)
     return -1;
  else 
     return 0;

}


/* ----------------------------------------------------	*/
/* compar for qsort                                     */
/* returns:						*/
/* +1 if num of homolog gene of ope1 is >		*/
/*        num of homolog gene of ope2			*/
/* -1 if num of homolog gene of ope1 is <		*/
/*        num of homolog gene of ope2			*/
/*  0 if num of homolog gene of ope1 is ==		*/
/*        num of homolog gene of ope2			*/
/* Note: if no homolog (NOT_G), the ope is inf, so it	*/
/* will be at the beginning of the sorted array		*/
/* ----------------------------------------------------	*/
static int sComparOpeGenNum(const void *ope1, const void *ope2)
{

  if (((*((OperonSeqPtr *)ope1))->homolog.gen_ptr == NOT_G) &&
      ((*((OperonSeqPtr *)ope2))->homolog.gen_ptr == NOT_G))
         return 0;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr == NOT_G)
     return -1;
  else if ((*((OperonSeqPtr *)ope2))->homolog.gen_ptr == NOT_G)
     return +1;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr->rel_number >
	(*((OperonSeqPtr *)ope2))->homolog.gen_ptr->rel_number)
     return 1;
  else if ((*((OperonSeqPtr *)ope1))->homolog.gen_ptr->rel_number <
	 (*((OperonSeqPtr *)ope2))->homolog.gen_ptr->rel_number)
     return -1;
  else 
     return 0;

}

/* ----------------------------------------------------	*/
/* compar for qsort                                     */
/* returns:						*/
/* +1 if ignore of homolog ope of gen1 is >		*/
/*        ignore of homolog ope of gen2			*/
/* -1 if ignore of homolog ope of gen1 is <		*/
/*        ignore of homolog ope of gen2			*/
/*  0 if ignore of homolog ope of gen1 is ==		*/
/*        ignore of homolog ope of gen2			*/
/* Note: if no homolog (NOT_O), the gen is inf, so it	*/
/* will be at the beginning of the sorted array		*/
/* ----------------------------------------------------	*/
static int sComparGenOpeNum(const void *gen1, const void *gen2)
{

  if (((*((GenomeSeqPtr *)gen1))->homol_ptr == NOT_O) &&
      ((*((GenomeSeqPtr *)gen2))->homol_ptr == NOT_O))
         return 0;
  else if ((*((GenomeSeqPtr *)gen1))->homol_ptr == NOT_O)
     return -1;
  else if ((*((GenomeSeqPtr *)gen2))->homol_ptr == NOT_O)
     return +1;
  else if ((*((GenomeSeqPtr *)gen1))->homol_ptr->ignore >
	(*((GenomeSeqPtr *)gen2))->homol_ptr->ignore)
     return 1;
  else if ((*((GenomeSeqPtr *)gen1))->homol_ptr->ignore <
	 (*((GenomeSeqPtr *)gen2))->homol_ptr->ignore)
     return -1;
  else 
     return 0;

}


/* ----------------------------------------------------	*/
/* compar for qsort : compare the relnumber of 2 operon	*/
/* gene pointers					*/
/* returns:						*/
/* +1 if ignore of ope1 is > ignore of ope2      	*/
/* -1 if ignore of ope1 is < ignore of ope2      	*/
/*  0 if ignore of ope1 is == ignore of ope2       	*/
/* ----------------------------------------------------	*/
static int sComparOpeNum(const void *ope1, const void *ope2)
{

  if ((*((OperonSeqPtr *)ope1))->ignore >
	(*((OperonSeqPtr *)ope2))->ignore)
	return 1;
  else if ((*((OperonSeqPtr *)ope1))->ignore <
	 (*((OperonSeqPtr *)ope2))->ignore)
	return -1;
  else 
	return 0;

}


/* ----------------------------------------------------	*/
/* compar for qsort : compare the class num of 2 operon	*/
/* gene pointers					*/
/* returns:						*/
/* +1 if class of ope1 is > class of ope2	      	*/
/* -1 if class of ope1 is < class of ope2	      	*/
/*  0 if class of ope1 is == class of ope2       	*/
/* ----------------------------------------------------	*/
static int sComparOpeClass(const void *ope1, const void *ope2)
{

			/* as NO_CLASS is < 0, the	*/
			/* non classified ope will be	*/
			/* in the first position in	*/
			/* array			*/
			/* as SINGLETON is < 0 and 	*/
			/* SINGLETON > NO_CLASS, the	*/
			/* singleton ope will be	*/
			/* after the NO_CLASS in	*/
			/* array			*/

  if ((*((OperonSeqPtr *)ope1))->class >
	(*((OperonSeqPtr *)ope2))->class)
	return 1;
  else if ((*((OperonSeqPtr *)ope1))->class <
	 (*((OperonSeqPtr *)ope2))->class)
	return -1;
  else 
	return 0;

}


/* ----------------------------------------------------	*/
/* compar for qsort : compare the class num of 2	*/
/* genomic gene pointers				*/
/* returns:						*/
/* +1 if class of gen1 is > class of gen2	      	*/
/* -1 if class of gen1 is < class of gen2	      	*/
/*  0 if class of gen1 is == class of gen2       	*/
/* ----------------------------------------------------	*/
static int sComparGenClass(const void *gen1, const void *gen2)
{

			/* as NO_CLASS is < 0, the	*/
			/* non classified gen will be	*/
			/* in the first position in	*/
			/* array			*/
			/* as SINGLETON is < 0 and 	*/
			/* SINGLETON > NO_CLASS, the	*/
			/* singleton ope will be	*/
			/* after the NO_CLASS in	*/
			/* array			*/

  if ((*((GenomeSeqPtr *)gen1))->class >
	(*((GenomeSeqPtr *)gen2))->class)
	return 1;
  else if ((*((GenomeSeqPtr *)gen1))->class <
	 (*((GenomeSeqPtr *)gen2))->class)
	return -1;
  else 
	return 0;

}

#if 0
/* ----------------------------------------------------	*/
/* for debugging purposes: sverif give various errors	*/
/* ----------------------------------------------------	*/
static void sverif(OperonSeqPtr *ope, int nb_ope, GenomeSeqPtr *gen, int nb_gen)
{
   int i;

   for (i = 0; i < nb_ope; i++)
      if (ope[i]->homolog.gen_ptr != NOT_G && ope[i]->homolog.gen_ptr->class == NO_CLASS)
	 fprintf(stderr,"---->ope %s have homolog gen %s and class %d (gen class %d)\n",
		ope[i]->mnemo, ope[i]->homolog.gen_ptr->gene_name,
		ope[i]->class, ope[i]->homolog.gen_ptr->class);

   fprintf(stderr, "====> dissimilarity between ope and homolog class\n");
   for (i = 0; i < nb_ope; i++)
      if (ope[i]->homolog.gen_ptr != NOT_G && ope[i]->homolog.gen_ptr->class != ope[i]->class)
	 fprintf(stderr,"---->ope %s have homolog gen %s and class %d (gen class %d)\n",
		ope[i]->mnemo, ope[i]->homolog.gen_ptr->gene_name,
		ope[i]->class, ope[i]->homolog.gen_ptr->class);

   fprintf(stderr, "====> dissimilarity between ope and homolog ptr\n");
   for (i = 0; i < nb_ope; i++)
      if (ope[i]->homolog.gen_ptr != NOT_G && ope[i]->homolog.gen_ptr->homol_ptr != ope[i])
	 fprintf(stderr,"---->ope %s have homolog gen %s and class %d (gen class %d)\n",
		ope[i]->mnemo, ope[i]->homolog.gen_ptr->gene_name,
		ope[i]->class, ope[i]->homolog.gen_ptr->class);

   for (i = 0; i < nb_gen; i++)
      if (gen[i]->homol_ptr != NOT_O && gen[i]->homol_ptr->class == NO_CLASS)
	 fprintf(stderr,"---->gen %s have homolog ope %s and class %d (ope class %d)\n",
		gen[i]->gene_name, gen[i]->homol_ptr->mnemo,
		gen[i]->class, gen[i]->homol_ptr->class);
   return;
}
#endif

/* ----------------------------------------------------	*/
/* ComputeR2Classes: find the class of genes in		*/
/* gen_kept of all ope[i_ope]				*/
/* ope[i_ope].gen_kept[0] is supposed to be the most	*/
/* similar gene with ope[i_ope]				*/
/* all ope must have their field homolog.gen_ptr valid	*/
/* or equal to NOT_G (must not be NULL)			*/
/* all gen must have their field homol_ptr valid	*/
/* or equal to NOT_O (must not be NULL)			*/
/* circular == 1 means that minn is at a distance 1 to	*/
/* maxn, else minn and maxn are not used		*/
/* operon_classe: 1: operon genes are in several R1	*/
/*                   classes				*/
/*                0: operon genes are all in the same	*/
/*                   R1 class				*/
/* returns: number of classes (-1 if error) 		*/
/* ----------------------------------------------------	*/
int ComputeR2Classes(OperonSeqPtr *ope, int nb_ope,
			GenomeSeqPtr *gen, int nb_gen,
			int nseuil,
			int operon_classe, 
			int gen_circular, int mingn, int maxgn,
			int ope_circular, int minon, int maxon,
			int verbose)
{

  int	nclassr2, nclassr1, nclass_old;

  if  (operon_classe == 0) {
     if (verbose)
	fprintf(stderr, "\n" COMPLETION_TAG "Computing R2 classes:\n");

     return sComputeR2ClassesOnly(ope, nb_ope, nseuil, gen_circular,
					mingn, maxgn, 0, verbose);
  }
  else {
     if (verbose)
	 fprintf(stderr, "\n" COMPLETION_TAG "Loop for computing R2 classes :\n");

     nclass_old = -2;			/* dummy initialisation	*/
     nclassr2 = -1;
     while (nclass_old != nclassr2) {
	nclass_old = nclassr2;
	nclassr2 = sComputeR2ClassesOnR1(ope, nb_ope, nseuil,
		ope_circular, gen_circular, mingn, maxgn,
		minon, maxon, verbose);

	sSetOpeSingleton(ope, nb_ope);

#if 0	/* debug	*/
	fprintf(stderr," ============> AFTER sComputeR2ClassesOnR1\n");
	PrintResultR2(ope, nb_ope, nclassr2, 0);
#endif

	if (nclass_old == -1)
	   sSetNOCLASS_for_gen_with_no_homol(gen, nb_gen);

	nclassr1 = sComputeR1ClassesOnR2(gen, nb_gen, nseuil,
		gen_circular, ope_circular, minon, maxon,
		mingn, maxgn, verbose);

	sSetGenSingleton(gen, nb_gen);

#if 0	/* debug	*/
	fprintf(stderr," ============> AFTER sComputeR1ClassesOnR2\n");
	PrintResultR2(ope, nb_ope, nclassr1, 0);
#endif

        if (verbose)
	    fprintf(stderr,
	       COMPLETION_TAG "Number of Class R2  = %d Number of Class R1 = %d\n",
	       nclassr2, nclassr1);
     }

     sSortOpeClassesAndSingleton(ope, nb_ope);

     return nclassr2;
  }

}

/* ----------------------------------------------------	*/
/* set the class of genomic genes with no homolog to	*/
/* NO_CLASS						*/
/* ----------------------------------------------------	*/
void sSetNOCLASS_for_gen_with_no_homol(GenomeSeqPtr *gen,	
					int nb_gen)
{
   int	i;
   for (i = 0; i < nb_gen; i++) 
      if (gen[i]->homol_ptr == NOT_O)
	 gen[i]->class = NO_CLASS;

   return;
}

/* ----------------------------------------------------	*/
/* sComputeR2ClassesOnly: computes class on genomic	*/
/* genes, the ope genes are assumed to be all in the	*/
/* same operon class					*/
/* the class are numbered from cur_class		*/
/* return the number of classes	found			*/
/* ----------------------------------------------------	*/
static int sComputeR2ClassesOnly(OperonSeqPtr *ope, int nb_ope, int nseuil,
			int circular, int minn, int maxn,
			int cur_clas, int verbose)
{
  int           (*compar) (const void *, const void *);
  int           i, last_class, i_prec_homol, i_first_homol,
		i_last_homol;

		/* first, sort operon genes on the rel_num of	*/
		/* their best genomic gene hit			*/
  compar = sComparGenNum;

  qsort(ope, nb_ope, sizeof(* ope), compar);

  		/* calcul des classes pour les homologues		*/
		/* attention tous les ope doivent avoir un homologue	*/
		/* ou bien doivent avoir homolog.gen_ptr == NOT_G	*/
		/* SURTOUT PAS NULL...sinon plante			*/
   
   last_class = cur_clas;


   i = 0;

   while(i < nb_ope && ope[i]->homolog.gen_ptr == NOT_G)
         i++;

   if (i >= nb_ope)
     return 0;

			/* precedent ope_ptr ayant un homologue	*/
   i_first_homol = i_prec_homol = i;

   ope[i]->class = ope[i]->homolog.gen_ptr->class = last_class;


   for ( ++i ; i < nb_ope; i++) {

     if(verbose)
	(void)fprintf(stderr, "\r" COMPLETION_TAG "sequence %d / %d\t\t", i+1, nb_ope);

				/* si pas d'homologue, on	*/
				/* saute au suivant		*/
     if (ope[i]->homolog.gen_ptr == NOT_G) {
	ope[i]->class = NO_CLASS;
	continue;
     }

			/* on regarde si la distance en gene	*/
			/* est adequate pour mettre le		*/
			/* ope[i]->homolog dans la meme classe	*/
			/* que le ope[i_prec_homol]->homolog	*/
			/* note: on verifie quand meme qu'ils	*/
			/* n'ont pas le meme rel_num (i.e. que	*/
			/* ce n'est pas le meme gene qui est	*/
			/* homologue a deux "operon genes"	*/
			/* successifs				*/
			/* sinon, on cree une nouvelle classe	*/
				/* ABS(diff)-1 < nseuil		*/
				/* car si on a			*/
				/* i   j			*/
				/* 12345...			*/
				/* on considere que la distance	*/
				/* en gene de i a j est de	*/
				/* j-i-1, cad 5-1-1->3 genes	*/
      if ((ope[i]->homolog.gen_ptr->rel_number !=
	     ope[i_prec_homol]->homolog.gen_ptr->rel_number)
	     && (ABS(ope[i]->homolog.gen_ptr->rel_number - 
	     ope[i_prec_homol]->homolog.gen_ptr->rel_number)-
	     1 <= nseuil)) {
	   ope[i]->homolog.gen_ptr->class =
			ope[i]->class = last_class;
      }
      else {
	   ope[i]->homolog.gen_ptr->class =
			ope[i]->class = ++last_class;
      }

      i_prec_homol = i;
   }

   i_last_homol = i_prec_homol;


   if (circular) {
			/* si circularite, et qu'il y a plus	*/
			/* d'une classe				*/
			/* mettre la classe du (des) dernier(s)	*/
			/* gene(s) au meme numero que la classe	*/
			/* du premier gene (ou l'inverse) si	*/
			/* leur distance est correcte		*/

      if (last_class > 0 &&
          (ABS(maxn - ope[i_last_homol]->homolog.gen_ptr->rel_number) +
	  ABS(ope[i_first_homol]->homolog.gen_ptr->rel_number-
	  minn) <= nseuil)) {

			/* descent from the last until the	*/
			/* class is not the same		*/
			/* and set the class of the last	*/
			/* elements to the first class		*/
	  i = i_last_homol;
	  while(i >= 0) {
	    if (ope[i]->homolog.gen_ptr != NOT_G) {
	        if (ope[i]->homolog.gen_ptr->class == last_class)
	           ope[i]->homolog.gen_ptr->class =
			ope[i]->class = 
			  ope[i_first_homol]->homolog.gen_ptr->class;
		else
		   break;
	     }
             i--;
	  }
			/* on a finalement une classe de moins	*/
	  last_class--;

	  compar = sComparOpeGenClass;

	  qsort(ope, nb_ope, sizeof(* ope), compar);

      }
   }


		/* retourne le nombre de classes trouvees	*/

   return (last_class - cur_clas +1);

}


/* ----------------------------------------------------	*/
/* sComputeR1ClassesOnly: computes class on operonic	*/
/* genes, the gen genes are assumed to be all in the	*/
/* same genomic class					*/
/* the class are numbered from cur_class		*/
/* return the number of classes	found			*/
/* ----------------------------------------------------	*/
static int sComputeR1ClassesOnly(GenomeSeqPtr *gen, int nb_gen, int nseuil,
			int circular, int minn, int maxn,
			int cur_clas, int verbose)
{
  int           (*compar) (const void *, const void *);
  int           i, last_class, i_prec_homol, i_first_homol,
		i_last_homol;

		/* first, sort genomic genes on the ignore	*/
		/* (i.e. rel_number) of				*/
		/* their best operon gene hit			*/
  compar = sComparGenOpeNum;

  qsort(gen, nb_gen, sizeof(* gen), compar);

  		/* calcul des classes pour les homologues		*/
		/* attention tous les gen doivent avoir un homologue	*/
		/* ou bien doivent avoir homol_ptr == NOT_G		*/
		/* SURTOUT PAS NULL...sinon plante			*/
   
#if 0 /* debug */
{
  int k;
       fprintf(stderr,"==================== R1R1R1 ============\n");
   for (k = 0; k < nb_gen; k++) {
	       fprintf(stderr,"after sComputeR1ClassesOnly: %s %5d %4d %s %5d %4d\n",
			       gen[k]->gene_name, gen[k]->rel_number, gen[k]->class,
			       gen[k]->homol_ptr->mnemo, gen[k]->homol_ptr->ignore, gen[k]->homol_ptr->class);

   }
}
#endif

   last_class = cur_clas;


   i = 0;

   while(i < nb_gen && gen[i]->homol_ptr == NOT_O)
         i++;

   if (i >= nb_gen)
     return 0;

			/* precedent gen_ptr ayant un homologue	*/
   i_first_homol = i_prec_homol = i;

   gen[i]->class = gen[i]->homol_ptr->class = last_class;


   for ( ++i ; i < nb_gen; i++) {

     if(verbose)
	(void)fprintf(stderr, "\r" COMPLETION_TAG "sequence %d / %d\t\t", i+1, nb_gen);

				/* si pas d'homologue, on	*/
				/* saute au suivant		*/
     if (gen[i]->homol_ptr == NOT_O) {
	gen[i]->class = NO_CLASS;
	continue;
     }

			/* on regarde si la distance en gene	*/
			/* est adequate pour mettre le		*/
			/* gen[i]->homolog dans la meme classe	*/
			/* que le gen[i_prec_homol]->homolog	*/
			/* note: on verifie quand meme qu'ils	*/
			/* n'ont pas le meme ignore (i.e.	*/
			/* rel_num (i.e. que			*/
			/* ce n'est pas le meme ope qui est	*/
			/* homologue a deux "genomic genes"	*/
			/* successifs				*/
			/* sinon, on cree une nouvelle classe	*/
				/* ABS(diff)-1 < nseuil		*/
				/* car si on a			*/
				/* i   j			*/
				/* 12345...			*/
				/* on considere que la distance	*/
				/* en gene de i a j est de	*/
				/* j-i-1, cad 5-1-1->3 genes	*/
      if ((gen[i]->homol_ptr->ignore !=
	     gen[i_prec_homol]->homol_ptr->ignore)
	     && (ABS(gen[i]->homol_ptr->ignore - 
	     gen[i_prec_homol]->homol_ptr->ignore)-
	     1 <= nseuil)) {
	   gen[i]->homol_ptr->class =
			gen[i]->class = last_class;
      }
      else {
	   gen[i]->homol_ptr->class =
			gen[i]->class = ++last_class;
      }

      i_prec_homol = i;
   }

   i_last_homol = i_prec_homol;


   if (circular) {
			/* si circularite, et qu'il y a plus	*/
			/* d'une classe				*/
			/* mettre la classe du (des) dernier(s)	*/
			/* ope(s) au meme numero que la classe	*/
			/* du premier ope (ou l'inverse) si	*/
			/* leur distance est correcte		*/

      if (last_class > 0 &&
          (ABS(maxn - gen[i_last_homol]->homol_ptr->ignore) +
	  ABS(gen[i_first_homol]->homol_ptr->ignore-
	  minn) <= nseuil)) {

			/* descent from the last until the	*/
			/* class is not the same		*/
			/* and set the class of the last	*/
			/* elements to the first class		*/
	  i = i_last_homol;
	  while(i >= 0) {
	    if (gen[i]->homol_ptr != NOT_O) {
	        if (gen[i]->homol_ptr->class == last_class)
	           gen[i]->homol_ptr->class =
			gen[i]->class = 
			  gen[i_first_homol]->homol_ptr->class;
		else
		   break;
	     }
             i--;
	  }
			/* on a finalement une classe de moins	*/
	  last_class--;

	  compar = sComparGenOpeClass;

	  qsort(gen, nb_gen, sizeof(* gen), compar);

      }
   }


		/* retourne le nombre de classes trouvees	*/

   return (last_class - cur_clas +1);

}


/* ----------------------------------------------------	*/
/* sComputeR2ClassesOnR1: computes class on genomic	*/
/* genes, taking into account the fact that ope genes	*/
/* are also classified in classes			*/
/* Note: ope[i]->class = NO_CLASS or a value in the	*/
/* range [012345...nclass_ope-1]			*/
/* minn is the minimum numero of ope, maxn is the	*/
/* maximum one						*/
/* returns the number of classes			*/
/* ----------------------------------------------------	*/
static int sComputeR2ClassesOnR1(OperonSeqPtr *ope, int nb_ope, int nseuil,
			int ope_circular, int gen_circular,
			int mingn, int maxgn,
			int minon, int maxon, int verbose)
{
  int           (*compar) (const void *, const void *);
  int           i, i_prec, last_class, cur_class, prec_class,
		i_first_class, i_last_class;

		/* sort operon genes on their class	*/
   compar = sComparOpeClass;

   qsort(ope, nb_ope, sizeof(* ope), compar);

  		/* calcul des classes pour les		*/
		/* homologues: attention tous les ope	*/
		/* doivent avoir un homologue ou bien	*/
		/* un homolog.gen_ptr == NOT_G		*/
		/* SURTOUT PAS NULL...sinon plante	*/

		/* les ope dont la class est NO_CLASS	*/
		/* ou SINGLETON	sont en premiers	*/
   i = 0;
   while (i < nb_ope && (ope[i]->class == NO_CLASS ||
			ope[i]->class == SINGLETON))
      i++;

   if (i >= nb_ope)
     return 0;

		/* a partir d'ici, ope ont forcement	*/
		/* un homolog				*/

   cur_class = 0;
   i_first_class = i;


   do {

      prec_class = ope[i]->class;
      i_prec = i;

      if(verbose)
	(void)fprintf(stderr, "\r" COMPLETION_TAG "sequence %d / %d\t\t", i+1, nb_ope);

      do {
        i++;
      } while(i < nb_ope && ope[i]->class == prec_class);

			/* i-i_prec est le nb d'ope entre i et i_prec-1	*/
      if (i-i_prec > 1) {
	 cur_class += sComputeR2ClassesOnly(&(ope[i_prec]), i-i_prec, nseuil,
			0, 0, 0, cur_class, 0);
         i_last_class = i-1;
      }
      else {			/* i == i_prec + 1		*/
				/* ope a forcement un homolog	*/
				/* ici, classe singleton	*/
	 ope[i_prec]->class =
		   ope[i_prec]->homolog.gen_ptr->class =  cur_class;
         cur_class++;
         i_last_class = i_prec;

      }

   } while (i < nb_ope);

   last_class = cur_class - 1;

   if(verbose)
      (void)fprintf(stderr, "\n");


   if (ope_circular) {
			/* si circularite, et qu'il y a plus	*/
			/* d'une classe				*/
			/* mettre la classe du (des) dernier(s)	*/
			/* gene(s) au meme numero que la classe	*/
			/* du premier gene (ou l'inverse) si	*/
			/* leur distance est correcte		*/

      if (i_last_class > 0 &&
			/* premier et dernier ope sont proches	*/
          (ABS(maxon - ope[i_last_class]->ignore) +
	  ABS(ope[i_first_class]->ignore-
	  minon) <= nseuil) &&
			/* et leurs homologues aussi	*/
          ((ABS(ope[i_last_class]->homolog.gen_ptr->rel_number -
	  ope[i_first_class]->homolog.gen_ptr->rel_number-1) <= nseuil) ||
			/* ou leurs homologues aussi	*/
			/* mais circular genome		*/
          (! gen_circular ||
	  (ABS(maxgn - ope[i_last_class]->homolog.gen_ptr->rel_number) +
	  ABS(ope[i_first_class]->homolog.gen_ptr->rel_number-
	  mingn) <= nseuil)))
	  ) {

			/* descent from the last until the	*/
			/* class is not the same		*/
			/* and set the class of the last	*/
			/* elements to the first class		*/
	  i = i_last_class;
	  while(i >= 0) {
	    if (ope[i]->class != NO_CLASS && ope[i]->class != SINGLETON) {
	        if (ope[i]->class == last_class)
	           ope[i]->class = ope[i]->homolog.gen_ptr->class =
			ope[i_first_class]->class;
		else
		   break;
	     }
             i--;
	  }
			/* on a finalement une classe de moins	*/
	  last_class--;

	  compar = sComparOpeGenClass;

	  qsort(ope, nb_ope, sizeof(* ope), compar);

      }
   }



   return last_class+1;

}


/* ----------------------------------------------------	*/
/* sComputeR1ClassesOnR2: computes class on operonic	*/
/* genes, taking into account the fact that gen genes	*/
/* are also classified in classes			*/
/* Note: gen[i]->class = NO_CLASS or a value in the	*/
/* range [012345...nclass_gen-1]			*/
/* minn is the minimum numero of gen, maxn is the	*/
/* maximum one						*/
/* returns the number of classes			*/
/* ----------------------------------------------------	*/
static int sComputeR1ClassesOnR2(GenomeSeqPtr *gen, int nb_gen,
			int nseuil,
			int gen_circular, int ope_circular,
			int minon, int maxon,
			int mingn, int maxgn,
			int verbose)
{
  int           (*compar) (const void *, const void *);
  int           i, i_prec, last_class, cur_class, prec_class,
		i_first_class, i_last_class;

		/* sort genomic genes on their class 	*/
   compar = sComparGenClass;

   qsort(gen, nb_gen, sizeof(* gen), compar);

  		/* calcul des classes pour les		*/
		/* homologues: attention tous les gen	*/
		/* doivent avoir un homologue		*/
		/* ou bien homol_ptr == NOT_G		*/
		/* SURTOUT PAS NULL...sinon plante	*/

		/* les gen dont la class est NO_CLASS	*/
		/* ou SINGLETON	sont en premiers	*/
   i = 0;
   while (i < nb_gen && (gen[i]->class == NO_CLASS || 
			 gen[i]->class == SINGLETON))
      i++;

   if (i >= nb_gen)
     return 0;

		/* a partir d'ici, gen ont forcement	*/
		/* un homolog				*/

   cur_class = 0;
   i_first_class = i;


   do {

      prec_class = gen[i]->class;
      i_prec = i;

      if(verbose)
	(void)fprintf(stderr, "\r" COMPLETION_TAG "sequence %d / %d\t\t", i+1, nb_gen);

      do {
        i++;
      } while(i < nb_gen && gen[i]->class == prec_class);
			
			/* tous les gen entre i_prec et i-1 sont	*/
			/* dans la meme classe				*/
			/* i-i_prec == le nb de gen entre i_prec et i-1	*/
      if (i-i_prec > 1) {

	 cur_class += sComputeR1ClassesOnly(&(gen[i_prec]), i-i_prec, nseuil,
			0, 0, 0, cur_class, 0);
         i_last_class = i-1;
      }
      else {			/* i == i_prec + 1		*/
				/* gen a forcement un homolog	*/
				/* ici, classe singleton	*/
	 gen[i_prec]->class =
		   gen[i_prec]->homol_ptr->class =  cur_class;
         cur_class++;
         i_last_class = i_prec;

      }

   } while (i < nb_gen);

   last_class = cur_class - 1;

   if(verbose)
      (void)fprintf(stderr, "\n");


   if (gen_circular) {
			/* si circularite, et qu'il y a plus	*/
			/* d'une classe				*/
			/* mettre la classe du (des) dernier(s)	*/
			/* ope(s) au meme numero que la classe	*/
			/* du premier ope (ou l'inverse) si	*/
			/* leur distance est correcte		*/

      if (i_last_class > 0 &&
			/* premier et dernier gen sont proches	*/
          (ABS(maxgn - gen[i_last_class]->rel_number) +
	  ABS(gen[i_first_class]->rel_number-
	  mingn) <= nseuil) &&
			/* et leurs homologues aussi	*/
          ((ABS(gen[i_last_class]->homol_ptr->ignore -
	  gen[i_first_class]->homol_ptr->ignore-1) <= nseuil) ||
			/* ou leurs homologues aussi	*/
			/* mais circular operon		*/
          (! ope_circular ||
	  (ABS(maxon - gen[i_last_class]->homol_ptr->ignore) +
	  ABS(gen[i_first_class]->homol_ptr->ignore-
	  minon) <= nseuil)))
	  ) {

			/* descent from the last until the	*/
			/* class is not the same		*/
			/* and set the class of the last	*/
			/* elements to the first class		*/
	  i = i_last_class;
	  while(i >= 0) {
	    if (gen[i]->class != NO_CLASS && gen[i]->class != SINGLETON) {
	        if (gen[i]->class == last_class)
	           gen[i]->class = gen[i]->homol_ptr->class =
			gen[i_first_class]->class;
		else
		   break;
	     }
             i--;
	  }
			/* on a finalement une classe de moins	*/
	  last_class--;

	  compar = sComparGenOpeClass;

	  qsort(gen, nb_gen, sizeof(* gen), compar);

      }
   }



   return last_class+1;

}



/* ----------------------------------------------------	*/
/* ComputeR1Classes: find the class of all operon genes	*/
/* which have an homolog in genomic genes		*/
/* returns: number of classes (-1 if error) 		*/
/* ----------------------------------------------------	*/
int ComputeR1Classes(OperonSeqPtr *ope, int nb_ope, int nseuil,
			int operon_circular, int minn, int maxn,
			int verbose)
{

  int		(*compar) (const void *, const void *);
  int		i, last_class, i_prec_homol, i_first_homol,
		i_last_homol;


  if (verbose)
     fprintf(stderr, "\n" COMPLETION_TAG "Computing operon genes classes:\n");

  compar = sComparOpeNum;

  qsort(ope, nb_ope, sizeof(* ope), compar);

  		/* voir combien de classes pour les homologues		*/
		/* attention tous les ope doivent avoir un homologue	*/
		/* ou bien doivent avoir homolog.gen_ptr == NOT_G	*/
		/* SURTOUT PAS NULL...sinon plante			*/
   
   last_class = 0;

   i = 0;

				/* if ignore is 0, operon must	*/
				/* be ignored, else it is its	*/
				/* relative position		*/
				/* (counted in gene on the	*/
				/* operon suite)		*/
   while(i < nb_ope && (ope[i]->ignore == 0 || ope[i]->homolog.gen_ptr == NOT_G)) {
	 if (ope[i]->homolog.gen_ptr == NOT_G)
	    ope[i]->class = NO_CLASS;
	 else {
	    ope[i]->class = ope[i]->homolog.gen_ptr->class = SINGLETON;
         }
	 i++;
   }

   if (i >= nb_ope)
     return 0;


				/* precedent ope_ptr ayant	*/
				/* une position relative 	*/
   i_first_homol = i_prec_homol = i;

   ope[i_first_homol]->class = last_class;

   for ( ++i ; i < nb_ope; i++) {

     if(verbose)
	(void)fprintf(stderr, "\r" COMPLETION_TAG "sequence %d / %d\t\t", i+1, nb_ope);

				/* si pas d'homologue	*/
				/* on passe au suivant	*/
     if (ope[i]->homolog.gen_ptr == NOT_G) {
	ope[i]->class = NO_CLASS;
	continue;
     }
     else if (ope[i]->ignore == -1 || ope[i]->ignore == IGNORE) {
				/* si pas de numero	*/
				/* on saute au suivant	*/
				/* en notant SINGLETON	*/
	 ope[i]->class = ope[i]->homolog.gen_ptr->class = SINGLETON;
	 continue;
     }
				/* ABS(diff)-1 < nseuil		*/
				/* car si on a			*/
				/* i   j			*/
				/* 12345...			*/
				/* on considere que la distance	*/
				/* en gene de i a j est de	*/
				/* j-i-1, cad 5-1-1->3 genes	*/
      if (ABS(ope[i]->ignore - 
	ope[i_prec_homol]->ignore)-1 <= nseuil)
	ope[i]->class = last_class;
      else
	ope[i]->class = ++last_class;

      i_prec_homol = i;
   }

   if(verbose)
	(void)fprintf(stderr, "\n");

   i_last_homol = i_prec_homol;

   if (operon_circular) {
			/* si circularite, et qu'il y a plus	*/
			/* d'une classe				*/
			/* mettre la classe du (des) dernier(s)	*/
			/* gene(s) au meme numero que la classe	*/
			/* du premier gene 	*/

      if (last_class > 0 &&
	 (ABS(maxn - ope[i_last_homol]->ignore) +
	 ABS(ope[i_first_homol]->ignore-minn) <= nseuil)) {

			/* descent from the last until the	*/
			/* class is not the same		*/
			/* and set the class of the last	*/
			/* elements to the first class		*/
	  i = i_last_homol;
	  while(i >= 0) {
	     if (ope[i]->homolog.gen_ptr != NOT_G) {
	        if (ope[i]->class == last_class)
	           ope[i]->class = ope[i_first_homol]->class;
		else
		   break;
	     }
             i--;
	  }
			/* la derniere classe est effacee...	*/
	  last_class--;

	  compar = sComparOpeClass;

	  qsort(ope, nb_ope, sizeof(* ope), compar);

      }

   }

			/* le nombre de classe est le numero	*/
			/* de la derniere classe + 1		*/
   return last_class+1;

}

/* ----------------------------------------------------	*/
/* sSortOpeClassesAndSingleton:				*/
/* purpose: sort and set the singleton			*/
/* ope array is assumed to be sorted on homolog gen	*/
/* classes, and sSetOpeSingleton apply a second		*/
/* sorting on each class based on the gen rel_num	*/
/* set the class of the operons which are		*/
/* in single couple with a gene to SINGLETON		*/
/* ----------------------------------------------------	*/
void sSortOpeClassesAndSingleton(OperonSeqPtr *ope, int nb_ope)
{
   int		(*compar) (const void *, const void *);
   int		i, i_prec, prec_class, j, cur_class;

   compar = sComparOpeGenClass;

   qsort(ope, nb_ope, sizeof(* ope), compar);


   i = 0;
   cur_class = 0;

   while (i < nb_ope && (ope[i]->homolog.gen_ptr == NOT_G ||
	  ope[i]->homolog.gen_ptr->class == NO_CLASS ||
	  ope[i]->homolog.gen_ptr->class == SINGLETON) )
      i++;

   if (i >= nb_ope)
     return ;

   compar = sComparOpeGenNum;

   do {

      prec_class = ope[i]->homolog.gen_ptr->class;
      i_prec = i;

      do {
        i++;
      } while(i < nb_ope && ope[i]->homolog.gen_ptr->class == prec_class);

            			/* i == i_prec+1		*/
				/* ici, classe singleton	*/
      if (i-i_prec == 1) {
	      ope[i_prec]->class = 
		   ope[i_prec]->homolog.gen_ptr->class = SINGLETON;
      }
      else {
	 for (j = i_prec; j < i; j++) {
	    ope[j]->class =
	    	ope[j]->homolog.gen_ptr->class = cur_class;
	 }
         
	 qsort(&(ope[i_prec]), i-i_prec, sizeof(* ope), compar);
         cur_class++;
      }


   } while (i < nb_ope);

   return;
}

/* ----------------------------------------------------	*/
/* sSetOpeSingleton:					*/
/* purpose: sort and set the singleton			*/
/* ope array is assumed to be sorted on homolog gen	*/
/* classes, and sSetOpeSingleton apply a second		*/
/* sorting on each class based on the gen rel_num	*/
/* set the class of the operons which are		*/
/* in single couple with a gene to SINGLETON		*/
/* ----------------------------------------------------	*/
void sSetOpeSingleton(OperonSeqPtr *ope, int nb_ope)
{
  int		(*compar) (const void *, const void *);
  int		i, i_prec, prec_class, j, cur_class;

   compar = sComparOpeGenClass;

   qsort(ope, nb_ope, sizeof(* ope), compar);


   i = 0;
   cur_class = 0;

   while (i < nb_ope && (ope[i]->homolog.gen_ptr == NOT_G ||
	  ope[i]->homolog.gen_ptr->class == NO_CLASS ||
	  ope[i]->homolog.gen_ptr->class == SINGLETON) )
      i++;

   if (i >= nb_ope)
     return ;

   compar = sComparOpeGenNum;

   do {

      prec_class = ope[i]->homolog.gen_ptr->class;
      i_prec = i;

      do {
        i++;
      } while(i < nb_ope && ope[i]->homolog.gen_ptr->class == prec_class);

            			/* i == i_prec+1		*/
				/* ici, classe singleton	*/
      if (i-i_prec == 1) {
	      ope[i_prec]->class = 
		   ope[i_prec]->homolog.gen_ptr->class = SINGLETON;
      }
      else {
	 for (j = i_prec; j < i; j++) {
	    ope[j]->class =
	    	ope[j]->homolog.gen_ptr->class = cur_class;
	 }
         
	 qsort(&(ope[i_prec]), i-i_prec, sizeof(* ope), compar);
         cur_class++;
      }


   } while (i < nb_ope);

   return;
}


/* ----------------------------------------------------	*/
/* sSetGenSingleton:					*/
/* purpose: sort and set the singleton			*/
/* gen array is assumed to be sorted on homolog ope	*/
/* classes, and sSetGenSingleton apply a second		*/
/* sorting on each class based on the ope ignore (i.e.	*/
/* rel_num						*/
/* set the class of the genomic genes which are		*/
/* in single couple with a ope to SINGLETON		*/
/* ----------------------------------------------------	*/
void sSetGenSingleton(GenomeSeqPtr *gen, int nb_gen)
{
  int		(*compar) (const void *, const void *);
  int		i, i_prec, prec_class, j, cur_class;

   compar = sComparGenOpeClass;

   qsort(gen, nb_gen, sizeof(* gen), compar);


   i = 0;
   cur_class = 0;

   while (i < nb_gen && (gen[i]->homol_ptr == NOT_O ||
	  gen[i]->homol_ptr->class == NO_CLASS ||
	  gen[i]->homol_ptr->class == SINGLETON) )
      i++;

   if (i >= nb_gen)
     return ;

   compar = sComparGenOpeNum;

   do {

      prec_class = gen[i]->homol_ptr->class;
      i_prec = i;

      do {
        i++;
      } while(i < nb_gen && gen[i]->homol_ptr->class == prec_class);

            			/* i == i_prec+1		*/
				/* ici, classe singleton	*/
      if (i-i_prec == 1) {
	      gen[i_prec]->class = 
		   gen[i_prec]->homol_ptr->class = SINGLETON;
      }
      else {
	 for (j = i_prec; j < i; j++) {
	    gen[j]->class =
	    	gen[j]->homol_ptr->class = cur_class;
	 }
         
	 qsort(&(gen[i_prec]), i-i_prec, sizeof(* gen), compar);
         cur_class++;
      }


   } while (i < nb_gen);

   return;
}
