/* ---------------------------------------------------	*/
/* combi.c						*/
/* <jompo> Jul 2000					*/
/* ---------------------------------------------------	*/

#include <stdio.h>
#include <stdlib.h>
#ifndef Darwin
#include <malloc.h>
#endif
#if (defined Linux) || (defined Darwin)
#include <string.h>		/* for memcpy	*/
#endif

#include "operon.h"
#include "class.h"
#include "util.h"
#include "tags.h"

#undef DEBIG			/* for debug printing	*/

#ifndef ABS
#define ABS(x)              ((x) >= 0  ? (x) : -(x))
#endif

typedef struct {
	char marked;		/* 0: this couple has	*/
				/* been tested, 1: it	*/
				/* has not been tested	*/
	short io;		/* operon index		*/
	short ig;		/* gene index		*/
	short next_diff_pg;	/* numero du prochain	*/
				/* DbleIndice avec un	*/
				/* gene different	*/
} DbleIndice, *DbleIndicePtr;

typedef struct {
	short io;		/* operon index		*/
	short ig;		/* gene index		*/
} DbleIndx, *DbleIndxPtr;

#define START_SIZE 5

typedef struct {
  DbleIndxPtr dbi;	/* array of operon index	*/
  int nb;		/* number of el. in dbi	*/
  int realsize;		/* real size in memory		*/
} Liste, *ListePtr;

typedef struct {
   DbleIndicePtr	dbi;
   int			size;
   OperonSeqPtr*	ope;
} Table, *TablePtr;


#define NEXT(t,i)	(t)->dbi[(i)].next_diff_pg
#define OPE(t,i)	sOpe[t->dbi[(i)].io]
#define GEN(t,i)	sOpe[t->dbi[(i)].io]->gen_kept[(t)->dbi[(i)].ig].gen_ptr
#define RELNUM(t,i)	GEN((t),(i))->rel_number
#define DISTGENRELNUM(t,i,j)	RELNUM(t,j) - RELNUM(t,i)

#define TESTED(t,i)	(t->dbi[i].marked == (char)1)

#define NO_NEXT	0

static OperonSeqPtr	*sOpe = NULL;

/* ---------------------------------------------------	*/
/* for qsort to sort on rel_number of gen_kept		*/
/* ---------------------------------------------------	*/
static int sComparDbleIndice(const void *dbli1, const void *dbli2)
{
   GenomeSeqPtr pg1, pg2;

   pg1 = sOpe[((DbleIndicePtr)dbli1)->io]->gen_kept[((DbleIndicePtr)dbli1)->ig].gen_ptr;
   pg2 = sOpe[((DbleIndicePtr)dbli2)->io]->gen_kept[((DbleIndicePtr)dbli2)->ig].gen_ptr;


   if ( pg1->rel_number > pg2->rel_number) {
      return +1;
   }
   else if ( pg1->rel_number < pg2->rel_number) {
      return -1;
   }
   else
      return 0;

}

/* ---------------------------------------------------	*/
/* alloue la table d'indice				*/
/* et la remplit					*/
/* ---------------------------------------------------	*/
TablePtr sCreateTable(OperonSeqPtr *ope, int nope, int seuil_score) 
{
   long			n;
   int			i, k, nostop, prec_i;
   GenomeSeqPtr		prec_pg;
   TablePtr		tptr;
   static Table		table;

   tptr = &table;

   n = 0;

			/* filtre les scores inf au	*/
			/* seuil, et ajuste en		*/
			/* consequence les		*/
			/* actual_gen_kept		*/
   for (i = 0; i < nope; i++) {
      nostop = 1;
      for(k = 0; k < ope[i]->actual_gen_kept && nostop; k++) {
	if (ope[i]->gen_kept[k].score < seuil_score) {
	  nostop = 0;
	  ope[i]->actual_gen_kept = k;
        }
      }
      n += ope[i]->actual_gen_kept;
   }
   
   if (n == 0) {
      MyError(stderr, 0, ERROR_TAG,
	 "=========== RESULT ====================\n");
      MyError(stderr, 1, ERROR_TAG,
	 "No couple operon/gene has a score > %d\n",
		seuil_score);
   }

   if (! (tptr->dbi=NEWN(DbleIndice, n))) {
      MyError(stderr, 1, ERROR_TAG,
	 "sCreateTable: cannot allocate for tptr->newd array(%d)\n",
	n);
   }

   tptr->size = n;
   tptr->ope = ope;
   sOpe = ope;

   n = 0;
   for (i = 0; i < nope; i++) {
      for (k = 0; k < ope[i]->actual_gen_kept; k++) {
         tptr->dbi[n].io = i;
         tptr->dbi[n].ig = k;
	 tptr->dbi[n].marked = (char) 0;
         n++;
      }
   }


#if 0
   for (i = 0; i < n; i++) {
      fprintf(stderr,"before %3d %3d-> %5d\n",tptr->dbi[i].io, tptr->dbi[i].ig,
	RELNUM(tptr,i));
   }
#endif

			/* needs initialisation of sOpe		*/
			/* before qsort				*/

   qsort((void *) (tptr->dbi), (size_t) n, (size_t) sizeof(*(tptr->dbi)),
	     sComparDbleIndice);

#if 0
   for (i = 0; i < n; i++) {
      fprintf(stderr,"after %3d %3d -> %5d (name %s)\n",tptr->dbi[i].io, tptr->dbi[i].ig,
        RELNUM(tptr,i),
	GEN(tptr,i)->fseqptr->name);
   }
#endif

			/* set the next indices	(for speed)	*/

   i = 0;
   prec_pg = GEN(tptr,i);
   prec_i  = 0;

			/* on se base sur le pointeur pour	*/
			/* pour savoir si on a un gene 		*/
			/* different 				*/
   while( ++i < tptr->size) {
	if (GEN(tptr,i) != prec_pg) {
	  for (k = i-1; k >= prec_i; k--) 
	     tptr->dbi[k].next_diff_pg = i;
	  prec_pg = GEN(tptr,i);
	  prec_i = i;
	}
   }
   for (i = prec_i; i < tptr->size; i++)
      tptr->dbi[i].next_diff_pg = NO_NEXT;			/* pas de next	*/
      
#ifdef DEBIG
   for (i = 0; i < n; i++) {
      fprintf(stderr,"[%d] %3d %3d %5d -> %5d (name %s)\n",
	i, tptr->dbi[i].io, tptr->dbi[i].ig, RELNUM(tptr,i),
	NEXT(tptr,i),
	GEN(tptr,i)->fseqptr->name);
   }
#endif
   return tptr;

}


/* ---------------------------------------------------	*/
/* fonctions sur les listes				*/
/* ---------------------------------------------------	*/
static void sFreeList(ListePtr li)
{
   if (li != NULL) {
     if (li->dbi)
        free(li->dbi);
     free(li);
   }
   return;

}

/* ---------------------------------------------------	*/
static ListePtr sNewList(int siz)
{
  ListePtr l;

  if (!(l=NEW(Liste)))
     MyError(stderr, 9, ERROR_TAG,
	"Cannot allocate list");

  if (!(l->dbi = NEWN(DbleIndx, siz)))
     MyError(stderr, 9, ERROR_TAG,
	"Cannot allocate array in list");

  l->realsize = siz;
  l->nb = 0;

  return l;
}

#if 0
/* ---------------------------------------------------	*/
static ListePtr sInitList(DbleIndicePtr pdbi)
{
  ListePtr	li;

  if (!(li=sNewList(START_SIZE)))
    return NULL;

  li->dbi[0].io = pdbi->io;
  li->dbi[0].ig = pdbi->ig;
  li->nb = 1;

  return li;
}
#endif

/* ---------------------------------------------------	*/
static void sAddList(ListePtr l, DbleIndicePtr pdbi)
{
  int j,k;

  if (l->realsize <= l->nb +1) {
      if (!(l->dbi=REALLOC(DbleIndx, l->dbi, l->realsize+START_SIZE)))
	 MyError(stderr, 9, ERROR_TAG,
	    "Cannot reallocate array in list");
      else
	 l->realsize = l->realsize+START_SIZE;
  }

			/* la liste doit etre triee	*/
			/* en io croissant		*/
  j = 0;
  while (j < l->nb && pdbi->io > l->dbi[j].io)
     j++;

		/* si j == nb, sort du for	*/
		/* et met t_io[nb] a new_i	*/
		/* sinon decale a droite	*/
		/* jusqu'a nb			*/
  for (k = l->nb; k > j; k--) {
    l->dbi[k].io = l->dbi[k-1].io;
    l->dbi[k].ig = l->dbi[k-1].ig;
  }

		/* et met new_i a la place	*/
		/* liberee			*/
  l->dbi[j].io = pdbi->io;
  l->dbi[j].ig = pdbi->ig;

  l->nb++;

  return;

}

/* ---------------------------------------------------	*/
static ListePtr sDupList(ListePtr src)
{
  ListePtr	dst;


  if (!(dst=sNewList(src->realsize)))
    return NULL;

  memcpy((void *)dst->dbi, (void *)src->dbi,
      (size_t) src->nb * sizeof(*(src->dbi)));


  dst->nb = src->nb;
  dst->realsize = src->realsize;

  return dst;

}

/* ---------------------------------------------------	*/
static void sVideList(ListePtr l)
{
   l->nb = 0;

   return;

}

/* ---------------------------------------------------	*/
int sIsOpeOfDbiInList(DbleIndicePtr dbiask, ListePtr l)
{
   int j;

   j = 0;
			/* tant que dbiask->io est <	*/
			/* a l->dbi[j].io		*/
			/* qui est triee en io		*/
			/* croissants			*/
   while (j < l->nb && l->dbi[j].io < dbiask->io) {
     j++;
   }

   if (l->dbi[j].io == dbiask->io)
      return 1;
   else
      return 0;

}

/* ---------------------------------------------------	*/
void sPrintOperon(ListePtr lis, int stat_on)
{
   int		i;
   float	len_ratio;
   static int	numop = 1;


   fprintf(stdout,"=== group %4d ===\n",numop);
   numop++;

   for(i=0;i<lis->nb;i++) {
     fprintf(stdout,"%s (num %4d) with %s (num %4d, score %3d",
	sOpe[lis->dbi[i].io]->mnemo,
	sOpe[lis->dbi[i].io]->ignore,
	sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].gen_ptr->gene_name,
	sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].gen_ptr->rel_number,
	sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].score);

     if (stat_on)
       (void)fprintf(stdout," prob %.3e", 
	  sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].stat);

     len_ratio = LENRATIO(sOpe[lis->dbi[i].io]->fseqptr->length, 
			sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].gen_ptr->fseqptr->length);


     (void)fprintf(stdout," dlen ratio %5.2f)\n", len_ratio);
   }
}

/* ---------------------------------------------------	*/
/* sCherche: recherche dans la table->dbi un prochain	*/
/* candidat						*/
/* ---------------------------------------------------	*/
static int sCherche(int dbistart, ListePtr lis, TablePtr table,
				int seuil_dist, int stat_on)
{
 int		dbj, found, retval, lim;
 ListePtr	lis2;

#ifdef DEBIG
int i;
#endif

     retval = 0;
     found = 0;

     dbj = NEXT(table,dbistart);	/* next dbi which gene diff	*/
					/* from istart gene		*/

					/* si fin de liste OU	*/
					/* si dbj a deja ete	*/
					/* "traite"		*/
     if (dbj == NO_NEXT || TESTED(table,dbj))
       return retval;

#ifdef DEBIG
{
fprintf(stderr,"sCherche:dbistart=%4d dbj=%4d  ", dbistart, dbj);
fprintf(stderr,"dist=%4d (seuil=%2d)\n",
	DISTGENRELNUM(table,dbistart, dbj), seuil_dist);
for(i=0;i<lis->nb;i++)
  fprintf(stderr,"%d(%d),",lis->dbi[i].io,
     sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].gen_ptr->rel_number);
fprintf(stderr,"\n");
}
#endif

     lim = NEXT(table,dbj);
     if (lim == NO_NEXT)
	   lim = table->size;
				/* Attention: DISTGENRELNUM	*/
				/* doit etre employe avec	*/
				/* le dbi le plus petit en	*/
				/* premier !!!!			*/
     while ( dbj < lim &&
	    DISTGENRELNUM(table,dbistart,dbj) <= seuil_dist) {

				/* si dbj a ete traite		*/
	if (TESTED(table,dbj)) {
#ifdef DEBIG
fprintf(stderr,"couple dbj %d deja traite\n", dbj);
#endif
	   dbj++;
	   continue;
				/* si l'operon de dbj n'est pas	*/
				/* dans la liste lis...		*/
	}
	else if (! sIsOpeOfDbiInList(&(table->dbi[dbj]), lis)) {

				/* on l'ajoute dans une liste	*/
				/* dupliquee			*/
#ifdef DEBIGA
fprintf(stderr,"OK1:dbistart=%3d(%3d),dbj=%3d(%3d): dist=%2d\n",
	dbistart,RELNUM(table,dbistart),dbj,RELNUM(table,dbj),
	DISTGENRELNUM(table,dbistart, dbj));
fprintf(stderr,"before Dup lis:");
for(i=0;i<lis->nb;i++)
  fprintf(stderr,"%d(%d),",lis->dbi[i].io,
     sOpe[lis->dbi[i].io]->gen_kept[lis->dbi[i].ig].gen_ptr->rel_number);
fprintf(stderr,"\n");
#endif

	    if (!(lis2 = sDupList(lis)))
	        MyError(stderr, 9, ERROR_TAG,
		   "sCherche:cannot duplicate liste, exiting...\n");

#ifdef DEBIGA
fprintf(stderr,"OK2:dbistart=%3d(%3d),dbj=%3d(%3d): dist=%2d\n",
	dbistart,RELNUM(table,dbistart),dbj,RELNUM(table,dbj),
	DISTGENRELNUM(table,dbistart, dbj));
fprintf(stderr,"OK2:lis2->nb=%d\n",lis2->nb);
for(i=0;i<lis2->nb;i++)
  fprintf(stderr,"%d(%d),",lis2->dbi[i].io,
     sOpe[lis2->dbi[i].io]->gen_kept[lis2->dbi[i].ig].gen_ptr->rel_number);
fprintf(stderr,"\n");
#endif

            sAddList(lis2, &(table->dbi[dbj]));


#ifdef DEBIG
fprintf(stderr,"OK3:dbistart=%3d(%3d),dbj=%3d(%3d): dist=%2d\n",
	dbistart,RELNUM(table,dbistart),dbj,RELNUM(table,dbj),
	DISTGENRELNUM(table,dbistart, dbj));
fprintf(stderr,"after sAddList:");
for(i=0;i<lis2->nb;i++)
  fprintf(stderr,"%d(%d),",lis2->dbi[i].io,
     sOpe[lis2->dbi[i].io]->gen_kept[lis2->dbi[i].ig].gen_ptr->rel_number);
fprintf(stderr,"\n");
#endif

				/* on envoie cette liste au prochain	*/
				/* gene (si il y en a un)		*/
	    found = sCherche(dbj, lis2, table, seuil_dist, stat_on);

#ifdef DEBIG
fprintf(stderr,"OK4:dbistart=%3d(%3d),dbj=%3d(%3d): dist=%2d\n",
	dbistart,RELNUM(table,dbistart),dbj,RELNUM(table,dbj),
	DISTGENRELNUM(table,dbistart, dbj));
fprintf(stderr,"after cherche:");
for(i=0;i<lis2->nb;i++)
  fprintf(stderr,"%d[%s](%d[%s]),",lis2->dbi[i].io,sOpe[lis2->dbi[i].io]->mnemo,
     sOpe[lis2->dbi[i].io]->gen_kept[lis2->dbi[i].ig].gen_ptr->rel_number,
     sOpe[lis2->dbi[i].io]->gen_kept[lis2->dbi[i].ig].gen_ptr->gene_name);
fprintf(stderr,"\n");
#endif
				/* on le marque comme traite	*/
            table->dbi[dbj].marked = (char) 1;

	    if (! found) {
				/* so we print operon only if	*/
				/* !found AND if number of	*/
				/* elements in liste lis is > 1	*/
				/* we will return 1		*/
	       if (lis2->nb > 1) {
#ifdef DEBIG
	fprintf(stdout,"not found print for dbistart=%d dbj=%d\n",dbistart,dbj);
#endif
	          sPrintOperon(lis2, stat_on);
	          retval = 1;
	       }
	    }
	    else {
				/* if found > 0, at least a	*/
				/* new list has been created	*/
				/* bigger than this one		*/
				/* so we go to the next 	*/
	       retval = 1;
	    }

	    sFreeList(lis2);

	}
	
	dbj++;

     }


     return retval;


}

/* ---------------------------------------------------	*/
/* Combinatorique: cherche en construisant un tableau	*/
/* range des meilleurs genes similaires des operons	*/
/* les operons possibles (pas forcement les meilleurs	*/
/* scores)						*/
/* ---------------------------------------------------	*/
int Combinatorique(OperonSeqPtr *ope, int nope,
		ClassParam *cparam, int stat_on)
{

   TablePtr	table;
   ListePtr	lis;
   int		i;

			/* enleve les gen_kept dont le	*/
			/* seuil de score est inf a 	*/
			/* cparam->seuil_score		*/
			/* puis cree la table		*/
			/* ope<->gen			*/
			/* par gen croissants en relnum	*/

   if (! (table = sCreateTable(ope, nope, cparam->seuil_score)))
	 MyError(stderr, 8, ERROR_TAG,
	    "Combinatorique: cannot allocate for indice table\n");


   lis = sNewList(START_SIZE);

   for (i = 0; i < table->size; i++) {
      if (table->dbi[i].marked == (char)0) {
	 sVideList(lis);
	 sAddList(lis,&(table->dbi[i]));
			/* on convertit le seuil nseuil	*/
			/* en vraie distance 		*/
         sCherche(i, lis, table, cparam->nseuil+1, stat_on);
      }
   }
   
   sFreeList(lis);

   return 0;
}

