/* ==================================================== */
/* @File fastp.h					*/
/* @history:						*/
/* @+	Apr. 97 <Gloup> first draft			*/
/* @+	Jul. 98 <jompo> second draft			*/
/* @+	Jul. 00 <jompo> third draft			*/
/* ==================================================== */

#ifndef _H_fastp

#define _H_fastp

/* -------------------------------------------- */
/* constantes programme				*/
/* -------------------------------------------- */

#include <sys/param.h>		/* for MAXPATHLEN	*/

#ifndef BUFSIZ
#define BUFSIZ 1024
#endif

#define STDALPHABET	 "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define DNA_ALPHABET	 "ACGT"

#define Alpha_Size	 26	    /* [A-Z]				*/					
#define Alpha_Bin_Size	 32	    /* 5 bits/symbol			*/

#define Pretty_Name_Len  10	    /* max length name for ali. output	*/
#define Pretty_Out_Len   50	    /* line length on ali. output	*/


#define Seq_Rotator	100		/* be verbose every 100 seq.	*/

#define ErrWarning	 0	    /* shell error codes		*/
#define ErrMemory	10
#define ErrSymbol	11
#define ErrReadSeq	12
#define ErrBadArg	13
#define ErrNotImp	14
#define ErrFilOpe	15
#define ErrFilAcc	16
#define ErrManySq	17

#define SelectFirst	1
#define SelectLast	2
#define SelectLong	3
#define SelectShort	4
#define SelectAlpha	5
#define SelectClique	10

#define SortNumberInc	SelectFirst
#define SortNumberDec	SelectLast
#define SortSizeDec	SelectLong
#define SortSizeInc	SelectShort
#define SortAlphaInc	SelectAlpha
#define SortAlphaDec	6


#define EOS		(char) -1    /* end of encoded sequence	*/
		    /* ------------------------ */
		    /* Program parameters	*/
		    /* ------------------------ */
typedef struct {
    int	 diag_integ;
    int  kuple;
    int  fuple;
    int  fast_lo_threshold;
    int	 sym_replace;
    int  opt_sym_replace;
    int  opt_sym_delete;
    int  short_output;			/* flag to ligthen output */
    int  do_fastp;			/* flag to do fastp	*/
    int  gen_std_fasta_format;		/* is std bank in fasta format	*/
    int  ope_std_fasta_format;		/* is operon file in fasta format	*/
    int  fast_opt_align;
    int  fastp_aliscore;		/* flag to compute score on alignment	*/
					/* rather than # of kuples		*/
    int  fastp_mxscore;			/* flag to compute score on maximum	*/
					/* subsum rather than on all shifted	*/
					/* alignment scoring			*/
    int  opt_verbose;
    int  opt_very_verbose;
    int  opt_only_sim;
    int  bstat;				/* stat: 1: ON, 0: OFF		*/
    /* int  store_scores;	*/		/* store scores: 1: YES, 0: NO(default)		*/
    int  store_aleas;			/* store shuffled seq scores: 1: YES, 0: NO(default)		*/
    int  max_gen_kept;
    int  exclude_self;			/* 0: all genomes genes are scanned, 1: genomes genes	*/
					/* with same name as the operon gene will be discarded	*/
					/* for this operon gene					*/
    char alphabet[Alpha_Size+2];
    char dbfile[MAXPATHLEN];
    char fast_matfile[MAXPATHLEN];
    char fast_alifile[MAXPATHLEN];
} Parameters;


/* -------------------------------------------- */
/* macros					*/
/* -------------------------------------------- */


#ifndef MIN
#define MIN(a, b) 	    ((a) < (b) ? (a) : (b))	
#endif
#ifndef MAX
#define MAX(a, b) 	    ((a) > (b) ? (a) : (b))	
#endif
#define ABS(x) 		    ((x) >= 0  ? (x) : -(x))	

#ifndef NEW
#define NEW(type)		(type *) malloc(sizeof(type))
#define NEWN(type, n)		(type *) malloc((n) * sizeof(type))
#define REALLOC(type, ptr, n) 	(type *) realloc(ptr, (n) * sizeof(type))
#define FREE(ptr)		free(ptr)
#endif


		    /* ------------------------ */
		    /* linked list of positions */
		    /* ------------------------ */
typedef struct _Cell {
    int		    pos;
    struct _Cell    *next;
} Cell;

		    /* ------------------------ */
		    /* HashHeader		*/
		    /* ------------------------ */
typedef struct {
    int	    alphasize;	    /* size of alphabet		    */
    int	    bitshift;	    /* log2(alphasize)		    */
    int	    wsize;	    /* 2^bitshift		    */
    int	    ksize;	    /* word size		    */
    Int32   tsize;	    /* (2^bitshift)^(ksize-1)	    */
    Int32   smask;	    /* symbol mask useful bits	    */
    Int32   wmask;	    /* word mask useful bits	    */
    char    alphabet[Alpha_Size+1];
} HashHeader;

		    /* ------------------------ */
		    /* HashTable		*/
		    /* ------------------------ */
typedef struct {
    HashHeader	header;	    /* header			    */
    Cell    **table;	    /* hashtable		    */
} HashTable;

		    /* ------------------------ */
		    /* HashVector		*/
		    /* ------------------------ */
typedef struct {
    HashHeader	header;	    /* header			    */
    Int32   *counts;	    /* word counts		    */
} HashVector;


		    /* ------------------------ */
		    /* Diagonal Table		*/
		    /* ------------------------ */
typedef struct {
    int diag_size;
    int diag_offset;
    int diag_integ;
    int	*diag;
} DiagTable;

#endif	/* _H_fastp */
