/* ----------------------------------------------------------	*/
/* operon.h : structures for operon				*/
/* ----------------------------------------------------------	*/

#include "fasta_io.h"

#define ORG_NAME_LEN	32
#define GEN_NAME_LEN	32
#define CTG_NAME_LEN	32
#define MNEMOCHAR	ORG_NAME_LEN+GEN_NAME_LEN+32	/* taille en char du mnemo	*/
							/* mnemo: 			*/
							/* <gene_ID>@<org_ID>@<contig_ID> */


#define LENRATIO(l1,l2)	((float)(((l1)>(l2)) ? ((float)(l1))/((float)(l2)) : ((float)(l2))/((float)(l1))))
/* old #define LENRATIO(l1,l2) (float)(ABS((l1 - l2)/(float)(l1))) */

#define IGNORE		0		/* ignore operon		*/
#define UNK_IGNORE      -1		/* flag : unknown ignore	*/
#define UNK_STRAND      'X'		/* flag : unknown strand	*/

#define NO_KEPT         0
#define KEPT            1

#define NO_CLASS	-3
#define SINGLETON	-2
#define NOBIJECTIF	-1

#define NOT_G	(GenomeSeqPtr)1L	/* to say that a OperonSeqPtr->homolog.gen_ptr	*/
					/* is not assigned (but not NULL)		*/
					/* i.e. this means that this OperonSeqPtr have	*/
					/* no GenomeSeqPtr homolog			*/

#define NOT_O	(OperonSeqPtr)1L	/* to say that a GenomeSeqPtr->homol_ptr  */
					/* is not assigned (but not NULL)               */
					/* i.e. this means that this GenomeSeqPtr have  */
					/* no OperonSeqPtr homolog                      */

#define BASE_BEST	10000		/* base of bestfit score, i.e. Fastp score	*/
					/* are in the [0-BASE_BEST] range and bestfit	*/
					/* score are added BASE_BEST to be in the	*/
					/* [BASE_BEST-BASE_BEST+HISCORE] range.		*/
					/* This allows to keep both scores in	*/
					/* the same array and to allow recognising if	*/
					/* a score is a fastp score or a bestfit score	*/
#define MYSHORT int

typedef unsigned short SCORE;
typedef struct OpSequence *OperonSeqPtr;

typedef struct {
   OperonSeqPtr ope_ptr;
   float	stat;
   SCORE	score;
} OpeKept, *OpeKeptPtr;

typedef struct {
   FastaSequencePtr	fseqptr;		/* corresponding FastaSequence		*/
   int			from;			/* beginning position in the chromosome */
   int			to;			/* ending position in the chromosome	*/
   MYSHORT		kept_by;		/* number of operon in which this gene	*/
   MYSHORT		class;			/* number of the class of genomic gene	*/
						/* is kept because similarity		*/
   OpeKeptPtr		ope_kept;		/* array of similar operon genes kept	*/
   OperonSeqPtr		homol_ptr;		/* homolog operonic gene		*/
   MYSHORT		max_ope_kept;		/* max # of kept similar operon genes	*/
   MYSHORT		actual_ope_kept;	/* actual # of kept similar ope. genes	*/
   MYSHORT		bestfitted;		/* gene has been bestfitted against	*/
						/* all ope: used only in the case "no	*/
						/* fastp"				*/
   char			mnemo[MNEMOCHAR];	/* gene mnemo				*/
   int			rel_number;		/* relative number of the gene on the	*/
						/* chromosome				*/
   char			contig_ID[CTG_NAME_LEN];/* ID du contig				*/
   char			org_name[ORG_NAME_LEN];	/* nom de l'organisme			*/
   char			gene_name[GEN_NAME_LEN];/* nom du gene				*/
   char			strand;			/* 'D': direct; 'C': complementary	*/
} GenomeSequence, *GenomeSeqPtr;

typedef struct {
   GenomeSeqPtr gen_ptr;
   float	stat;
   SCORE	score;
} GenKept, *GenKeptPtr;

typedef struct OpSequence {
   FastaSequencePtr	fseqptr;		/* pointer to the Genome sequence	*/
   int			ignore;			/* to include or ignore in search	*/
						/* can also be rel_number of operon	*/
   MYSHORT		kept_by;		/* number of genome gene in which this	*/
						/* gene is kept because similarity	*/
   MYSHORT		max_gen_kept;		/* max # of kept similar genomic genes	*/
   MYSHORT 		actual_gen_kept;	/* actual # of kept similar genes	*/
   MYSHORT 		class;			/* number of the class of operon gene	*/
   GenKeptPtr		gen_kept;		/* array of similar genome gene kept 	*/
   GenKept		homolog;		/* homolog genomic gene		 	*/
   char			mnemo[MNEMOCHAR];	/* operon mnemo	*/
   char			strand;			/* 'D': direct; 'C': complementary	*/
						/* operon gene				*/
} OperonSequence;
 

/* Prototypes	*/

OperonSeqPtr NewOperonSequence(int gen_kept);
GenomeSeqPtr NewGenomeSequence(int ope_kept);
Bool ReadOperonSequence(FILE *streamin, OperonSeqPtr ope, int standard_fasta_format);
void FreeOperonSequence(OperonSeqPtr ope);
void FreeGenomeSequence(GenomeSeqPtr genoseq);
Bool ReadGenomeSequence(FILE *streamin, GenomeSeqPtr genoseq, int standard_fasta_format);
void WriteGenomeSequence(FILE *stream, GenomeSeqPtr genoseq, int len_line);
void KeepGeneInOpeF(OperonSeqPtr p_ope, GenomeSeqPtr p_gen, float score);
void KeepOpeInGeneF(GenomeSeqPtr p_gen, OperonSeqPtr p_ope, float score);
void KeepOpeInGeneB(GenomeSeqPtr p_gen, OperonSeqPtr ope_ptr, float fscore, float stat);
void WriteKeptGene(OperonSeqPtr p_ope, float len_thresh, int correction_score,
		 float normscore, int stat_on, int short_output, FILE *fout);
int ComparBestFitScoreGenKept(const void *gen_ptr1, const void *gen_ptr2);
void set_NULL2NOT_O_Gene(GenomeSeqPtr *gen, int nb_gen);
void SetAllOpeHomologs(OperonSeqPtr *p_ope, int nb_ope, float len_thresh, int seuil_score, int bijectif, int verbose);
void SetAllOpeLastHomologs(OperonSeqPtr *p_ope, int nb_ope, float len_thresh, int seuil_score, int bijectif, int verbose);
void SetOneHomolog(OperonSeqPtr ope, float len_thresh, int seuil_score, int bijectif);
int FindInfScoreInOpeKept(GenomeSeqPtr p_gen, int max);
int IsGenInOpeKeptR(OperonSeqPtr p_ope, GenomeSeqPtr p_gen, int min);
void ResetAllScores(OperonSeqPtr *p_ope, int nb_ope, int corr);
void ComputeStatForOperon(OperonSeqPtr ope_ptr, int *aleascore, int n_alea);
void sort_kept_ope_by_score_in_gene(GenomeSeqPtr p_gen);
