/* ==================================================== */
/* @File opscan.c					*/
/* @history:						*/
/* @+	Jul. 98 <jompo> from dbc to opscan		*/
/* @+	Mar. 00 <jompo> first draft opscan		*/
/* @+	May. 03 <jompo> second draft opscan		*/
/* ==================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Gtypes.h"
#include "fasta_io.h"
#include "opscan.h"
#include "opscan_help.h"
#include "util.h"
#include "fastp_hash.h"
#include "fastp_core.h"
#include "bali.h"
#include "operon.h"
#include "class.h"
#include "combi.h"
#include "stat.h"
#include "tags.h"

#if ((defined IRIX) || (defined Linux) )
#include <getopt.h>
#elif (defined Darwin)
#include <unistd.h>
#endif

#if ((defined SOLARIS) || (defined SunOS))
extern char     *optarg;
extern int      optind;
#endif


#ifndef BUFSIZ
#define BUFSIZ		1024
#endif

#define MAX_GEN_SEQ	5000		/* starting length for array gen	*/
					/* but this length can increase after	*/
					/* by increment of MAX_GEN_SEQ		*/
#define INC_GEN_SEQ	2000

#define MAX_OPE_SEQ	200		/* starting length for array ope	*/
					/* but this length can increase after	*/
					/* by increment of INC_OPE_SEQ	*/
#define INC_OPE_SEQ	1000		/* increment for widen the ope array	*/

#define HISCORE		10000		/* higher score	- a score is in the	*/
					/* 0-HISCORE range			*/

#define NORMSCORE	100

#undef ROTATOR


/* -------------------------------------------- */
/* get int argument				*/
/* -------------------------------------------- */

static int sGetIArg(char *opt, char *arg)
{
    int     iarg;
    char    buffer[BUFSIZ];

    if (   arg
	&& (sscanf(arg, "%d", &iarg) == 1)
	&& (iarg >= 0))
	return iarg;

    (void) sprintf(buffer,  "[GetIArg] bad argument -%s %s",
			    opt, (arg ? arg : "<NULL>"));

    (void) Erreur(buffer, ErrBadArg);

    return -1; /* never reached */
}

/* -------------------------------------------- */
/* get float argument				*/
/* -------------------------------------------- */

static float sGetFArg(char *opt, char *arg)
{
    float     farg;
    char    buffer[BUFSIZ];

    if (   arg && (sscanf(arg, "%f", &farg) == 1))
	return farg;

    (void) sprintf(buffer,  "[GetFArg] bad argument -%s %s",
			    opt, (arg ? arg : "<NULL>"));

    (void) Erreur(buffer, ErrBadArg);

    return -1.0; /* never reached */
}

/* -------------------------------------------- */
/* get char argument				*/
/* -------------------------------------------- */

static int sGetCArg(char *opt, char *arg)
{
    char    buffer[BUFSIZ];

    if (   arg
	&& (sscanf(arg, "%s", buffer) == 1))
	return (int) *buffer;

    (void) sprintf(buffer,  "[GetCArg] bad argument -%s %s",
			    opt, (arg ? arg : "<NULL>"));

    (void) Erreur(buffer, ErrBadArg);

    return -1; /* never reached */
}

#if 0
/* -------------------------------------------- */
/* argument error				*/
/* -------------------------------------------- */

static int sBadArg(int carg, char *varg, int status)
{
    char    buffer[BUFSIZ];

    (void) sprintf(buffer, "option -%c Bad value: %s", carg, varg);

    (void) Erreur(buffer, status);

    return status;
}
#endif

#define PP (void)fprintf(stderr, HEADER_TAG

/* -------------------------------------------- */
/* printout general header			*/
/* -------------------------------------------- */

static void sPrintHeader(int argn, char *argv[], int nb_ope, int nb_gen,
			 Parameters *parms)
{
	int i;


	PP	"/------------------------------------------\n");
	PP	"/ Opscan Version 0.9 Fev. 2000             \n");
	PP	"/------------------------------------------\n");
	PP	"/ running cmd    : ");
		for (i = 0 ; i < argn ; i++)
		    (void)fprintf(stderr, "%s ", argv[i]);
	(void)fprintf(stderr, "\n");		    
	PP	"/ database               : %s\n", parms->dbfile);
	PP	"/ number of operon seq   : %d\n", nb_ope);
	PP	"/ number of genome seq   : %d\n", nb_gen);
	if (parms->do_fastp == 0)
	  PP	"/ option -N : fastp will not be run\n");
	if (parms->bstat)
	  PP	"/ option -Z : statistics will be computed\n");
	if (parms->exclude_self)
	  PP	"/ option -w : genes with same name will be discarded from best hits\n");
	PP	"/------------------------------------------\n");
}
#undef PP

#define PP (void)fprintf(stderr, HEADER_TAG
/* -------------------------------------------- */
/* printout fastp header			*/
/* -------------------------------------------- */

static void sPrintFastpHeader(Parameters *parms)
{

	PP	"/------------------------------------------\n");
	PP	"/ Fastp part:				    \n");
	PP	"/------------------------------------------\n");
	PP	"/ alphabet               : %s\n", parms->alphabet);
	PP	"/ kuple                  : %d\n", parms->kuple);
	PP	"/ fastp diag integ.      : %d\n", parms->diag_integ);
	PP	"/ fastp lower threshold  : %d\n", parms->fast_lo_threshold);

	PP	"/ unknown symbol : ");
	
	if (parms->opt_sym_replace)
	    (void)fprintf(stderr, "replace by %c\n", parms->sym_replace);	
	else if (parms->opt_sym_delete)
	    (void)fprintf(stderr, "delete\n");	
	else
	    (void)fprintf(stderr, "raise error\n");
	    	
 	PP "\n");
}

#undef PP


/* -------------------------------------------- */
/* printout bestfit header			*/
/* -------------------------------------------- */

#define PP (void)fprintf(stderr, HEADER_TAG

static void sPrintBestfitHeader(int nope, int ngen, int fastp_done, char *bestmatfile, 
			 int fglobal, int wopen, int wext)
{
	PP	"/------------------------------------------		 \n");
	PP	"/   Bestfit part                           		 \n");
	PP	"/------------------------------------------		 \n");
	PP	"/ # of operon genes: %d                                 \n", nope);
	if (fastp_done)
	PP	"/ Max # of genome genes by operon: %d                    \n", ngen);
	else
	PP	"/ # of genome genes by operon: %d                        \n", ngen);
	PP	"/ algorithm      : %s\n", (fglobal ? "nws (global)" : "bestfit (local)"));
	PP	"/ scoring matrix : %s\n", bestmatfile				   );  
	PP	"/ gap scheme     : %.1f + (L-1) * %.1f\n",
					(float) wopen / SCORENORM, 
					(float) wext / SCORENORM		   );
	PP	"\n"								   );
}

#undef PP

#define PP (void)fprintf(stderr, HEADER_TAG

/* -------------------------------------------- */
/* printout class header			*/
/* -------------------------------------------- */

static void sPrintClassHeader( ClassParam *cparms)
{
	PP	"/------------------------------------------		 \n");
	PP	"/   Class part                           		 \n");
	PP	"/------------------------------------------		 \n");
	PP	"/ Bestfit score threshold for \"homologs\": %d\n",
						(int)(cparms->seuil_score));
	PP	"/ length ratio threshold : ");
	if (cparms->len_thresh < 0.0)
	(void)fprintf(stderr, " NO LENGTH RATIO THRESHOLD\n");
	else
	(void)fprintf(stderr, " %f\n", cparms->len_thresh);
	PP	"/ Maximum number of genes separating two genes          \n");
	PP      "/   to be in the same class:%d                          \n",
							cparms->nseuil);
	PP	"/ One to one relation: %s\n",  (cparms->bijectif) ? "YES":"NO");
	PP	"/ Singletons are %s\n",  (cparms->output_singleton) ?
							"output":"not output");
	PP	"/ Operon set is %s\n",  (cparms->ope_circular) ? "circular":"linear");
	PP	"/ Chromosome is %s\n",  (cparms->circular) ? "circular":"linear");
	PP	"/ Compute also classes within operon genes:%s\n",
					(cparms->operon_class) ? "YES":"NO");
	PP	"/ %s output intra-operon classes (see README.opscan)\n",
					(cparms->outR1) ? "Do":"Do not ");
}

#undef PP

#define PP (void)fprintf(stderr, BBH_TAG

/* -------------------------------------------- */
/* printout BBH header			*/
/* -------------------------------------------- */

static void sPrintBBhHeader( ClassParam *cparms)
{
	PP	"/------------------------------------------		 \n");
	PP	"/   Double best hits part                    		 \n");
	PP	"/------------------------------------------		 \n");
	PP	"/ Bestfit score threshold for \"homologs\": %d\n",
						(int)(cparms->seuil_score));
	PP	"/ length ratio threshold : ");
	if (cparms->len_thresh < 0.0)
	(void)fprintf(stderr, " NO LENGTH RATIO THRESHOLD\n");
	else
	(void)fprintf(stderr, " %f\n", cparms->len_thresh);
}

#undef PP

/* ----------------------------------------------------	*/
/* PrintResultR2: imprime les classes resultats	*/
/* ----------------------------------------------------	*/
void PrintResultR2(OperonSeqPtr *ope, int nb_ope, int nclass, int stat_on)
{
   int i_ope;
   float len_ratio;

#define PN (void)fprintf(stdout, CL_NUMBER
#define PP (void)fprintf(stdout, CLASS_TAG
#define PD (void)fprintf(stdout, CDESC_TAG

   PN "Number of classes: %d\n", nclass);

		/* all ope must have homolog.gen_ptr = homolog or NOT_G */

   for (i_ope = 0; i_ope < nb_ope; i_ope++) {


      if (ope[i_ope]->homolog.gen_ptr != NOT_G && ope[i_ope]->class != SINGLETON) {


         PP "%4d %s %4d %4d",
		ope[i_ope]->class,
	    	ope[i_ope]->mnemo, ope[i_ope]->ignore,
	 	ope[i_ope]->class);

	 (void)fprintf(stdout," %s %4d %5.1f",
	    	ope[i_ope]->homolog.gen_ptr->gene_name,
	    	ope[i_ope]->homolog.gen_ptr->rel_number,
	    	(float)ope[i_ope]->homolog.score * (float)NORMSCORE/(float)HISCORE);

	 if (stat_on)
	   (void)fprintf(stdout," %5.3f", ope[i_ope]->homolog.stat);

	 len_ratio = LENRATIO(ope[i_ope]->fseqptr->length,
			ope[i_ope]->homolog.gen_ptr->fseqptr->length);
 
	 (void)fprintf(stdout," %5.2f\n", len_ratio);
         
	 PD "       %s: %s\n",
			ope[i_ope]->mnemo, ope[i_ope]->fseqptr->comment);
	 PD "       %s: %s\n",
			ope[i_ope]->homolog.gen_ptr->mnemo,
			ope[i_ope]->homolog.gen_ptr->fseqptr->comment);

      }
   }

   return;
}

#undef PP
#undef PD
#undef PO
#undef PN

/* ----------------------------------------------------	*/
/* sPrintSingletons: imprime les singletons       	*/
/* ----------------------------------------------------	*/
static void sPrintSingletons(OperonSeqPtr *ope, int nb_ope,
				int stat_on)
{
   int		i_ope, nsing;
   float	len_ratio;

#define PP (void)fprintf(stdout, SINGL_TAG
#define PN (void)fprintf(stdout, SINGL_NUMBER
#define PD (void)fprintf(stdout, SINGL_DESC

   nsing = 0;
   for (i_ope = 0; i_ope < nb_ope; i_ope++)
       if (ope[i_ope]->homolog.gen_ptr != NOT_G &&
	   ope[i_ope]->homolog.gen_ptr->class == SINGLETON)
	      nsing++;

   PN "Single operon/genomic gene couples: %d\n", nsing);

		/* all ope must have homolog.gen_ptr = homolog or NOT_G */

   for (i_ope = 0; i_ope < nb_ope; i_ope++) {


      if (ope[i_ope]->homolog.gen_ptr != NOT_G &&
	  ope[i_ope]->class == SINGLETON) {

         PP "%s %4d %4d %s %4d %3d %5.1f",
	    	ope[i_ope]->mnemo, ope[i_ope]->ignore,
		ope[i_ope]->class,
	    	ope[i_ope]->homolog.gen_ptr->mnemo,
	    	ope[i_ope]->homolog.gen_ptr->rel_number,
	    	ope[i_ope]->homolog.gen_ptr->class,
	    	(float)ope[i_ope]->homolog.score * (float)NORMSCORE / (float)HISCORE);

	 if (stat_on)
            (void)fprintf(stdout," %5.3f", ope[i_ope]->homolog.stat);

	 len_ratio = LENRATIO(ope[i_ope]->fseqptr->length,
			ope[i_ope]->homolog.gen_ptr->fseqptr->length);

	 (void)fprintf(stdout," %5.2f\n", len_ratio);

	 PD "       %s: %s\n",
			ope[i_ope]->mnemo, ope[i_ope]->fseqptr->comment);
	 PD "       %s: %s\n",
			ope[i_ope]->homolog.gen_ptr->mnemo,
			ope[i_ope]->homolog.gen_ptr->fseqptr->comment);
      }
   }

   return;

#undef PP
#undef PN
#undef PD
}

/* ----------------------------------------------------	*/
/* sPrintBBH: imprime les double best hits       	*/
/* ----------------------------------------------------	*/
static void sPrintBBH(OperonSeqPtr *ope, int nb_ope,
				int stat_on)
{
   int		i_ope, nbbh;
   float	len_ratio;

#define PP (void)fprintf(stdout, BBH_TAG
#define PN (void)fprintf(stdout, BBH_NUMBER
#define PD (void)fprintf(stdout, SINGL_DESC


   nbbh = 0;
   for (i_ope = 0; i_ope < nb_ope; i_ope++)
       if (ope[i_ope]->homolog.gen_ptr != NOT_G && 
	   ope[i_ope]->homolog.gen_ptr != NULL)
	      nbbh++;

   PN "Operon/genomic gene double best hits: %d\n", nbbh);

		/* all ope must have homolog.gen_ptr = homolog or NOT_G */

   for (i_ope = 0; i_ope < nb_ope; i_ope++) {


      if (ope[i_ope]->homolog.gen_ptr != NOT_G &&
	  ope[i_ope]->homolog.gen_ptr != NULL) {

         PP "%s %4d %s %4d %5.1f",
	    	ope[i_ope]->mnemo, ope[i_ope]->ignore,
	    	ope[i_ope]->homolog.gen_ptr->mnemo,
	    	ope[i_ope]->homolog.gen_ptr->rel_number,
	    	(float)ope[i_ope]->homolog.score * (float)NORMSCORE / (float)HISCORE);

	 if (stat_on)
            (void)fprintf(stdout," %5.3f", ope[i_ope]->homolog.stat);

	 len_ratio = LENRATIO(ope[i_ope]->fseqptr->length,
			ope[i_ope]->homolog.gen_ptr->fseqptr->length);

	 (void)fprintf(stdout," %5.2f\n", len_ratio);

	 PD "       %s: %s\n",
			ope[i_ope]->mnemo, ope[i_ope]->fseqptr->comment);
	 PD "       %s: %s\n",
			ope[i_ope]->homolog.gen_ptr->mnemo,
			ope[i_ope]->homolog.gen_ptr->fseqptr->comment);
      }
   }

   return;

#undef PP
#undef PN
#undef PD
}

/* ----------------------------------------------- */
/* fit s2 into s1, padding with spaces		   */
/* ----------------------------------------------- */
static char *sFitName(char *s1, char *s2)
{
    int i;

    (void) strncpy(s1, s2, Pretty_Name_Len);
    s1[Pretty_Name_Len] = '\000';

    for (i = 0 ; i < Pretty_Name_Len ; i++)    
	if (! s1[i])
	    s1[i] = ' ';

    return s1;
}

/* ----------------------------------------------- */
/* printout alignment				   */
/* Note: sequence must have been deindexed before  */
/* ----------------------------------------------- */
static void sPrintAli(FastaSequence *query, FastaSequence *dbseq, char *pretty)
{
    int	    i, j, jmax, up, low, pages, reste;
    char    *c_up, *c_low;
    char    namq[Pretty_Name_Len+2], 
	    namb[Pretty_Name_Len+2];

    (void) sFitName(namq, query->name);        
    (void) sFitName(namb, dbseq->name);        

    i     = (int)(strlen(pretty)) / 2;
    pages = i / Pretty_Out_Len;
    reste = i % Pretty_Out_Len;
    
    if (reste)
	pages++;
    
    c_up  =  pretty;
    c_low =  c_up + 1;
    
    up = low = 1;
   	   	
    for (i = 1 ; i <= pages ; i++) {
	jmax = (reste && (i == pages) ? reste : Pretty_Out_Len);
	printf("\n>%s %5d ", namq, up);
   	for (j = 0 ; j < jmax ; j++, c_up += 2) {
	    putchar(*c_up);
	    if (IS_ALPHA(*c_up))
		up++;
	}
	putchar('\n');
	printf(">%s %5d ", namb, low);
   	for (j = 0 ; j < jmax ; j++, c_low += 2) {
	    putchar(*c_low);
	    if (IS_ALPHA(*c_low))
		low++;
	}
	putchar('\n');
    }	


}



/* -------------------------------------------- */
/* printout footer				*/
/* -------------------------------------------- */

#define PP (void)fprintf(stderr, SUMMARY_TAG

static void sPrintFooter(int nb_ope, int nb_gen, int nb_ali_fast, int nb_ali_best,
			int nb_ali2, int nb_fastp_gen_elim)
{
	int		ngentot;
	MyPsPtr		mypsptr;

	ngentot = nb_gen+nb_fastp_gen_elim;
	
(void)fprintf(stderr,	"\n");
PP      "/------------------------------------------\n");
PP	" Total # of operon sequences read: %d\n", nb_ope);
PP	" Total # of genome sequences read: %d\n", ngentot);
PP	" # of correct length sequences eliminated after fastp: %d\n",
		nb_fastp_gen_elim);
PP	" fraction of correct sequences kept: %.1f\n",
		((float) (ngentot-nb_fastp_gen_elim) * 100. / (float) nb_gen));
PP	" Comparisons space: %d\n", nb_ope*ngentot);
PP	" Total # of fastp comparisons performed: %d\n", nb_ali_fast);
PP	" Total # of bestfit alignments performed in first pass: %d\n", nb_ali_best);
PP	" Total # of bestfit alignments performed in second pass: %d\n", nb_ali2);
/*	PP	" Pre-filtering savings: %.1f %%\n", save); */
PP	" User Cpu time: %.2f s\n", UserCpuTime());
PP	" Sys  Cpu time: %.2f s\n", SystemCpuTime());
	mypsptr = GiveProcessStatus(stderr);
PP	" Memory used: %d Kb\n", (int)mypsptr->mem_ko);
PP	" End of program\n");
}

#undef PP

/* ---------------------------------------------------- */
static float ComputeTabStat(char *tscore, int i_ope, int ngen, int score)
{
   int i, n;

   tscore += (i_ope*ngen);
   
   n = 0;

   for (i = 0; i < ngen; i++)
	if ((int)score < (int)tscore[i])
	    n++;
   
   return (float)n/(float)ngen;

}

/* ---------------------------------------------------- */
static void sPrintAlignments(OperonSeqPtr *ope, int nb_ope,
		ScorMat *pmatrix, BestParam *pbparms,
		char *alphabet)
{
   int		i_ope;
   char		*pretty;

   for (i_ope = 0; i_ope < nb_ope; i_ope++) {
      if (ope[i_ope]->homolog.gen_ptr != NOT_G &&
	  ope[i_ope]->homolog.gen_ptr->class != SINGLETON) {

	 NwsAlign(ope[i_ope]->fseqptr->seq,
		  ope[i_ope]->homolog.gen_ptr->fseqptr->seq,
                  ope[i_ope]->fseqptr->length,
		  ope[i_ope]->homolog.gen_ptr->fseqptr->length,
		  pmatrix->weight,  pbparms->wopen, pbparms->wext, &pretty,
                                pbparms->fglobal); 
      
    	 (void) DeIndexSequence(ope[i_ope]->fseqptr->seq, alphabet);
    	 (void) DeIndexSequence(ope[i_ope]->homolog.gen_ptr->fseqptr->seq, alphabet);
	 sPrintAli(ope[i_ope]->fseqptr, ope[i_ope]->homolog.gen_ptr->fseqptr, pretty);
    	 (void) IndexSequence(ope[i_ope]->fseqptr->seq, alphabet);
    	 (void) IndexSequence(ope[i_ope]->homolog.gen_ptr->fseqptr->seq, alphabet);
      }
   }

}

/* ---------------------------------------------------- */
/* BestfitBestGenAllOpe: run besfit between all		*/
/* best genes of operon genes and all operon genes	*/
/* Note: bijectif is always TRUE when calling this	*/
/* function						*/
/* seuil_score:						*/
/*  do not keep result if bestfit score < seuil_score	*/
/* Note:                                       		*/
/* seuil_score must be incremented before by   		*/
/* BASE_BEST as we compare bestfit scores      		*/
/*							*/
/* bijectif:						*/
/*  if bijectif, SetOneHomolog will keep as homolog	*/
/*  for an operon gene "ope" only the genomic genes	*/
/*  whose best hit is the operon gene "ope"		*/
/* ---------------------------------------------------- */
static int sBestfitBestGenAllOpe(OperonSeqPtr *ope, int nb_ope,
			ScorMat *pmatrix, BestParam *pbparms,
			Parameters *pparms, float len_thresh,
			int seuil_score_best, int bijectif,
			int *nb_ali)
{

  GenomeSeqPtr	gen_ptr;
  int		i, j, k, found;
  long		dist;
  float		fscore, fdist, fnorm;

  *nb_ali = 0;

#define MONITOR_INV_BESTFIT	0

#define PP   (void)fprintf(stderr, COMPLETION_TAG

  if (pparms->opt_verbose)
     PP "Bestfit of the best genomic gene of operon genes against operon genes\n");

					/* ----------------------------	*/
					/* in case of fastp done	*/
					/* ----------------------------	*/
  if (pparms->do_fastp) {
					/* for all operon gene		*/
   for (i = 0; i < nb_ope; i++) {
					/* if ope is not OK		*/
					/* OR if best score of ope	*/
					/* is < seuil_score_best	*/
					/* jump to next			*/
     if ((! (ope[i]->fseqptr->ok)) || (int)ope[i]->gen_kept[0].score < seuil_score_best) {
	ope[i]->homolog.gen_ptr = NOT_G;
	ope[i]->homolog.score   = (SCORE)0;
	ope[i]->homolog.stat   = -1.0;
#if MONITOR_INV_BESTFIT
	fprintf(stderr, "ope %s best score (%d) is inf to threshold %d\n",
		ope[i]->mnemo, (int)ope[i]->gen_kept[0].score, seuil_score_best);
#endif
	continue;
     } 
					/* best gene of ope[i]	*/
     gen_ptr = ope[i]->gen_kept[0].gen_ptr;

					/* if no gen_ptr, next		*/
     if (!gen_ptr) {
	ope[i]->homolog.gen_ptr = NOT_G;
	ope[i]->homolog.score   = (SCORE)0;
	ope[i]->homolog.stat   = -1.0;
#if MONITOR_INV_BESTFIT
	fprintf(stderr, "ope %s has no gen_kept[0].gen_ptr\n");
#endif
	continue;
     }


					/* fastp has been done, so ope_kept	*/
					/* must be filled, so if no score<BASE_BEST	*/
					/* and >-1, all scores are besfit score	*/
					/* if gen_ptr			*/
					/* has already been "bestfited"	*/
					/* against all its "fastp" ope	*/
					/* -1- verify that its best hit	*/
					/* is ope[i]			*/
					/* -2- jump to the next ope	*/
/* PATCH A ENLEVER */
/*     sort_kept_ope_by_score_in_gene(gen_ptr); */

     found = FindInfScoreInOpeKept(gen_ptr, BASE_BEST);

     if ( found == -1) {		/* no found score < BASE_BEST	*/
					/* i.e. all "fastp" ope of	*/
					/* gen_ptr have been bestfitted	*/

					/* then test if ope[i] is the	*/
					/* best ope of gen_ptr		*/
	if (ope[i] == gen_ptr->ope_kept[0].ope_ptr && 
	    (len_thresh < 0.0 || LENRATIO(ope[i]->fseqptr->length,
		gen_ptr->fseqptr->length) <= len_thresh)
	    ) {
					/* we are sure that the score	*/
					/* is >= seuil_score_best	*/
					/* because it is the same as	*/
					/* ope[i]->gen_kept[0].score	*/
	    ope[i]->homolog.gen_ptr = gen_ptr;
	    ope[i]->homolog.score   = gen_ptr->ope_kept[0].score;
	    ope[i]->homolog.stat   = gen_ptr->ope_kept[0].stat;
	    ope[i]->homolog.gen_ptr->homol_ptr = ope[i];
#if MONITOR_INV_BESTFIT
	fprintf(stderr, "ope %s is the gen_ptr->ope_kept[0].ope_ptr of %s (all ope_kept[i] bestfitted)\n",
		ope[i]->mnemo, gen_ptr->mnemo);
#endif
	}
	else {
					/* if ope[i] is not the best	*/
					/* ope of gen_ptr (its best	*/
					/* gene), then no homolog for	*/
					/* ope[i]			*/
	    ope[i]->homolog.gen_ptr = NOT_G;
	    ope[i]->homolog.score   = (SCORE)0;
	    ope[i]->homolog.stat   = -1.0;
#if MONITOR_INV_BESTFIT
	fprintf(stderr, "ope %s is not the gen_ptr->ope_kept[0].ope_ptr of %s (all ope_kept[i] bestfitted)\n",
		ope[i]->mnemo, gen_ptr->mnemo);
#endif
	}

					/* next ope			*/
	continue;
     }
					/* else, we must bestfit all 		*/
					/* ope[i]->gen_kept[k].gen_ptr with	*/
					/* score<BASE_BEST with all their ope_kept	*/
					/* note: k is the first gen_kept with	*/
					/* score<BASE_BEST (and for all		*/
					/* following k until			*/
					/* ope[i]->actual_gen_kept 		*/
					/* ope[i]->gen_kept[k] have		*/
					/* also score<BASE_BEST			*/
/* PATCH A ENLEVER, TROUVER LE BUG , les scores des ope_kept du gen_kept[0] de MYPN_001150 (num 113)
ne sont pas ranges en ordre decroissant !!! */
/*      for (k = 0; k < ope[i]->actual_gen_kept; k++) 
	 sort_kept_ope_by_score_in_gene(ope[i]->gen_kept[k].gen_ptr);
 */ 
     for (k = found; k < ope[i]->actual_gen_kept; k++) {

					/* Note: the bestfit score is also	*/
					/* kept in the ope_kept array of the	*/
					/* genome gene ope[i]->gen_kept[0]	*/
					/* if ope was a fastp hit for this	*/
					/* genome gene				*/
        gen_ptr = ope[i]->gen_kept[k].gen_ptr;

        for (j = 0; j < gen_ptr->actual_ope_kept; j++) {

					/* if bestfit between gen_ptr and	*/
					/* its ope_kept[j] has been done, next.	*/
					/* OR 					*/
					/* if ope_kept[j] has already an	*/
					/* homolog, or have no homolog (NOT_G)	*/
					/* AND bijectif is ON			*/
					/* continue				*/
					/* next...				*/

	   if ((int)gen_ptr->ope_kept[j].score > BASE_BEST ||
		(bijectif && gen_ptr->ope_kept[j].ope_ptr->homolog.gen_ptr != NULL))
	      continue;

					/* is the gene already bestfitted	*/
					/* against ope_kept[j] ope_ptr ?	*/

	   fscore = IsGenInOpeKeptR(gen_ptr->ope_kept[j].ope_ptr, gen_ptr, BASE_BEST);

	   if (fscore == -1) {
					/* gen_ptr not already "bestfitted"	*/
					/* against ope[j]			*/
					/* bestfit to do */
	      (*nb_ali)++; 
#ifdef ROTATOR /* too much output	*/
              if(pparms->opt_verbose)
	         (void)fprintf(stderr,"\r");
		 PP " genomic gene of operon gene %4d / %4d\t\t", i+1, nb_ope);
#endif	   
	      dist = NwsAlign(gen_ptr->ope_kept[j].ope_ptr->fseqptr->seq, gen_ptr->fseqptr->seq,
				gen_ptr->ope_kept[j].ope_ptr->fseqptr->length, gen_ptr->fseqptr->length,
				pmatrix->weight,  pbparms->wopen,  pbparms->wext,
				NULL, pbparms->fglobal);

	      fdist = (float) dist/ SCORENORM;

	      fnorm = (pbparms->fglobal ? (float) MAX2(gen_ptr->ope_kept[j].ope_ptr->fseqptr->length,
						gen_ptr->fseqptr->length)
				 : (float) MIN2(gen_ptr->ope_kept[j].ope_ptr->fseqptr->length,
						gen_ptr->fseqptr->length));

	      fscore = (float)HISCORE - (fdist * (float)HISCORE / fnorm) + BASE_BEST;

	      KeepOpeInGeneB(gen_ptr, gen_ptr->ope_kept[j].ope_ptr, fscore, -1.0);

	   }
   
#if 0
	   PP "Operon: %s Genome %s Dist: %.1f Score: %3d\n",
				gen_ptr->ope_kept[j].ope_ptr->fseqptr->name, gen_ptr->fseqptr->name,
				 fdist, fscore); 
#endif

        }				/* end for j		*/

     }				/* end for k		*/
    
					/* all kept genomic genes of ope[i]	*/
					/* have been "bestfitted" against their	*/
					/* fastp hits so we can set if there	*/
					/* is an homolog or not			*/

     SetOneHomolog(ope[i], len_thresh, seuil_score_best, bijectif);
     if (ope[i]->homolog.gen_ptr == NULL)
	    PP "sBestfitBestGenAllOpe: This should not happens - homolog == NULL\n");

   }				/* end for i		*/

 }
 else {
					/* ---------------------------	*/
					/* no fastp done, so bestfit	*/
					/* must be run between all	*/
					/* genes and all operon genes	*/
					/* ---------------------------	*/

					/* for all operon gene		*/
   for (i = 0; i < nb_ope; i++) {
					/* if ope is not OK		*/
					/* OR if best score of ope	*/
					/* is < seuil_score_best	*/
					/* jump to next			*/
					/* Note: seuil_score_best is	*/
					/* already corrected with	*/
					/* BASE_BEST			*/
     if ((! (ope[i]->fseqptr->ok)) || (int)ope[i]->gen_kept[0].score < seuil_score_best) {
	ope[i]->homolog.gen_ptr = NOT_G;
	ope[i]->homolog.score   = (SCORE)0;
	ope[i]->homolog.stat   = -1.0;
	continue;
     }
					/* best gene of ope[i]	*/
     gen_ptr = ope[i]->gen_kept[0].gen_ptr;

					/* if no gen_ptr, next		*/
     if (!gen_ptr) {
	ope[i]->homolog.gen_ptr = NOT_G;
	ope[i]->homolog.score   = (SCORE)0;
	ope[i]->homolog.stat   = -1.0;
	continue;
     }
					/* fastp has not been done, so ope_kept	*/
					/* can be partially filled with		*/
					/* bestfit score that we do not need to	*/
					/* recompute				*/


					/* for all gene kept			*/
     for (k = 1; k < ope[i]->actual_gen_kept; k++) {

					/* Note: the bestfit score is also	*/
					/* kept in the ope_kept array of the	*/
					/* genome gene ope[i]->gen_kept[0]	*/
					/* if ope was a fastp hit for this	*/
					/* genome gene				*/
        gen_ptr = ope[i]->gen_kept[k].gen_ptr;

					/* if gene has already been bestfitted	*/
					/* against all operon genes		*/
	if (gen_ptr->bestfitted) 
	   continue;

					/* for all ope      			*/
        for (j = 0; j < nb_ope; j++) {

				

					/* if i==j, nothing to do, next...	*/
	   if (j == i)
	      continue;

					/* if bijectif && ope[j] has already 	*/
					/* an homolog, or have no homolog	*/
					/* (NOT_G) continue			*/

	   if (bijectif && ope[j]->homolog.gen_ptr != NULL)
	     continue;


					/* if bestfit between gen_ptr and	*/
					/* ope[j] has been done, next..		*/

	   fscore = IsGenInOpeKeptR(ope[j], gen_ptr, BASE_BEST);

	   if (fscore == -1) {
					/* gen_ptr not already "bestfitted"	*/
					/* against ope[j]			*/

					/* bestfit to do */
	      (*nb_ali)++; 

#ifdef ROTATOR /* too much output	*/
              if(pparms->opt_verbose)
	         (void)fprintf(stderr, "\r");
		 PP "genomic gene of operon gene %4d / %4d\t\t", i+1, nb_ope);
#endif	   
	      dist = NwsAlign(ope[j]->fseqptr->seq, gen_ptr->fseqptr->seq,
				ope[j]->fseqptr->length, gen_ptr->fseqptr->length,
				pmatrix->weight,  pbparms->wopen,  pbparms->wext,
				NULL, pbparms->fglobal);

	      fdist = (float) dist/ SCORENORM;

	      fnorm = (pbparms->fglobal ? (float) MAX2(ope[j]->fseqptr->length,
						gen_ptr->fseqptr->length)
				 : (float) MIN2(ope[j]->fseqptr->length,
						gen_ptr->fseqptr->length));

	      fscore = (float)HISCORE - (fdist * (float)HISCORE / fnorm) + BASE_BEST;

					/* A FINIR JOMPO	*/
	      KeepOpeInGeneB(gen_ptr, ope[j], fscore, -1.0);

	   }
   


#if 0
	   PP "Operon: %s Genome %s Dist: %.1f Score: %3d\n",
				ope[i_ope]->fseqptr->name, gen_ptr->fseqptr->name,
				 fdist, fscore); 
#endif


        }				/* end for j		*/
	
	gen_ptr->bestfitted = 1;

     }					/* end for k		*/
    
					/* all genomic genes of ope[i]		*/
					/* have been "bestfitted" against 	*/
					/* all possible operon genes, so	*/
					/* so we can set if there		*/
					/* is an homolog or not			*/

     SetOneHomolog(ope[i], len_thresh, seuil_score_best, bijectif);

   }					/* end for i		*/
 }

#if RECURS_HOMOLOG
 SetAllOpeLastHomologs(ope, nb_ope, len_thresh, seuil_score_best,
			bijectif, pparms->opt_verbose);
#endif

 return 0;
}

#undef PP

#define ON_VERBOSE  if (parms.opt_verbose)
#define ON_VERY_VERBOSE  if (parms.opt_very_verbose)

#define ON_ONLY_SIMILARITY  if (parms.opt_only_sim)

/* ----------------------------------------------- */
/* main						   */
/* ----------------------------------------------- */

int main(argn, argv)
	int  argn;
	char *argv[];
{
	int	      carg, rarg, errflag, max_operon_seq, max_gen_seq, nb_ope_tot,
	    	      i_ope, i_gen, maxlen_ope, minlen_ope, maxlen_gen, nb_ope, nb_gen,
		      nb_ali_fast, nb_fastp_gen_elim, nerrseq,
		      nb_ali2, nb_ali_best, ngen_todo, fastbest, n_ali,
		      norme = 0, score, i_new_gen, ndel,
		      min_gen_rel_number, max_gen_rel_number,
		      min_ope_rel_number, max_ope_rel_number,
		      nclass_gen, nclass_ope = 1;
	int	      *bstat = NULL;
	int	     (*compar) (const void *, const void *);
	float	      fscore = 0.0, farg, fdist, fnorm;
	FILE          *operon_filin;
	OperonSeqPtr  *ope, ope_temp;
	GenomeSeqPtr  *gen, gen_temp, gen_ptr;

					/* specific fastp	*/
	Parameters    parms;
	ClassParam    cparms;
	DiagTable     *diags = NULL;
	HashTable     *hashtab = NULL;
	FILE	      *fast_alistream;
	char	      operon_query_file[512], *shufseq;

					/* specific bestfit	*/
	BestParam     bparms;
	ScorMat       matrix;
	long          dist;
	char          bestmatfile[MAXPATHLEN];
	char	      *tab_alea = NULL;
	
	/* ---------------------------- */    
	/* defaut options  		*/

	errflag			= 0;
	parms.sym_replace	= (char)'X';
	parms.opt_sym_replace	= 1;
	parms.gen_std_fasta_format	= 0;	/* should be "mnemo format", see README.opscan	*/
	parms.ope_std_fasta_format	= 0;	/* should be "mnemo format", see README.opscan	*/
	parms.opt_sym_delete	= 0;
	parms.opt_verbose	= 0;
	parms.opt_very_verbose	= 0;
	parms.opt_only_sim	= 0;
	parms.max_gen_kept      = MAX_GEN_KEPT;
	parms.short_output      = 0;	/* no short output by default	*/

	/* ---------------------------- */    
	/* fastp default options	*/
		
	parms.diag_integ	= 2;
	parms.kuple		= -1;
	parms.fuple		= -1;
	parms.fast_opt_align	= 0;
	parms.do_fastp    	= 1;
	parms.fast_lo_threshold	= DEFAULT_FAST_LO_THRESHOLD;
	parms.bstat             = 0;	/* no stats by default		*/
	/* OLD ???? parms.store_scores      = 0; */	/* scores are not stored	*/
					/* by default			*/
	parms.store_aleas      = 0;	/* shuffled scores are not stored	*/
					/* by default			*/

	/* ---------------------------- */    
	/* class default options	*/
	
	cparms.do_bbh = 0;		/* no double best hits by default	*/
	cparms.outR1 = 0;
	cparms.nseuil = DEFAULT_CLASS_DIST;
	cparms.circular = 0;
	cparms.ope_circular = 0;
	cparms.bijectif = 1;
	cparms.combi = 0;		/* combinatory search in place of bijective	*/
	cparms.operon_class = 0;
	cparms.seuil_score = DEFAULT_CLASS_SCORE_SEUIL;
	cparms.output_singleton = 1;	/* sort les singletons	*/
	cparms.len_thresh	= 1.3;


	/* ---------------------------- */    
	/* bestfit default options	*/
		
	bparms.wopen       = 1.2 * SCORENORM;
	bparms.wext        = 0.8 * SCORENORM;
	bparms.fglobal     = 0;		/* local	*/
	bparms.best_print  = 0;	
	bparms.printmatrix = 0;


	/* ----------------------------	*/
	/* used in fastp		*/

	(void)strcpy(parms.alphabet, STDALPHABET);
	parms.dbfile[0]  = '\000';
	parms.fast_matfile[0] = '\000';
	parms.fast_alifile[0] = '\000';
	parms.fastp_aliscore = 1;
	parms.fastp_mxscore = 0;
	parms.exclude_self = 0;


	/* ----------------------------		*/
	/* used in bestfit			*/
	/* and in fastp if fastp_aliscore	*/
	/* or fastp_mxscore			*/

	(void) strcpy(bestmatfile, "<internal> Identity");
			/* Identity matrix:		*/
			/* matrix->score[i][j]:		*/
			/*   used in fastp_mxscore	*/
			/*    1 if i == j		*/
			/*   -1 if i != j		*/
			/* matrix->simscore[code(i,j)]:	*/
			/*   used in fastp_aliscore &	*/
			/*   0.0 if i == j		*/
			/*   1.0 if i != j		*/
			/* matrix->weight[code(i,j)]:	*/
			/* used in bestfit		*/
			/*   == distance		*/
	(void) IdentityMatrix(&matrix);

	operon_filin = NULL;

        /* ---------------------------- */    
        /* parse cmdline arguments      */
        
        while ((carg = getopt(argn, argv, "BCDEFHIJK:L:M:NO:PQR:STUVXZabce:ghi:k:lmn:o:p:r:st:vxy")) != -1) {

            
            switch(carg) {

		case 'N' :		/* no fastp			*/
		    parms.do_fastp = 0;
		    break;

		case 'I' :		/* print matrix 			*/
		    bparms.printmatrix = 1;
		    break;

		case 'J' :		/* short output format  		*/
		    parms.short_output = 1;
		    break;

		case 'X' :		/* exclude genome gene with same	*/
					/* name as the operon gene concerned	*/
					/* useful when comparing a genome with	*/
					/* itself				*/
		    parms.exclude_self = 1;
		    break;

		case 'x' :		/* fastp score is computed	*/
					/* as # of kuples rather	*/
					/* than alignment score 	*/
		    parms.fastp_aliscore = 0;
		    parms.fastp_mxscore = 0;
		    break;

		case 'y' :		/* fastp score is computed	*/
					/* as score of maximum subsum	*/
					/* of shifted sequences 	*/
					/* rather than alignment score 	*/
					/* or # of kuples 		*/
		    parms.fastp_aliscore = 0;
		    parms.fastp_mxscore = 1;
		    break;

		case 'Z' :		/* fastp or bestfit stats	*/
		    parms.bstat = 1;
		    break;

		case 'T' :		/* in bestfit case, for  	*/
					/* storing aleas (huge memory	*/
					/* needed)			*/
		    parms.store_aleas = 1;
		    break;

		/* OLD ???? case 'X' : */		/* in bestfit case, for  	*/
					/* storing scores (huge memory	*/
					/* needed)			*/
		/* OLD ????    parms.store_scores = 1;
		    break;
		 */

		case 'F' :		/* genomic sequence are in	*/
					/* standard fasta format	*/
					/* not in "mnemo format"	*/
					/* (see README.opscan)		*/
		    parms.gen_std_fasta_format = 1;
		    break;

		case 'E' :		/* operon sequences are in	*/
					/* standard fasta format	*/
					/* not in "mnemo format"	*/
					/* (see README.opscan)		*/
		    parms.ope_std_fasta_format = 1;
		    break;

		case 'D' :		/* unknown symbol: DO NOT delete  */
		    parms.opt_sym_delete = 0;
		    break;

		case 'K' :		/* # of kept similar genes	*/
					/* by one operon gene		*/
		   parms.max_gen_kept = sGetIArg("K", optarg);
		   if (parms.max_gen_kept < 1) {
		      (void)fprintf(stderr,
			  REMARK_TAG "option -K: number is < 1, setting it to %d\n",
			  MAX_GEN_KEPT);
		      parms.max_gen_kept = MAX_GEN_KEPT;
		   }
		   break;

		case 'O' :		/* arg : name of operon query	*/
					/* file (fasta like format)	*/
		    if (sscanf(optarg, "%s", operon_query_file) < 0) {
		       (void)fprintf(stderr, ERROR_TAG
			 "option -O: Cannot read queries file name in %s\n",
			 optarg);
		       exit(1);
		    }
		    if (!(operon_filin=fopen(operon_query_file,"r"))) {
		       (void)fprintf(stderr, ERROR_TAG
			 "option -O: Cannot open file %s\n",operon_query_file);
		       exit(1);
		    }
		    break;

                case 'h' :              /* help             */
					/* specific to opscan */
		    PrintHelpOpscan();
		    			/* options of fastp	*/
		    PrintFastpHelp();
		    			/* options of bestfit	*/
		    PrintBestfitHelp();
		    exit(0);
		    break;

		case 'i' :		/* integration    */
		    parms.diag_integ = sGetIArg("i", optarg);
		    break;

		case 'k' :		/* kuple	  */
		    parms.kuple = sGetIArg("k", optarg);
		    break;

		case 'L' :		/* fast_lo_threshold	    */
					/* get the score in the		*/
					/* 0-100 range			*/
		    parms.fast_lo_threshold = sGetIArg("L", optarg);
                    if (parms.fast_lo_threshold >= 100 || parms.fast_lo_threshold < 0) {
		       (void) Erreur("Lower threshold for Fastp is not in the 0-100 range", 1);
		    }
					/* corrects score in the	*/
					/* 0-HISCORE range		*/
		    parms.fast_lo_threshold = (int)((parms.fast_lo_threshold * HISCORE)/(float)NORMSCORE);
		    break;


		case 'p' :		/* print fastp align	    */
		    (void) Erreur("Not yet implemented: -p", ErrNotImp);
		    if (! CheckAccess(optarg, "w"))
			exit(ErrFilAcc);
		    (void)strcpy(parms.fast_alifile, optarg);
		    parms.fast_opt_align = 1;
		    break;

		case 'R' :		/* symbol replace    */
		    parms.sym_replace = sGetCArg("R", optarg);
		    parms.opt_sym_replace = 1;
		    break;

		case 'r' :		/* ratio length threshold	    */
		    cparms.len_thresh = sGetFArg("r", optarg);
		    if (cparms.len_thresh < 1.0)
		       cparms.len_thresh = -1.0;
		    break;

		case 'v' :		/* verbose	    */
		    parms.opt_verbose = 1;
		    break;

		case 'V' :		/* very verbose	    */
		    parms.opt_very_verbose = 1;
		    break;


					/* --------------------	*/
					/* bestfit arguments	*/
					/* --------------------	*/

		case 'e' :		/* gap extension    */
		    farg = sGetFArg("e", optarg);
		    bparms.wext = (int) (farg * SCORENORM);
		    break;

		case 'g' :		/* global           */
		   bparms.fglobal = 1;
		   break;

		case 'l' :		/* local            */
		   bparms.fglobal = 0;
		   break;

		case 'M' :		/* matrix           */
		   (void)strcpy(bestmatfile, optarg);
		   if (! (ReadScoreMatrix(bestmatfile, &matrix)))
		      exit(1);
		   break;

		case 'o' :		/* gap opening      */
		   farg = sGetFArg("o", optarg);
		   bparms.wopen = (int) (farg * SCORENORM);
		   break;

		case 'P' :		/* print bestfit align      */
		   bparms.best_print = 1;
		   break;

					/* --------------------	*/
					/* class arguments	*/
					/* --------------------	*/

		case 'b' :		/* no "bijective":		*/
					/* two or more ope_gen can be	*/
					/* "homologs" with the same	*/
					/* genomic gene			*/
					/* default: 0 [Off] i.e.	*/
					/* for a "homologic" relation	*/
					/* to be set between a ope_gen	*/
					/* and a geno_gen: geno_gen	*/
					/* must be the most similar to	*/
					/* ope_gen AND ope_gen must be	*/
					/* the most similar to geno_gen	*/
		   cparms.bijectif = 0;
		   break;

		case 'a' :		/* operons are searched in 	*/
					/* a combinatory manner		*/
					/* cannot be combined with	*/
					/* bijectif			*/
		    cparms.combi = 1;
		    break;

		case 'c' :		/* genomic chromosome is	*/
					/* circular			*/
					/* default: 0 [Off] i.e.	*/
					/* linear chromosome		*/
		   cparms.circular = 1;
		   break;

		case 'C' :		/* sortir aussi les classes	*/
					/* R1 en plus des classes R2	*/
					/* i.e. classes intra-operonic	*/
		   cparms.outR1 = 1;
		   break;

		case 'n' :		/* nombre de genes adjacents	*/
					/* (seuil) pour faire partie	*/
					/* d'une classe			*/
		   cparms.nseuil = sGetIArg("n", optarg);
		   break;

		case 't' :		/* bestfit score threshold	*/
					/* (0-100) for a couple operon	*/
					/* gene/genomic gene to be	*/
					/* considered as homolog 	*/
		   cparms.seuil_score = sGetIArg("t", optarg);
                   if (cparms.seuil_score >= 100 || cparms.seuil_score < 0) {
		       (void) Erreur("Lower threshold for bestfit is not in the 0-100 range", 1);
		   }
		   break;


		case 'Q' :		/* class are to be computed also on operon genes    */
		    cparms.operon_class = 1;
		    break;

		case 'U' :		/* operon is circular    */
		    cparms.ope_circular = 1;
		    break;

		case 's' :		/* Ne pas sortir les		*/
					/* singletons			*/
		   cparms.output_singleton = 0;
		   break;

		case 'S' :		/* basic search, no classes are 	*/
					/* computed, only fastp+bestfit runs	*/
					/* and best similar genomic genes are	*/
					/* output				*/
		    parms.opt_only_sim = 1;
		    break;

		case 'H' :		/* double best hit BBH			*/
					/* no classes are computed, only	*/
					/* fastp+bestfit runs    		*/
					/* and double best hits are reported	*/
		    cparms.do_bbh = 1;
		    break;



		case '?' :		/* misusage	    */
		    errflag++;
		    break;

	    }
	}

	if (bparms.printmatrix)
	   PrintMatrix(&matrix, stderr);

	/* ---------------------------- */    
	/* adapt fastp options		*/
	
	if (parms.kuple <= 0) { /* only STDALPHABET [A-Z] */
	       parms.kuple = 2;
	}
	

	/* ---------------------------- */    
	/* bestfit options		*/
	
	/* bestfit score threshold	*/
	/* corrects score in the	*/
	/* 0-HISCORE range		*/

	cparms.seuil_score = (int)((cparms.seuil_score * HISCORE)/(float)NORMSCORE);

	/* ---------------------------- */    
	/* check options		*/
	
	if (parms.opt_sym_replace && parms.opt_sym_delete) {
	    Erreur("-D and -R options are incompatibles", 0);
	    errflag++;
	}
	
	if (   parms.opt_sym_replace
	    && (! strchr(parms.alphabet, parms.sym_replace) ||
		  (char)(parms.sym_replace) == '\0')) {
	    Erreur("invalid -R value for this alphabet", 0);
	    errflag++;
	}

	if (operon_filin == NULL) {
	    Erreur("No operon file given (-O option)",0);
	    ExitUsage(1);
	}

	if (cparms.bijectif && cparms.combi) {
	    Erreur(
	       "Option -a cannot be combined with bijective homolog relation\n",
		0);
	    Erreur(
	       "Setting combinatory option only (no bijective relation)\n", 0);
            cparms.bijectif = 0;

	}

	if (cparms.bijectif == 0 && cparms.do_bbh == 1) {
	    Erreur(
	       "Option -H (double best hits) must be combined with bijective homolog relation\n",
		0);
	    Erreur(
	       "Setting relations to bijective relations\n", 0);
            cparms.bijectif = 1;

	}

	if (cparms.outR1 && ! cparms.operon_class) {
	   Erreur("Output of intra-operonic classes (option -C) needs -Q option (operon classes)",0);
	   Erreur("Setting the -Q option...", 0);
	   cparms.operon_class = 1;
	}



	/* ---------------------------- */    
	/* should remain 0/1 arguments	*/

	rarg = argn - optind;

	if (rarg == 0) {
	    (void)strcpy(parms.dbfile, "<stdin>");
	}
	else if (rarg == 1) {
	    (void)strcpy(parms.dbfile, argv[optind]);
	    if (! AssignToStdin(parms.dbfile))
		exit(ErrFilOpe);
	}
	else
	    errflag++;

	if (errflag) 
	    ExitUsage(1);

	/* ---------------------------- */    
	/* read first seq if needed	*/
	

	nb_ope = 0;
	nb_ope_tot = 0;
	minlen_ope = 100000000;
	maxlen_ope = -1;
	max_operon_seq = MAX_OPE_SEQ;
	max_ope_rel_number = -100000000;
	min_ope_rel_number =  100000000;

	if ((ope = NEWN(OperonSeqPtr,max_operon_seq)) ==  NULL) {
	    (void) Erreur(
		"Not enough memory for allocating operon sequences !!!",
		ErrManySq);
	}

	ope_temp  = NewOperonSequence(parms.max_gen_kept);

#define PPCOMPL (void)fprintf(stderr, COMPLETION_TAG
#define PPER (void)fprintf(stderr, ERROR_TAG

	while (ReadOperonSequence(operon_filin, ope_temp,
				       parms.ope_std_fasta_format)) {
	   
	       nb_ope_tot++;

	       if (ope_temp->ignore == IGNORE) {
		 ON_VERBOSE {
		    PPCOMPL "operon gene %s (num %d): ignored (ignore=%d)\n",
			  ope_temp->mnemo, nb_ope_tot, ope_temp->ignore);
		 }
		 continue;
	       }

	   
#ifdef ROTATOR /* too much output	*/
	       if (   (parms.opt_verbose)
		   && ((nb_ope % Seq_Rotator) == (Seq_Rotator - 1)))
		    (void)fprintf(stderr, "\r" COMPLETION_TAG "reading operon sequence %4d\t\t", nb_ope+1);
#endif		    
	       if (nb_ope >= max_operon_seq-1) {
		   ope = REALLOC(OperonSeqPtr, ope, max_operon_seq+INC_OPE_SEQ);
		   if (! ope) {
		      (void) Erreur(
			 "Too many operons in operon file, not enough memory",
			 ErrManySq);
                      break;
                   }
		   max_operon_seq += INC_OPE_SEQ;
	       }

    	       if (! ope_temp->fseqptr->ok) {
		   PPER "IO error with sequence #%d\n", nb_ope+1);
		   exit(2);
	       }

    	       if (ope_temp->fseqptr->length <= 0) {
		      PPER "Length error with sequence #%d (name %s) length = %d\n",
		      nb_ope+1,
		      ope_temp->fseqptr->name,
		      (int)(ope_temp->fseqptr->length));
		   exit(2);
	       }
	       else {
		  if (ope_temp->fseqptr->length > maxlen_ope)
		     maxlen_ope = ope_temp->fseqptr->length;
                  if (ope_temp->fseqptr->length < minlen_ope)
		     minlen_ope  = ope_temp->fseqptr->length;
	       }

/* debug */
#if 0
		WriteOperonSequence(stderr, ope_temp, 50);
#endif
	       (void) UpperStr(ope_temp->fseqptr->seq);
	    
	       if (parms.opt_sym_replace)
		   (void) ReplaceInStr(ope_temp->fseqptr->seq, parms.alphabet,
				    parms.sym_replace);
	       if (parms.opt_sym_delete) {
		   ndel =  DeleteInStr(ope_temp->fseqptr->seq, parms.alphabet);
		   if (ndel > 0)
		      ope_temp->fseqptr->length -= ndel;
	       }
	    
	   				/* must index sequence for	*/
					/* fastp AND besfit		*/ 
	       if (! IndexSequence(ope_temp->fseqptr->seq, parms.alphabet)) {
		   PPER "Error in operon #%d %s\n",
				 nb_ope+1, ope_temp->fseqptr->name);
	   	   exit(2);
	       }

					/* take min and max rel_number	*/

	       if (ope_temp->ignore > max_ope_rel_number)
		   max_ope_rel_number = ope_temp->ignore;

	       if (ope_temp->ignore < min_ope_rel_number)
		   min_ope_rel_number = ope_temp->ignore;

	       ope[nb_ope++] = ope_temp;
	       ope_temp = NewOperonSequence(parms.max_gen_kept);

	}

        FreeOperonSequence(ope_temp);
	fclose(operon_filin);

	/* ---------------------------- */    
	
	/* ON_VERBOSE */
	PPCOMPL "\n" );
	PPCOMPL "/ %d valid operon sequences read (total read: %d)\n",
		 nb_ope, nb_ope_tot);

	if (nb_ope <= 0) {
	   PPER "No operon genes in operon file !\n");
	   exit(3);
	}

	/* ---------------------------- */    
	/* read DB			*/

	nb_gen = 0;
	maxlen_gen = 0;
	nb_fastp_gen_elim = 0;
	max_gen_rel_number = -1000000;
	min_gen_rel_number =  10000000;

	gen_temp = NewGenomeSequence(parms.max_gen_kept);


	max_gen_seq = MAX_GEN_SEQ;

	if ((gen = NEWN(GenomeSeqPtr,max_gen_seq)) ==  NULL) {
	    (void) Erreur(
		  "Not enough memory for allocating genome sequences !!!",
		  ErrManySq);
	}

	ON_VERBOSE
	    PPCOMPL "/ reading Genome DB %s\n", parms.dbfile);
	
	RewindFastaDB(stdin);

	nerrseq = 0;

	while (ReadGenomeSequence(stdin, gen_temp,
				     parms.gen_std_fasta_format)) {

#ifdef ROTATOR /* too much output	*/
	    if (   (parms.opt_verbose)
		&& ((nb_gen % Seq_Rotator) == (Seq_Rotator - 1))) {
		    (void)fprintf(stderr, "\r");
		    PPCOMPL "reading genome sequence %4d\t\t", nb_gen+1);
            }
#endif		    
	    if (nb_gen >= max_gen_seq-1) {
		 gen = REALLOC(GenomeSeqPtr, gen, max_gen_seq+INC_GEN_SEQ);
		 if (! gen) {
		     (void) Erreur(
			    "Too many sequences in DB, not enough memory",
			    ErrManySq);
                      break;
                 }
		 max_gen_seq += INC_GEN_SEQ;
	    }
	    	
    	    if (! gen_temp->fseqptr->ok) {
		PPER "IO error with genome sequence #%d\n",
					nb_gen+1);
		nerrseq++;
		continue;
	    }

    	    if (gen_temp->fseqptr->length <= 0) {
		PPER "Length error : genome sequence #%d (name:%s) length = %d\n",
			nb_gen+1,
			gen_temp->fseqptr->name,
			(int)(gen_temp->fseqptr->length));
		nerrseq++;
		continue;
	    }

/* debug */
#if 0
		WriteGenomeSequence(stderr, gen_temp, 50);
#endif



	    (void) UpperStr(gen_temp->fseqptr->seq);
	    
	    if (parms.opt_sym_replace)
		(void) ReplaceInStr(gen_temp->fseqptr->seq, parms.alphabet,
				    parms.sym_replace);
	    if (parms.opt_sym_delete) {
		ndel = DeleteInStr(gen_temp->fseqptr->seq, parms.alphabet);
		if (ndel > 0)
		      gen_temp->fseqptr->length -= ndel;
	    }


	   				/* must index sequence for	*/
					/* fastp AND besfit		*/ 

	    if (! IndexSequence(gen_temp->fseqptr->seq, parms.alphabet)) {
		PPER "Error in sequence #%d %s\n",
				 nb_gen+1, gen_temp->fseqptr->name);
		nerrseq++;
		continue;
	    }

	    if (gen_temp->fseqptr->length > maxlen_gen)
		maxlen_gen = gen_temp->fseqptr->length;

					/* take min and max rel_number	*/

	    if (gen_temp->rel_number > max_gen_rel_number)
		   max_gen_rel_number = gen_temp->rel_number;

	    if (gen_temp->rel_number < min_gen_rel_number)
		   min_gen_rel_number = gen_temp->rel_number;

	    gen[nb_gen++] = gen_temp;

	    gen_temp = NewGenomeSequence(parms.max_gen_kept);


	}	    

	FreeGenomeSequence(gen_temp);

	/* ON_VERBOSE */
	PPCOMPL "\n");
	PPCOMPL "/ %d genome sequences (total read: %d)\n", nb_gen,
			nb_gen+nerrseq);

	if (nb_gen <= 0) {
	   PPER "No genome genes in genome file !\n");
	   exit(4);
	}


	sPrintHeader(argn, argv, nb_ope, nb_gen, &parms);

	/* ---------------------------- */    
	/* initialize various counters	*/

	nb_ali_fast = 0;
	nb_ali2 = 0;

	if (parms.bstat) {
	   if (!(bstat = NEWN(int, nb_gen)))
		 MyError(stderr,2,
		     "\nMemory error: cannot allocate for bstat (source:%s line %d)\n",
		     __FILE__, __LINE__);

	   InitsShufseq(maxlen_gen + 1);
        }

	/* ---------------------------- */    
	/* fastp run                    */

	if (parms.do_fastp != 0) {
	 
	  /* ---------------------------- */    
	  /* diagonals allocation 	  */

	  if (! (diags = NewDiagTable(maxlen_ope * maxlen_gen + 1)))
	    exit(ErrMemory);

	  /* ---------------------------- */
	  /* verify options               */

	  /* ---------------------------- */    
	  if (parms.fastp_aliscore || parms.fastp_mxscore) {
	     parms.diag_integ = diags->diag_integ = 0;
	  }
	  else if (parms.diag_integ >= 0)
	     diags->diag_integ = parms.diag_integ;
          else
	      MyError(stderr,11,
		 "%s: diagonal integration must be positive (option -i). Exiting...\n",
	         argv[0]);

	  /* ---------------------------- */    
	  /* print header                 */

	  sPrintFastpHeader(&parms);
	  

	  /* ---------------------------- */    
	  /* hash allocation		  */


	  if (! (hashtab = NewHashTable(parms.alphabet, parms.kuple)))
	    exit(ErrMemory);

	

	 /* ---------------------------- */    
	 /* open output file if needed	*/
	
	  fast_alistream = (parms.fast_opt_align  ?
			      fopen(parms.fast_alifile, "w") : NULL);

	 /* ---------------------------- */    
	 /* fastp loop on entries	*/
	
	  ON_VERBOSE
	      PPCOMPL "/ fastp run on operon genes (%d sequences)\n", nb_ope);


    						/* --------------------	*/
						/* fastp loop on	*/
						/* operon genes		*/
						/* --------------------	*/
						/* if parms.do_fastp	*/
						/* else only bestfit	*/
						/* run, see below	*/
						/* --------------------	*/

	  for (i_ope = 0 ; i_ope < nb_ope; i_ope++) {

#ifdef ROTATOR /* too much output	*/
	    ON_VERBOSE {
		   (void)fprintf(stderr, "\r");
		   PPCOMPL "sequence %4d / %4d\t\t",
		   i_ope+1, nb_ope);
	    }
#endif
	    if (! (ope[i_ope]->fseqptr->ok ))
		continue;

	    (void) ResetHashTable(hashtab);

	    (void) HashUserSeq(hashtab, ope[i_ope]->fseqptr->seq);

	    diags->diag_offset = ope[i_ope]->fseqptr->length;


						/* --------------------	*/
						/* fastp loop on	*/
						/* genome genes		*/
    						/* --------------------	*/
            n_ali = 0;
	    for (i_gen = 0 ; i_gen < nb_gen; i_gen++) {
    						

		if (! (gen[i_gen]->fseqptr->ok ))
		    continue;
		    
						/* ----------------------------	*/
						/* exclude gene	with		*/
						/* same name as ope[i_ope]	*/
						/* ----------------------------	*/
		if (parms.exclude_self)
		   if (! strcmp(gen[i_gen]->fseqptr->name, ope[i_ope]->fseqptr->name))
		      continue;
		    

						/* ---------------- */
						/* Pas de fastp si */
						/* seuil longueur   */
						/* pas bon	    */
						/* ---------------- */
                if (cparms.len_thresh > 0.0 && LENRATIO(ope[i_ope]->fseqptr->length,
				gen[i_gen]->fseqptr->length) > cparms.len_thresh) {
		  continue;
		}

		nb_ali_fast++;
				    
		diags->diag_size = ope[i_ope]->fseqptr->length +
						gen[i_gen]->fseqptr->length;

		ComputeDgnl(gen[i_gen]->fseqptr->seq, hashtab, diags);
		
		fastbest = IntegrateDgnl(diags, &score);		


		if (parms.fastp_aliscore == 1) {
				/* compute fasta score on alignment if needed	*/

				/* returned score is between 0.0 and 1.0	*/
		    fscore = Aliscore(ope[i_ope]->fseqptr->seq, gen[i_gen]->fseqptr->seq,
				ope[i_ope]->fseqptr->length, gen[i_gen]->fseqptr->length,
				fastbest - ope[i_ope]->fseqptr->length, &matrix);

		    fscore *= (float)HISCORE;

		}
		else if (parms.fastp_mxscore == 1) {
				/* compute fasta score on maximum subsum if needed	*/

		    fscore = HspScore(ope[i_ope]->fseqptr->seq, gen[i_gen]->fseqptr->seq,
				ope[i_ope]->fseqptr->length, gen[i_gen]->fseqptr->length,
				fastbest - ope[i_ope]->fseqptr->length, &matrix);
		    fscore *= (float)HISCORE;
#if 1
{
			/* comparaison des 4 scores: fastp diag, fasta decalage (fs),	*/
			/* fasta max subsum (fscore) et bestfit (bs)		*/
float fs, bs, fp, mx;
		mx = (fscore*100.0)/(float)HISCORE;
		norme =   MIN(ope[i_ope]->fseqptr->length,
				gen[i_gen]->fseqptr->length) -
			        parms.kuple;
		if (score > norme)
		      score = norme;

		fp = (score * 100.0) / (float) norme;

		fs = (float)100.0 * Aliscore(ope[i_ope]->fseqptr->seq, gen[i_gen]->fseqptr->seq,
			ope[i_ope]->fseqptr->length, gen[i_gen]->fseqptr->length,
			fastbest - ope[i_ope]->fseqptr->length, &matrix);
		fdist = (float)NwsAlign(ope[i_ope]->fseqptr->seq,
				gen[i_gen]->fseqptr->seq,
				ope[i_ope]->fseqptr->length,
				gen[i_gen]->fseqptr->length,
				matrix.weight,  bparms.wopen,
				bparms.wext, NULL, bparms.fglobal)/SCORENORM;
		fnorm = (bparms.fglobal ?
			    (float) MAX2(ope[i_ope]->fseqptr->length,
			                 gen[i_gen]->fseqptr->length)
				 : (float) MIN2(ope[i_ope]->fseqptr->length,
						gen[i_gen]->fseqptr->length));
		
		bs = (float)100.0 - (fdist * (float)100.0 / fnorm);
		fprintf(stderr,"fastpscore=%7.3f aliscore=%7.3f mxscore=%7.3f bestscore=%7.3f\n",fp, fs,mx,bs);
}
#endif

		}
		else {
				/* score is # of matching kuples	*/

		    norme =   MIN(ope[i_ope]->fseqptr->length,
				gen[i_gen]->fseqptr->length) -
			        parms.kuple;
		
		    if (score > norme)
		        score = norme;

		    fscore = (score * (float)HISCORE) / (float) norme;
		
		}

#if 0
		(void)fprintf(stderr,"ope %5d / gen %5d fscore: %6.2f (low thresh:%6.2f)\n",
				i_ope+1, i_gen+1, fscore * (float)NORMSCORE / (float)HISCORE,
				parms.fast_lo_threshold  * (float)NORMSCORE / (float)HISCORE);
#endif

#if 0				
		if (parms.fastp_aliscore == 1 || parms.fastp_mxscore == 1) {
		   PPCOMPL
			"\n%-20s:%-20s:%f (minlen %5d fastbest %d shift %d)",
			ope[i_ope]->fseqptr->name,
			gen[i_gen]->fseqptr->name,
			fscore,
			MIN(ope[i_ope]->fseqptr->length, gen[i_gen]->fseqptr->length),
			fastbest, fastbest-ope[i_ope]->fseqptr->length);
		}
		else {
		   PPCOMPL
			"\n%-20s:%-20s:%f (score:%5d norme %5d fastbest %d shift %d)",
			ope[i_ope]->fseqptr->name,
			gen[i_gen]->fseqptr->name,
			fscore, score, norme, fastbest, fastbest-ope[i_ope]->fseqptr->length);
		}
#endif

		if (parms.bstat) {

		    shufseq = ShuffleSeqInStatic(gen[i_gen]->fseqptr->seq,
						    gen[i_gen]->fseqptr->length, -1);
		    ComputeDgnl(shufseq, hashtab, diags);
		
		    fastbest = IntegrateDgnl(diags, &score);		

					/* same norme for shuffled seq	*/
		    if (parms.fastp_aliscore == 1) {
				/* returned score is between 0.0 and 1.0	*/
		       bstat[n_ali] = (int)((float)HISCORE * Aliscore(ope[i_ope]->fseqptr->seq, shufseq,
				ope[i_ope]->fseqptr->length, gen[i_gen]->fseqptr->length,
				fastbest - ope[i_ope]->fseqptr->length, &matrix));
		    }
		    else if (parms.fastp_mxscore == 1) {
				/* returned score is between 0.0 and 1.0	*/
		       bstat[n_ali] = (int)((float)HISCORE * HspScore(ope[i_ope]->fseqptr->seq, shufseq,
				ope[i_ope]->fseqptr->length, gen[i_gen]->fseqptr->length,
				fastbest - ope[i_ope]->fseqptr->length, &matrix));
		    }
		    else {
		       if (score > norme)
		          score = norme;

		       bstat[n_ali] = (int)((score * (float)HISCORE) / (float)norme);
		    }
#if 0
		    (void)fprintf(stderr,"XX alea score %d\n",bstat[n_ali]);
#endif
		    n_ali++;
		}



		if (fscore > parms.fast_lo_threshold) {

#if 0						/* debug 	*/
                    (void)fprintf(stderr,
				"\nope %d / gen %d \t\t->kept with score:%f (internal stored score: %d)",
				i_ope+1, i_gen+1, fscore*(float)NORMSCORE/(float)HISCORE,
				(int)fscore);
#endif

					/* ----------------------------	*/
		      			/* keep gen[i_gen] for		*/
					/* comparing with ope[i_ope]	*/
					/* ----------------------------	*/
		    KeepGeneInOpeF(ope[i_ope], gen[i_gen], fscore);
		    KeepOpeInGeneF(gen[i_gen], ope[i_ope], fscore);

		}

	    }   /* end for i_gen	*/
	   
	    if (parms.bstat)
		ComputeStatForOperon(ope[i_ope], bstat, n_ali);



	   }   /* end for i_ope	*/

	   ON_VERBOSE {
	       (void)fprintf(stderr,"\n");
	   }



	   /* ---------------------------- */
	   /* print kept gene if very verbose	*/

	   ON_VERY_VERBOSE {
	     (void)fprintf(stderr,"\n");
	     PPCOMPL "List of similar genes after Fastp:\n");
	     for (i_ope = 0; i_ope < nb_ope; i_ope++)
				/* write kept gene without correction	*/
				/* and even if len ratio is bad		*/
	       WriteKeptGene(ope[i_ope], -1.0, 0, (float)NORMSCORE/(float)HISCORE, parms.bstat,
				parms.short_output, stderr);
	   }


	   /* --------------------------------- */
	   /* free genome sequence which	*/
	   /* are not similar to at least	*/
	   /* one operon			*/

	  i_new_gen = 0;

	  for (i_gen = 0 ; i_gen < nb_gen ; i_gen++) {
	    if (gen[i_gen]->kept_by == NO_KEPT) {
		(void) FreeGenomeSequence(gen[i_gen]);
		gen[i_gen] = NULL;
	    }
	    else {
	       gen[i_new_gen] = gen[i_gen];
	       i_new_gen++;
	    }
	  }

	  nb_fastp_gen_elim = nb_gen - i_new_gen;
	  nb_gen = i_new_gen;
	  
	  /* ---------------------------- */
	  /* close output files		*/
	
	  if (fast_alistream)
	    (void)fclose(fast_alistream);


	}	/* end if (parms.do_fastp != 0)	 	*/



	/* ---------------------------- */    
	/* bestfit stuff		*/


	nb_ali_best = 0;

	compar = ComparBestFitScoreGenKept;


	sPrintBestfitHeader(nb_ope,
			(parms.do_fastp != 0) ? ope[0]->max_gen_kept : nb_gen,
			parms.do_fastp, bestmatfile, bparms.fglobal,
			bparms.wopen, bparms.wext);

	ON_VERBOSE
	      PPCOMPL "/ bestfit run on %d operon genes each against %d genomic genes\n",
			nb_ope, (parms.do_fastp != 0) ? ope[0]->max_gen_kept: nb_gen);

    						/* --------------------	*/
						/* bestfit loop on	*/
						/* operon genes		*/
						/* --------------------	*/


						/* if scores to store	*/
						/* free bstat		*/
	if (parms.bstat && parms.store_aleas)
	    FREE(bstat);

	/* OLD ????
	if (parms.store_scores) {
	   tab_scor = NEWN(char, nb_ope*nb_gen);
	   ON_VERBOSE
	      PPCOMPL "/ Allocating %d octets for score storage\n",
			nb_ope*nb_gen);
	   if (! tab_scor)
	      MyError(stderr,4,
		 "Not enough memory for scores, re-run %s with no -X option, exiting...\n",
	         argv[0]);
	}
	*/

	if (parms.bstat && ! parms.do_fastp && parms.store_aleas) {

	   ON_VERBOSE
	      PPCOMPL "/ Allocating %d octets for shuffled score storage\n",
			nb_ope*nb_gen);

	   tab_alea = NEWN(char, nb_ope*nb_gen);
	   if (! tab_alea)
	      MyError(stderr,5,
	      "Not enough memory for shuffled scores, re-run %s with no -T option, exiting...",
	      argv[0]);
	}

	for (i_ope = 0 ; i_ope < nb_ope; i_ope++) {

#ifdef ROTATOR /* too much output	*/
	    ON_VERBOSE {
	       (void)fprintf(stderr, "\r");
	       PPCOMPL "sequence %4d / %4d\t\t", i_ope+1, nb_ope);
	    }
#endif
	    if (! (ope[i_ope]->fseqptr->ok))
		continue;




						/* --------------------	*/
						/* if fastp has been	*/
						/* done	bestfit over	*/
						/* fastp-similar genome	*/
						/* genes with operon	*/
						/* gene ope		*/
						/* --------------------	*/
						/* if fastp has NOT	*/
						/* been done bestfit	*/
						/* over all genome	*/
						/* genes		*/
						/* --------------------	*/
	    
	    ngen_todo = (parms.do_fastp != 0) ? 
				ope[i_ope]->actual_gen_kept :
				nb_gen;

						/* --------------------	*/
						/* bestfit loop       	*/
						/* --------------------	*/

    	    n_ali = 0;
	    for (i_gen = 0 ; i_gen < ngen_todo; i_gen++) {

		if (parms.do_fastp != 0) {
						/* if fastp done, 	*/
						/* bestfit is only	*/
						/* need for kept	*/
						/* genome genes		*/
		   gen_ptr = ope[i_ope]->gen_kept[i_gen].gen_ptr;

						/* and excluded genes	*/
						/* have been already	*/
						/* excluded		*/
		}
		else {
						/* if fastp not done, 	*/
						/* bestfit is to be	*/
						/* done for all		*/
						/* genome genes		*/
		   gen_ptr = gen[i_gen];
						/* ----------------------------	*/
						/* exclude gene	with		*/
						/* same name as ope[i_ope]	*/
						/* ----------------------------	*/
		   if (parms.exclude_self)
		      if (! strcmp(gen[i_gen]->fseqptr->name, ope[i_ope]->fseqptr->name))
		         continue;

						/* ----------------------------	*/
						/* must control if genome	*/
						/* gene is ok because it has	*/
						/* not been controled in fastp	*/
						/* ----------------------------	*/
		   if (! (gen_ptr->fseqptr->ok))
		       continue;
						/* ----------------------------	*/
						/* we must control length	*/
						/* of genome gene		*/
						/* because it has		*/
						/* not been controled in fastp	*/
						/* ----------------------------	*/
                   if (cparms.len_thresh > 0.0 && LENRATIO(ope[i_ope]->fseqptr->length,
				gen_ptr->fseqptr->length) > cparms.len_thresh) {
		      continue;
		}

		}
		    

		nb_ali_best++;
				   
		dist = NwsAlign(ope[i_ope]->fseqptr->seq,
				gen_ptr->fseqptr->seq,
				ope[i_ope]->fseqptr->length,
				gen_ptr->fseqptr->length,
				matrix.weight,  bparms.wopen,
				bparms.wext, NULL, bparms.fglobal);

		fdist = (float) dist/ SCORENORM;

		fnorm = (bparms.fglobal ?
			    (float) MAX2(ope[i_ope]->fseqptr->length,
			                 gen_ptr->fseqptr->length)
				 : (float) MIN2(ope[i_ope]->fseqptr->length,
						gen_ptr->fseqptr->length));
		

		fscore = (float)HISCORE - (fdist * (float)HISCORE / fnorm);


#if 0
		ON_VERBOSE {
		    PPCOMPL "Operon: %s Genome %s Dist: %.1f Score: %.1f\n",
			ope[i_ope]->fseqptr->name, gen_ptr->fseqptr->name,
				 fdist, fscore * (float)NORMSCORE / (float)HISCORE); 
		}
#endif



					/* ----------------------------	*/
					/* pour ne pas confondre	*/
					/* score bestfit		*/
					/* et score fastp		*/
					/* ----------------------------	*/
		fscore += BASE_BEST;
					/* ----------------------------	*/
					/* for stats, if fastp has	*/
					/* not been done before (else	*/
					/* stats are computed on fasta	*/
					/* scores)			*/
					/* ----------------------------	*/

		if (parms.bstat && ! parms.do_fastp) {

		   dist = NwsAlign(ope[i_ope]->fseqptr->seq,
				ShuffleSeqInStatic(gen_ptr->fseqptr->seq,
					gen_ptr->fseqptr->length, -1),
				ope[i_ope]->fseqptr->length,
				gen_ptr->fseqptr->length,
				matrix.weight,  bparms.wopen,
				bparms.wext, NULL,
				bparms.fglobal);

		   if (parms.store_aleas) {
		      tab_alea[i_ope*nb_gen+i_gen] = (char)(BASE_BEST + (float)HISCORE -
					 (((float) dist/ SCORENORM) * (float)HISCORE / fnorm));
		   }
		   else {
		      bstat[n_ali] = BASE_BEST + (float)HISCORE -
			    (((float) dist/ SCORENORM) * (float)HISCORE / fnorm);
		   }
		   n_ali++;
		   

		}

					/* ----------------------------	*/
					/* garde le score bestfit a	*/
					/* la place du score fastp	*/
					/* ----------------------------	*/

		if (parms.do_fastp != 0) {	
					/* ----------------------------	*/
					/* fastp has been done before	*/
					/* so ope[i_ope]->gen_kept is	*/
					/* not empty			*/
					/* ----------------------------	*/

		    ope[i_ope]->gen_kept[i_gen].score = (SCORE)((int)fscore);
		    KeepOpeInGeneB(gen_ptr, ope[i_ope], fscore, -1.0);
	        }
		else {

		    KeepGeneInOpeF(ope[i_ope], gen_ptr, fscore);
		    KeepOpeInGeneF(gen_ptr, ope[i_ope], fscore);

		}


					/* ----------------------------	*/
	    }   			/* end for i_gen		*/
					/* ----------------------------	*/
	   
					/* tri des scores			*/

	    qsort ((void *) (ope[i_ope]->gen_kept),
			(size_t) (ope[i_ope]->actual_gen_kept),
			(size_t) sizeof(*(ope[i_ope]->gen_kept)), compar);

	    if (parms.bstat && ! parms.do_fastp)
		ComputeStatForOperon(ope[i_ope], bstat, n_ali);
	    

	    ON_ONLY_SIMILARITY
                  WriteKeptGene(ope[i_ope], cparms.len_thresh, BASE_BEST,
				(float)NORMSCORE/(float)HISCORE, parms.bstat,
				parms.short_output, stdout);

					/* ----------------------------	*/
	}   				/* end for i_ope		*/
					/* ----------------------------	*/


        ON_VERBOSE PPCOMPL "\n");

	if (parms.bstat)
		FREE(bstat);


	/* ---------------------------- */
	/* print kept gene if very verbose	*/

	ON_VERY_VERBOSE {
	     (void)fprintf(stderr,"\n");
	     PPCOMPL "List of similar genes after Bestfit:\n");
				/* write kept gene with		*/
				/* BASE_BEST correction   	*/
				/* and even if len ratio is bad	*/
	     for (i_ope = 0; i_ope < nb_ope; i_ope++)
	       WriteKeptGene(ope[i_ope], -1.0, BASE_BEST,
				(float)NORMSCORE/(float)HISCORE,
				parms.bstat, parms.short_output, stderr);
	}
	


	ON_ONLY_SIMILARITY {
		exit(0);
	}

				/* if BBH	*/
	if (cparms.do_bbh)
	   sPrintBBhHeader(&cparms);
	else			/* Classes	*/
	   sPrintClassHeader(&cparms);

#if 0
	fprintf(stderr, "\nList of homologs before classes:\n");
	for (i_ope = 0; i_ope < nb_ope; i_ope++)
	   if (! (ope[i_ope]->homolog.gen_ptr == NOT_G || ope[i_ope]->homolog.gen_ptr == NULL))
	      fprintf(stderr, "homol: %s %s %d\n", ope[i_ope]->mnemo,
			ope[i_ope]->homolog.gen_ptr->mnemo,
			(int)(ope[i_ope]->homolog.score *
			(float)NORMSCORE/(float)HISCORE));
	   else
	      fprintf(stderr, "homol: %s NULL\n", ope[i_ope]->mnemo);
#endif

	
					/* ----------------------------	*/
					/* if combi, the relation is	*/
					/* not bijective		*/
					/* ----------------------------	*/
	if (cparms.combi) {
					/* must reset the scores	*/
					/* in order to have the correct	*/
					/* threshold			*/
	   ResetAllScores(ope, nb_ope, - BASE_BEST);

	   Combinatorique(ope, nb_ope, &cparms, parms.bstat);

	   exit(0);
	}
					/* ----------------------------	*/
					/* if bijective, bestfit must	*/
					/* be done between all best	*/
					/* genes of all operon genes	*/
					/* and all operon genes		*/
					/* ----------------------------	*/
	else if (cparms.bijectif) {
	      if (parms.bstat && ! parms.do_fastp) {
	         ComputeTabStat(tab_alea, i_ope, nb_gen, (int)fscore);
	      }
	      else if (sBestfitBestGenAllOpe(ope, nb_ope, &matrix, &bparms, &parms,
				cparms.len_thresh, cparms.seuil_score + BASE_BEST,
				cparms.bijectif, &nb_ali2)) {
	         MyError(stderr, 0, ERROR_TAG, 
		 "*error*: Cannot complete bestfit between best genomic genes and all operon genes");
	      }
	}
	else {
					/* ----------------------------	*/
					/* compute "homologs"		*/
					/* with no bijection		*/
					/* ----------------------------	*/
	      SetAllOpeHomologs(ope, nb_ope, cparms.len_thresh,
			cparms.seuil_score + BASE_BEST, cparms.bijectif,
			parms.opt_verbose);


	}

					/* ----------------------------	*/
					/* output BBH and exit 		*/
					/* if user asked		*/
					/* ----------------------------	*/
	if (cparms.do_bbh) {
					/* ---------------------------- */    
					/* print footer       		*/

	    sPrintFooter(nb_ope, nb_gen, nb_ali_fast,
			nb_ali_best, nb_ali2,
			nb_fastp_gen_elim);
	
					/* ---------------------------- */    
					/* correct the bestfit scores	*/
	
	    ResetAllScores(ope, nb_ope, - BASE_BEST);

	    sPrintBBH(ope, nb_ope, parms.bstat);

            if (bparms.best_print) {
	        sPrintAlignments(ope, nb_ope, &matrix, &bparms, parms.alphabet);
	    }

#if 0 /* debug */
	fprintf(stderr, "\nList of homologs after classes:\n");
	for (i_ope = 0; i_ope < nb_ope; i_ope++)
	   if (! (ope[i_ope]->homolog.gen_ptr == NOT_G || ope[i_ope]->homolog.gen_ptr == NULL))
	      fprintf(stderr, "homol: %s %s %d\n", ope[i_ope]->mnemo,
					ope[i_ope]->homolog.gen_ptr->mnemo,
					(int)(ope[i_ope]->homolog.score *
					(float)NORMSCORE/(float)HISCORE));
	   else
	      fprintf(stderr, "homol: %s NULL\n", ope[i_ope]->mnemo);
#endif
	   exit(0);
	}
					/* ----------------------------	*/
					/* etablissement des classes 	*/
					/* operoniques			*/
					/* ----------------------------	*/
	if (cparms.operon_class) {
	      nclass_ope = ComputeR1Classes(ope, nb_ope, cparms.nseuil,
			cparms.ope_circular, min_ope_rel_number, 
			max_ope_rel_number, parms.opt_verbose);
	}
	else 
	    nclass_ope = 1;

	if (cparms.outR1) {
	         /* sortir les classes R1 i.e. intra-operoniques */
	         (void)fprintf(stdout, SUBOPE_NUMBER "Intra-operon classes: %d\n",
			nclass_ope);
              for (i_ope = 0; i_ope < nb_ope; i_ope++) 
		 fprintf(stdout, SUBOPE_TAG "%s %4d class: %4d (%s have homolog)\n",
				ope[i_ope]->mnemo, ope[i_ope]->ignore, ope[i_ope]->class,
				ope[i_ope]->homolog.gen_ptr != NOT_G ? "does":"does not");
	}

					/* ----------------------------	*/
					/* etablissement des classes	*/
					/* genomiques			*/
					/* ----------------------------	*/

					/* first, set the NULL		*/
					/* homol_ptr of gen to NOT_O	*/
	set_NULL2NOT_O_Gene(gen, nb_gen);

	nclass_gen = ComputeR2Classes(ope, nb_ope, gen, nb_gen, cparms.nseuil, 
			cparms.operon_class, cparms.circular,
			min_gen_rel_number, max_gen_rel_number,
			cparms.ope_circular, min_ope_rel_number, max_ope_rel_number,
			parms.opt_verbose);





	/* ---------------------------- */    
	/* print footer       		*/

	sPrintFooter(nb_ope, nb_gen, nb_ali_fast,
			nb_ali_best, nb_ali2,
			nb_fastp_gen_elim);

					
	/* ---------------------------- */    
	/* correct the bestfit scores	*/


	
	ResetAllScores(ope, nb_ope, - BASE_BEST);

	PrintResultR2(ope, nb_ope, nclass_gen, parms.bstat);

					/* si les singletons sont filtres	*/
					/* ils ne sont pas imprimes au dessus	*/
					/* on les imprime explicitement ici	*/
	if (cparms.output_singleton)
	      sPrintSingletons(ope, nb_ope, parms.bstat);

        if (bparms.best_print) {
	     sPrintAlignments(ope, nb_ope, &matrix, &bparms, parms.alphabet);
	}

	/* ---------------------------- */    
	/* free all buffers needed	*/
	/* for bestfit			*/

	FreeTableBuffers();

	/* ---------------------------- */    
	/* if needs of real sequences	*/
	/* should DeIndexSequence 	*/
	/* operon and genome genes	*/


	/* ---------------------------- */    
	/* free fastp stuff		*/
	/* only now because we need it	*/
	/* in sBestfitBestGenAllOpe	*/

	if (parms.do_fastp != 0) {
	  (void) FreeDiagTable(diags);
	  (void) FreeHashTable(hashtab);

	}

	/* ---------------------------- */    
	/* free sequence stuff		*/

	for (i_ope = 0 ; i_ope < nb_ope ; i_ope++)
	    (void) FreeOperonSequence(ope[i_ope]);

	free(ope);

	for (i_gen = 0 ; i_gen < nb_gen ; i_gen++)
	    (void) FreeGenomeSequence(gen[i_gen]);

	free(gen);

	exit(0);

	return 0;
}

#undef ON_VERBOSE


