/* ==================================================== */
/* @file: opscan_help.c					*/
/* Help & usage						*/
/* @history:						*/
/* @+	Apr. 97 <AV> first draft			*/
/* @+	Fev2000 <jompo> second draft			*/
/* ==================================================== */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Gtypes.h"
#include "opscan.h"
#include "opscan_help.h"
#include "operon.h"
#include "fastp.h"
#include "bali.h"
#include "class.h"
#include "tags.h"

/* ----------------------------------------------- */
/* printout opscan help				   */
/* ----------------------------------------------- */

#define PP fprintf(stderr,HELP_TAG

void PrintHelpOpscan(void)
{
	PP	"------------------------------------------			\n");
	PP	" Opscan with Operon Query  0.1 (mars 2000)                    \n");
	PP	"------------------------------------------			\n");
	PP	"synopsis :							\n");
	PP	"  Fastp then NWS (without end gap penalties) operon search in genome	\n");
	PP	"  database starting with an operon file			\n");
        PP      "                                                               \n");
	PP	"usage: 							\n");
	PP	"  opscan [options] -O operon_query [db_file]	\n");
        PP      "                                                               \n");
        PP      "options:                                                       \n");
        PP      "(*): not yet implemented                                       \n");
        PP      "----- General options   ------------------                     \n");
	PP	"-N           : no fastp, fastp is not run as a prefilter	\n");
	PP	"               i.e. only bestfit will be run between all	\n");
	PP	"               operon sequences and all genome sequences	\n");
	PP	"		(default: fastp is run as a prefilter)		\n");
	PP	"		Note: this option slows the run        		\n"); 
        PP      "                                                               \n");
	PP	"-F           : genomic sequences are in standard fasta format  \n");
	PP	"                (not in \"mnemo format\" (see README.opscan))	\n");
	PP	"                the genomic genes are supposed to be contiguous\n");
	PP      "                and in the order they are placed on the chromosome\n");
        PP      "                                                               \n");
	PP	"-E           : operon sequences are in standard fasta format   \n");
	PP	"                (not in \"mnemo format\" (see README.opscan))	\n");
	PP	"                the operon genes are supposed to be contiguous \n");
	PP      "                and in the order they are placed on the chromosome\n");
        PP      "                                                               \n");
	PP	"-n           : maximum number of genes separating two genes	\n");
	PP	"               to be considered in the same class		\n");
        PP      "               default: 5                                      \n");
        PP      "                                                               \n");
	PP	"-r <float>   : [r]atio length threshold (float)		\n");
	PP	"               ratio length threshold between operon and genomic\n");
	PP	"               genes to be candidate to be \"homolog\",	\n");
	PP	"               i.e. genomic gen is retained for comparison with\n");
	PP      "               operon ope, if ratio of the longest sequence	\n");
	PP	"               length divided by the shortest one is less than\n");
	PP      "               \"ratio length threshold\"	\n");
	PP	"               In other words, if one sequence is greater than r\n");
	PP	"               times the other, or lesser than 1/r times the other,\n");
	PP	"               they won't be compared, and they never can be \"homolog\"\n");
	PP	"               Example: -r 2 means that genomic gene g and operon\n");
	PP	"               gene o won't be compared if length of g is greater\n");
	PP      "               than twice or lesser than 0.5 times length of o\n");
	PP	"               (no threshold when r is less than 1.0)		\n");
	PP	"               default: -r 1.3                                 \n");
	PP	"								\n");     
	PP	"-c           : chromosome is circular [default: Off]    	\n");
        PP      "                                                               \n");
	PP	"-b           : not [b]ijective:  two or more operon genes can  \n");
	PP      "               be \"homolog\" with the same genomic gene\n");
	PP	"                [default: Off, i.e. one to one	relation\n");
	PP      "                between operon gene and genomic gene: \n");
	PP      "                the genomic gene must be the most similar to the \n");
	PP      "                operon gene AND the operon gene must be the most\n");
	PP      "                similar to the genomic gene]\n");
        PP      "                                                               \n");
	PP	"-S           : no classes are computed, only similarity    	\n");
        PP      "               between operon genes and genome genes are       \n");
        PP      "               output                                          \n");
        PP      "                                                               \n");
	PP	"-H           : no classes are computed, only double best    	\n");
        PP      "               hits between operon genes and genome genes are  \n");
        PP      "               output                                          \n");
        PP      "                                                               \n");
	PP	"-s           : Do not output [s]ingletons [Default: Off]    	\n");
        PP      "               (single couples operon/genomic genes are        \n");
        PP      "               output by default, but they are not considered	\n");
	PP      "               as classes)		                        \n");
        PP      "                                                               \n");
	PP	"-Q           : classes are computed also for operon genes  	\n");
        PP      "                                                               \n");
	PP	"-C           : output also R1 class (i.e. sub-operons in the  	\n");
	PP	"               operon genes)				  	\n");
        PP      "                                                               \n");
	PP	"-U           : operon gene set is circular [default: Off]    	\n");
        PP      "                                                               \n");
	PP	"-K integer   : number of similar sequences to keep in first run\n");
	PP      "               (fastp or bestfit) for one entry\n");
	PP	"               (default:%d)					\n", MAX_GEN_KEPT);
        PP      "                                                               \n");
	PP	"-X           : genomes gene(s) with same name as an operon gene\n");
        PP      "               are discarded from the best hits for this operon\n");
        PP      "               gene                                            \n");
        PP      "                                                               \n");
        PP      "-R char      : [R]eplace unknown symbol by 'char'              \n");
        PP      "               default: -R 'X'              \n");
	PP      "               Note: alphabet is [A-Z]                         \n");
        PP      "               see also: -D                                    \n");
        PP      "                                                               \n");
        PP      "-D           : [D]elete unknown symbol                  \n");
        PP      "               default: off               \n");
        PP      "               see also: -R                                    \n");
        PP      "                                                               \n");
	PP	"-Z           : statistics are computed on shuffled genomic     \n");
	PP      "               sequences    \n");
        PP      "               If fastp is to be done, statistics are computed \n");
        PP      "               on fastp scores of shuffled genomic sequences   \n");
        PP      "               If only bestfit (-N option) statistics are      \n");
        PP      "               computed on bestfit scores of shuffled genomic \n");
        PP      "               sequences (this option doubles computing time)  \n");
        PP      "               Statistics consist of a number which is the     \n");
        PP      "               percentage of random shuffled genomic sequences \n");
        PP      "               that shows better (or equal) score than the     \n");
        PP      "               sequence of interest                   \n");
        PP      "                                                               \n");
        PP      "-T           : s[T]ores bestfit shuffled scores                \n");
        PP      "               Speeds up, but take huge memory                 \n");
        PP      "                                                               \n");
        PP      "-a           : [experimental] combinatoric search of operons   \n");
        PP      "               based upon a systematic search of all potential \n");
        PP      "               operons (even if genomic genes are not the      \n");
        PP      "               bijective \"homologs\" of the operon genes, but \n");
        PP      "               only in the list of kept genes of the operon    \n");
        PP      "               genes)                                          \n");
        PP      "                                                               \n");
	PP	"-v           : [v]erbose					\n");
	PP	"               completion verbosity				\n");
	PP	"               default = Off		    			\n");
	PP	"								\n");
	PP	"-V           : [Very] verbose					\n");
	PP	"               completion + huge verbosity			\n");
	PP      "               (alignments are output)                         \n");
	PP	"               default = Off		    			\n");
	PP	"								\n");
	PP      "-J           : short output, line have following format:       \n");
	PP      "                <ope> <gen> <score> <len_ratio> <nbgen>        \n");
	PP      "                where <nbgen> is the number of gene kept with  \n");
	PP      "                this <ope>                                     \n");
	PP	"								\n");
        PP      "-h           : [H]elp - print <this> help                      \n");
        PP      "                                                               \n");
        PP      "file formats                                                   \n");
        PP      "                                                               \n");
	PP	"operon_query:   \"mnemo format\" (special fasta format special)  \n");
	PP	"								\n");
        PP      "db_file:        \"mnemo format\" (special fasta format special)  \n");
	PP	"								\n");
	PP      "                 (see README.opscan)				\n");
        PP      "--------------------------------                               \n");


	return;
}

#undef PP

#define PP fprintf(stderr, HELP_TAG

void PrintFastpHelp(void)
{
	PP	"------------------------------------------			\n");
	PP	" options for the Fastp part of Opscan          		\n");
	PP	"------------------------------------------			\n");
	PP	"								\n");      
	PP	"----- Scoring options  ------------------			\n");
	PP	"								\n");
	PP	"-k integer   : [K]uple size (integer)				\n");
	PP	"               default: -k 2                                   \n");
	PP	"								\n");      
	PP	"-L integer   : [L]ower threshold (integer:0-100)		\n");
	PP	"               similarity lower threshold in fastp for		\n");
	PP	"		subsequent exclusion in bestfit search	 	\n");
	PP	"               [default: -L %d]     \n", DEFAULT_FAST_LO_THRESHOLD);
	PP	"		Range: 0 (no similarity)-100 (complete		\n");
	PP	"		similarity)					\n");
	PP	"								\n");      
	PP	"-x           : Fastp score is computed as the number of       	\n");
	PP	"               kuples which match for the best shift between	\n");
	PP	"               the two sequences (default: off, the Fastp score\n");
	PP	"               is computed based on the alignment score of the \n");
	PP	"               best shift alignment)                           \n");
	PP	"								\n");      
	PP	"-y           : Fastp score is computed as the maximum scoring 	\n");
	PP	"               segment for the best shift between the two	\n");
	PP	"               sequences (default: off, the Fastp score	\n");
	PP	"               is computed based on the alignment score of the \n");
	PP	"               best shift alignment)                           \n");
	PP	"								\n");      
	PP	"-i integer   : Only in case of -x option, the integer gives    \n");
	PP      "               the diagonal [I]ntegration in fastp computations\n");
	PP	"               default:  -i 2	(when -x option, else no	\n");
	PP	"                                integration is done)		\n");
	PP	"								\n");      
	PP	"----- Report options          ------------------		\n");
	PP	"								\n");
	PP	"-p filename  : [P]rint out alignments into filename(*)		\n");
	PP	"               default = Off					\n");
	PP	"								\n");
}

#undef PP



/* ----------------------------------------------- */
/* printout help				   */
/* ----------------------------------------------- */

#define PP fprintf(stderr, HELP_TAG

void PrintBestfitHelp(void)
{
	PP	"------------------------------------------			\n");
	PP	" options for bestfit     					\n");
	PP	"------------------------------------------			\n");
	PP	"								\n");      
	PP	"-l           : [L]ocal alignment (BestFit)			\n");
	PP	"               default = On 					\n");
	PP	"-g           : [G]lobal alignment				\n");
	PP	"               default = Off					\n");
	PP	"								\n");      
	PP	"-t <integer> : bestfit score threshold (0-100) for a couple	\n");
	PP      "               operon gene / genomic gene to be considered as  \n");
	PP      "               homolog	[default %d]		\n", DEFAULT_CLASS_SCORE_SEUIL);
        PP      "                                                               \n");
	PP	"----- Gap & Weights options  ------------------		\n");
	PP	"-M filename  : Scoring [M]atrix				\n");
	PP	"               default = <internal> Identity			\n");
	PP	"-I           : print matrix on stderr       			\n");
        PP      "                                                               \n");
	PP	"-o ff        : Gap [O]pening weigth				\n");
	PP	"               default = 1.2					\n");
	PP	"               Please note that it is a ratio multiplying the  \n");
	PP	"               lowest replacement score in the similarity      \n");
	PP	"               matrix                                          \n");
	PP	"-e ff        : Gap [E]xtension weigth				\n");
	PP	"               default = 0.8					\n");
	PP	"               Please note that it is a ratio multiplying the  \n");
	PP	"               lowest replacement score in the similarity      \n");
	PP	"               matrix                                          \n");
	PP	"								\n");      
	PP	" Note on gaps: opscan use distances to align sequences, and    \n");
	PP	" similarity log odds are translated to distances between amino-\n");
	PP	" acids between 0 (identity) and 100 (the lowest score in the   \n");
        PP      " similarity matrix). So an open gap ratio of 1.2 give a        \n");
        PP      " distance of 120, and an extension gap ratio of 0.8 give a     \n");
	PP	" distance of 80.                                               \n");
	PP	"								\n");      
	PP	"----- Report options          ------------------		\n");
	PP	"-P           : [P]rint out alignment				\n");
	PP	"               default = Off					\n");
	PP	"------------------------------------------			\n");
	PP	"file formats							\n");
	PP	"								\n");
	PP	"matrix_file:    blast format					\n");
	PP	"--------------------------------				\n");
	PP	"								\n");
}

#undef PP

/* ----------------------------------------------- */
/* printout usage and exit			   */
/* ----------------------------------------------- */

#define PP fprintf(stderr, HELP_TAG

int ExitUsage(int stat)
{
	PP	"usage: opscan [options] -O <operon_file> [db_file]\n");
	PP	"type \"opscan -h\" for help			\n");

	if (stat)
	    exit(stat);
	    
	return stat;
}


#undef PP
