/* ------------------------------------------------------------	*/
/* stat.c : part of opscan					*/
/* Jul 00: <jompo> Atelier de BioInformatique			*/
/* ------------------------------------------------------------	*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>		/* for time()			*/
#include <time.h>
#include <math.h>		/* for rand()			*/

#define REALLOC(type, ptr, n)     (type *) realloc(ptr, (n) * sizeof(type))

#include "util.h"
#include "stat.h"
#include "tags.h"

static char *sShufseq   = NULL;
static int  sShufseqLen = 0;
static int  sSeed       = -1;

/* ------------------------------------------------------------	*/
/* InitsShufseq: initialise sShufseq et met sShufseqLen a la 	*/
/* longueur allouee						*/
/* (+1 pour le '\0')						*/
/* ------------------------------------------------------------	*/
void InitsShufseq(int len)
{
    if (! (sShufseq = NEWN(char, (size_t) (len + 1))))
       MyError(stderr, 7, ERROR_TAG,
	  "InitsShufseq: not enough memory for shuffled sequence\n");

    sShufseqLen = len + 1;
    return;
}

/* ------------------------------------------------------------	*/
/* FreeShufseq: libere sShufseq					*/
/* ------------------------------------------------------------	*/
void FreeShufseq(void)
{
   if (sShufseq)
      free(sShufseq);

   sShufseqLen = 0;

   return;
}

/* ------------------------------------------------------------	*/
/* ShuffleSeqInStatic: genere une sequence shuffle en mono	*/
/* dans la statique sShufSeq a partir de s,			*/
/* retourne le pointeur sur la statique				*/
/* la longueur est donnee pour gagner du temps..		*/
/* ------------------------------------------------------------	*/
char *ShuffleSeqInStatic(char *s, int slen, int seed)
{
   char         temp;
   int          i, randoffset;
#ifdef SunOS
   int RAND_MAX = 1<<31;
#endif

   if (sShufseqLen < slen + 1) {
      
      if (! (sShufseq = REALLOC(char, sShufseq, 2*slen)))
	 MyError(stderr, 8, ERROR_TAG,
	    "NewShuffleSeq: not enough memory for shuffled sequence\n");

      sShufseqLen = 2*slen;

   }

				/* Initialise une seule fois	*/
				/* la graine aleatoire		*/
   if (seed < 0 && sSeed < 0) {
       sSeed = (int) time(0) & 0xfffffff;
       srand (seed);
   }
   else if (seed > 0) {
      sSeed = (int) seed;
      srand (seed);
   }

				/* we are sure to not copy 	*/
				/* more than slen char !!!	*/
   (void) strncpy(sShufseq, s, slen);
   sShufseq[slen] = '\0';

 
      				/* on ne va pas jusqu'a 1	*/
                                /* il faut que random/big < 1	*/
                                /* or random max == RAND_MAX	*/

   for (i = 0; i < slen-1; i++) {
      randoffset = ((double)rand()/((double)RAND_MAX+1))*(slen-i);
      temp = sShufseq[i];
      sShufseq[i] = sShufseq[i+randoffset];
      sShufseq[randoffset] = temp;
   }

   return sShufseq;

}

