/* ==================================================== */
/* @file: util.c					*/
/* general utilities functions				*/
/* @history:						*/
/* @+	Apr. 97 <AV> first draft			*/
/* ==================================================== */

#include <stdio.h>

/* #include <sys/statfs.h>*/
/* #include <time.h>       */        /* clock_t(sys/siginfo.h needs) */
/* #include <fcntl.h>    */          /* for O_RDONLY                 */
/* #include <errno.h> */             /* for errno                    */
/* #include <setjmp.h> */             /* for longjmp                  */



#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>              /* for errno                    */
#include <unistd.h>		/* for R_OK	*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/param.h>		/* for MAXPATHLEN	*/

#if ((!(defined Darwin)) && (!(defined SunOS)) && (!(defined Linux)))
#include <sys/procfs.h>              /* for prpsinfo or prusage_t */
#endif

#if 0
#include <malloc.h>
#endif

#include <stdarg.h>

#ifdef IRIX
/* #include <sys/statfs.h> */        /* for struct statfs            */
#else
#ifdef OSF1     /* DEC */
#include <sys/mount.h>
#endif /* endif OSF1     DEC */
#endif /* endif IRIX */


#include <fcntl.h>             /* for O_RDONLY                 */


#include "Gtypes.h"
#include "util.h"
#include "tags.h"

static MyPs sMy;		/* process usage */


/* -------------------------------------------- */
/* error reporting				*/
/* -------------------------------------------- */

int Erreur(char *msg, int stat)
{
	if (stat == 0)
	   (void)fprintf(stderr, WARNING_TAG "*Warning* [%d] %s\n", stat, msg);
	else
	   (void)fprintf(stderr, ERROR_TAG "*Error* [%d] %s\n", stat, msg);
 	
	if (stat > 0)
	   exit(stat);

	(void)fflush(stderr);
	
	return stat;
}

/* -------------------------------------------------------------------- */
void MyError(FILE *stream, int stat, char *tag, char *fmt, ...)
{
   va_list args;

   va_start(args,fmt);
   (void)fprintf(stream, tag);
   (void)vfprintf(stream, fmt, args);
   (void)fputc((int)'\n', stream);
   va_end(args);
   (void)fflush(stream);

   if (stat > 0)
      exit(stat);

   return;

}

/* -------------------------------------------- */
/* memory error					*/
/* -------------------------------------------- */

int MemoryErreur(char *where, int stat)
{
    char buffer[BUFSIZ];

    (void)sprintf(buffer, "Not enough memory in %s", where);

    return Erreur(buffer, stat);	
}

/* -------------------------------------------- */
/* open a file and assign it to stdin		*/
/* -------------------------------------------- */

int AssignToStdin(char *filename)
{
    char buffer[256];
	
    if (! freopen(filename, "r", stdin)) {
	(void)sprintf(buffer, "cant open file %s to read", filename);
	return Erreur(buffer, 0);
    }

    return 1;
}

/* -------------------------------------------- */
/* check file access				*/
/* -------------------------------------------- */

int CheckAccess(char *filename, char *mode)
{
    int	    amode = 0;
    FILE    *stream;
    char    buffer[256];

    if (strchr(mode, 'r'))  amode |= R_OK;
    if (strchr(mode, 'w'))  amode |= W_OK;
    if (strchr(mode, 'x'))  amode |= X_OK;
    if (strchr(mode, 'f'))  amode |= F_OK;
    
    if (amode & W_OK) {	/* if creation is needed */
	if ((stream = fopen(filename, "w")))
	    (void)fclose(stream);
    }
    
    if (access(filename, amode)) {
	(void)sprintf(buffer, "cant open file %s in mode %s", filename, mode);
	return Erreur(buffer, 0);
    }

    return 1;
}

/* ---------------------------------------------------- */
/* Trim leading spaces from string 'str'		*/
/* ---------------------------------------------------- */
char *StrTrimLeading(char *str)
{
    char *sb, *sn;
	
    if (! (str && isspace((int)(*str))))
	return str;
		
    for (sb = sn = str ; isspace((int)(*sn)) ; sn++) 
	/* nop */ ;

    while (*sn)
	*sb++ = *sn++;

    *sb = '\000';
	
    return str;
}

/* ---------------------------------------------------- */
/* uppercase string					*/
/* ---------------------------------------------------- */
char *UpperStr(char *str)
{
    char *s;
    
    for (s = str ; *s ; s++)
	if (IS_LOWER(*s))
	    *s = TO_UPPER(*s);

    return str;
}

/* ---------------------------------------------------- */
/* lowercase string					*/
/* ---------------------------------------------------- */
char *LowerStr(char *str)
{
    char *s;
    
    for (s = str ; *s ; s++)
	if (IS_UPPER(*s))
	    *s = TO_LOWER(*s);

    return str;
}

/* ---------------------------------------------------- */
/* delete bad symbols in string				*/
/* return number of deletion				*/
/* ---------------------------------------------------- */
int DeleteInStr(char *str, char *alphabet)
{
    char *s, *ss;
    int	 ndel;
   
    ndel = 0;
    for (s = ss = str ; *s ; s++)
	if (strchr(alphabet, *s)) {
	    *ss++ = *s;
	    ndel++;
	}

    *ss = '\000';

    return ndel;
}

/* ---------------------------------------------------- */
/* replace bad symbols in string			*/
/* ---------------------------------------------------- */
char *ReplaceInStr(char *str, char *alphabet, int repchar)
{
    char *s;
    
    for (s = str ; *s ; s++)
	if (! strchr(alphabet, *s))
	    *s = (char)repchar;

    return str;
}

/* ---------------------------------------------------- */
/* return user CPU time in seconds			*/
/* ---------------------------------------------------- */
float UserCpuTime(void)
{
    float  cpu;
    struct rusage rusage;
     	
    (void) getrusage(RUSAGE_SELF, &rusage);

    cpu =    (float) rusage.ru_utime.tv_sec 
	  + ((float) rusage.ru_utime.tv_usec / 1000000.);

    return cpu;
}

/* ---------------------------------------------------- */
/* return system CPU time in seconds			*/
/* ---------------------------------------------------- */
float SystemCpuTime(void)
{
    float  cpu;
    struct rusage rusage;
     	
    (void) getrusage(RUSAGE_SELF, &rusage);

    cpu =    (float) rusage.ru_stime.tv_sec 
	  + ((float) rusage.ru_stime.tv_usec / 1000000.);

    return cpu;
}

/* ---------------------------------------------------- */
/* return memory used in kb (IRIX)			*/
/* ---------------------------------------------------- */
float MemoryArena(void)
{
#if 0
#ifdef IRIX
    struct mallinfo info;

    info = mallinfo();
    
    return (float) info.arena / 1024.;
#else
    return -1.;
#endif
#else
    return -1.;
#endif
}

/*--------------------------------------------------------------*/
/* GiveProcessStatus: retourne un pointeur sur une structure	*/
/* contenant la memoire, le user cpu_time en seconde et		*/
/* nanoseconde du programme courant				*/
/* erreur sur sterr						*/
/*--------------------------------------------------------------*/


MyPsPtr GiveProcessStatus(FILE *sterr)
{
#if (! defined Darwin && ! defined Linux)
   char		active_proc_path[MAXPATHLEN];
   int		fildes;
#endif

#ifdef IRIX
   int		 retval;
#endif
   pid_t	prog_pid;

#if (defined  IRIX)
   prusage_t	my_prusage, *my_prusagePtr = &my_prusage;
#elif (defined Darwin) || (defined Linux)
   struct rusage	my_prusage, *my_prusagePtr = &my_prusage;
   int			pagesize = 0;
#elif (defined OSF1)     /* DEC */
   prpsinfo_t	my_prpsinfo, *my_prpsinfoPtr = &my_prpsinfo;
#endif

	/* get the pid of current program	*/

   sMy.u_cpu_time  = -1;
   sMy.nu_cpu_time = -1;
   sMy.mem_ko      = -1;

   prog_pid = getpid();

#if (defined Darwin) || (defined Linux)
   if (getrusage(RUSAGE_SELF, my_prusagePtr) != -1) { 
      sMy.u_cpu_time  = (int)(my_prusagePtr->ru_utime.tv_sec);
      sMy.nu_cpu_time = (int)(my_prusagePtr->ru_utime.tv_usec);
      /* sMy.mem_ko      = my_prusagePtr->ru_maxrss; */
      pagesize = getpagesize();
      sMy.mem_ko      = (int)(my_prusagePtr->ru_minflt*pagesize)/1024;
   }
   else {
      if (sterr)
	 fprintf(sterr,"GiveProcessStatus: cannot retrieve rusage !");
   }
   fprintf(stderr,"ru_maxrss %d  ru_ixrss %d  ru_idrss; %d ru_isrss %d\n",
   	(int)(my_prusagePtr->ru_maxrss),
   	(int)(my_prusagePtr->ru_ixrss),(int)(my_prusagePtr->ru_idrss),
	(int)(my_prusagePtr->ru_isrss));
   fprintf(stderr,"ru_minflt %d  ru_majflt %d  ru_nswap; %d pagesize=%d\n",
   	(int)(my_prusagePtr->ru_minflt),
	(int)(my_prusagePtr->ru_majflt),
	(int)(my_prusagePtr->ru_nswap),(int)(pagesize));
   return &sMy;

#else
   if (prog_pid == (pid_t)-1) {
      if (sterr)
	 fprintf(sterr,"GiveProcessStatus: cannot retrieve pid !");
      return NULL;
   }

	/* cette routine donne la taille memoire		*/
	/* et le cpu time					*/
	/* du process appelant					*/
	/* et de pid prog_pid					*/
	/* et retourne NULL:					*/
	/* 	si erreur d'ouverture du active_proc_path, i.e.	*/
	/*			/debug/"pid"			*/
	/*			ou plus recemment /proc/"pid"	*/
	/* 	si le process est fini				*/
	/*			ou n'existe pas			*/
	/*   (difficilement possible vu que c'est le program	*/
	/*    qui tourne et qui appelle cette routine!!!)	*/

				/* ---------------------------- */
				/* copie le nom du directory	*/
				/* des process			*/
				/* ---------------------------- */
   strcpy(active_proc_path,"/proc/");

				/* ---------------------------- */
				/* set the active proc name	*/
				/* it is the "id" file that is	*/
				/* in the /proc directory	*/
				/* ---------------------------- */
   sprintf(active_proc_path, "/proc/%d",
	   (int)prog_pid);

   fildes = open (active_proc_path, O_RDONLY);

				/* ---------------------------- */
   if (fildes == -1) {	/* open failed			*/ 
				/* ---------------------------- */
         if (errno == ENOENT) {
           if (sterr)
	      fprintf(sterr,
		"GiveProcessStatus:: program of pid %d has finished !!!\n",
		(int)prog_pid);
	   return NULL;	/* or non existent	*/
         }
         else {
            if (sterr)
	       fprintf(sterr,
		"GiveProcessStatus:: Can't open %s (errno:%d)",
		active_proc_path, errno);
	    return NULL;		/* unknown error	*/
         }
   }
           
#if (defined IRIX)
   retval = ioctl(fildes, PIOCUSAGE, my_prusagePtr);

   if (retval == -1) {
				/* ---------------------------- */
         if (errno == ENOENT) {	/* fildes non existent		*/
				/* ---------------------------- */
            if (sterr)
	       fprintf(sterr,
		"GiveProcessStatus:: program of pid %d has finished\n",
		(int)prog_pid);
	   return NULL;	/* or non existent	*/
         }
	 else {
           if (sterr)
	      fprintf(sterr,
		"GiveProcessStatus:: Can't ioctl PIOCUSAGE %s errno :%d",
		active_proc_path, errno);
				/* ---------------------------- */
	   return NULL;		/* cannot ioctl PIOCUSAGE	*/
				/* ---------------------------- */
         }
   }
   else {
	 sMy.u_cpu_time = my_prusagePtr->pu_utime.tv_sec;
	 sMy.nu_cpu_time = (int)(my_prusagePtr->pu_utime.tv_nsec/1000000);
					/* pu_size is unsigned long */
	 sMy.mem_ko = (int)((float)(my_prusagePtr->pu_size)*4.096);
   }


#elif (defined OSF1)     /* DEC */
   retval = ioctl(fildes, PIOCPSINFO, my_prpsinfoPtr);

   if (retval == -1.0) {
				/* ---------------------------- */
         if (errno == ENOENT) {	/* fildes non existent		*/
				/* ---------------------------- */
            if (sterr)
	       fprintf(sterr,
		"GiveProcessStatus:: program of pid %d has finished\n",
		(int)prog_pid);
	   return NULL;		/* or non existent	*/
         }
	 else {
           if (sterr)
	      fprintf(sterr,
		"GiveProcessStatus:: Can't ioctl PIOCPSINFO %s errno :%d",
		active_proc_path, errno);
				/* ---------------------------- */
	   return NULL;	/* cannot ioctl PIOCPSINFO	*/
				/* ---------------------------- */
         }
   }
   else {
	 sMy.u_cpu_time = my_prpsinfoPtr->pr_time.tv_sec;
	 sMy.nu_cpu_time = (int)(my_prpsinfoPtr->pr_time.tv_nsec/1000000));
	 sMy.mem_ko = ((int)((float)(my_prpsinfoPtr->pr_size)*4.096);
         if ((int)my_prpsinfoPtr->pr_zomb == 1) {
					/* zombie process	*/
	    if (sterr)
	       fprintf(sterr,
              "program of pid %d is zombie process <%c>",
	       (int)prog_pid, my_prpsinfoPtr->pr_sname);
         }
   }
#endif

   close(fildes);

   return &sMy;
#endif
}


