#!/bin/bash

organism=$1
delta=$2
prefixR=../../../$organism/11Blocks/Delta$delta/OrthBlocks/
prefixG=../../../$organism/01Genomes
suffix=.orth.pairs

echo "1name1	2name2	3#chr1	4#chr2	5#sizeInNuclG1	6#sizeInNuclG2	7#RBH	8#gen1	9#gen2	10%simi	11#SB	12#inv1	13#inv2	14#g/SB	15#SBw2	16#nuclInSB1	17#nuclInSB2	18#gInSB1	19#gInSB2	20#interSB1	21#interSB2" > $prefixR/../Summary${organism:0:1}$delta.txt

for f in $prefixR*$suffix
do 
# 1,2 : the two names: nameSH='name1.name2'
name=${f%%$suffix}
nameSH=${name##$prefixR}
name1=${nameSH:0:4}
name2=${nameSH:5:4}
# 3,4 : nb of scf / genomes
nb1=`awk '{if (NR==1) print NF}' $prefixG/$name1.ch`
nb2=`awk '{if (NR==1) print NF}' $prefixG/$name2.ch`
# 5,6 : size of the genomes in nucleotides
sizeG1=`awk '(NR==1){som=0}(NR==2){chr=$3}{if (chr==$3){en=$5}else{som+=en;chr=$3}}END{som+=$5;print som}' $prefixG/$name1.def`
sizeG2=`awk '(NR==1){som=0}(NR==2){chr=$3}{if (chr==$3){en=$5}else{som+=en;chr=$3}}END{som+=$5;print som}' $prefixG/$name2.def`
# 7 : nb of BDBH
nbBDBH=`awk 'END{print NR}' $prefixR/../../../10RBH/$name1.$name2.opsc.shrt`
# 8,9 : nb of genes / genomes
nbgen1=`awk 'END{print $(NF-1)}' $prefixG/$name1.def`
nbgen2=`awk 'END{print $(NF-1)}' $prefixG/$name2.def`
echo -n -e "$name1\t$name2\t$nb1\t$nb2\t$sizeG1\t$sizeG2\t$nbBDBH\t$nbgen1\t$nbgen2\t"

# 10 : average similarity between syntenic homolog
awk '{sum+=$11} END{printf("%.4f\t",sum/NR);}' $f
# 11 : nb of Synteny Blocks
awk 'END{printf("%d\t",NR);}' $name.orth.synt.start_end
# 12 : nb of time where 2 consecutive blocks along G1 are on the same chromo in G2
awk '(NR==1){chr1=$1;chr2=$2;cpt=0}{if($1!=chr1){chr1=$1;chr2=$2} else {if($2!=chr2){chr1=$1;chr2=$2}else{cpt+=1}}} END{printf("%d\t",cpt);}' $name.orth.synt
# 13 : nb of time where 2 consecutive blocks along G2 are on the same chromo in G1
sort -n -k 2 -k 6 $name.orth.synt > temp
awk '(NR==1){chr1=$1;chr2=$2;cpt=0}{if($2!=chr2){chr2=$2;chr1=$1} else {if($1!=chr1){chr2=$2;chr1=$1}else{cpt+=1}}} END{printf("%d\t",cpt);}' temp
rm temp
# 14 : average of gene per Blocks 
awk '{sum+=$11+$20} END{printf("%.4f\t",sum/(2*NR));}' $name.orth.synt.start_end
# 15 : nb of Blocks with 2 anchors
awk '{if($21==2){sum+=1}} END{printf("%d\t",sum);}' $name.orth.synt.start_end
# 16 : nb de nucl inside a block in G1
awk '(NR==1){b1=$6;b2=$10;chr1=$1}{if($1==chr1){if ($6<=b2) {if (b2<$10) {b2=$10}}else{sum+=b2-b1+1;b1=$6;b2=$10}}else{sum+=b2-b1+1;b1=$6;b2=$10;chr1=$1}} END{sum+=b2-b1+1;printf("%d\t",sum)}' $name.orth.synt.start_end
# 17 : nb de nucl inside a block in G2
sort -n -k 13 $name.orth.synt.start_end > temp
awk '(NR==1){b1=$15;b2=$19;chr2=$2}{if($2==chr2){if ($15<=b2) {if (b2<$19) {b2=$19}}else{sum+=b2-b1+1;b1=$15;b2=$19}}else{sum+=b2-b1+1;b1=$15;b2=$19;chr2=$2}} END{sum+=b2-b1+1;printf("%d\t",sum)}' temp
# 18 : nb de genes within block in G1
awk '(NR==1){b1=$4;b2=$8}{if ($4<=b2) {if (b2<$8) {b2=$8}}else{sum+=b2-b1+1;b1=$4;b2=$8}} END{sum+=b2-b1+1;printf("%05d\t",sum)}' $name.orth.synt.start_end
# 19 : nb de genes within block in G2
awk '(NR==1){b1=$13;b2=$17}{if ($13<=b2) {if (b2<$17) {b2=$17}}else{sum+=b2-b1+1;b1=$13;b2=$17}} END{sum+=b2-b1+1;printf("%05d\t",sum)}' temp
# 20 : interBloc size (nucl) in G1
awk '(NR==1){b2=$10;chr1=$1}{if($1==chr1){if ($6<=b2) {if (b2<$10) {b2=$10}}else{sum+=$6-b2+1;b2=$10}}else{b2=$10;chr1=$1}} END{printf("%d\t",sum)}' $name.orth.synt.start_end
# 21 : interBloc size (nucl) in G2
awk '(NR==1){b2=$19;chr2=$2}{if($2==chr2){if ($15<=b2) {if (b2<$19) {b2=$19}}else{sum+=$15-b2+1;b2=$19}}else{b2=$19;chr2=$2}} END{printf("%d\n",sum)}' temp
rm temp
#nbfeatIN2=`awk '(NR==1)...END{print sum}
#percent= echo "($nbfeatIN/($nbfeat1+$nbfeat2))*100" | bc -l	
#echo -n -e "$percent"
done >> $prefixR/../Summary${organism:0:1}$delta.txt

