#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
## SynChro reconstruct pairwise synteny blocks for a given Delta

import re, os, sys
import n01Genomes, n02Homologs, n03Summary, n04SVGPicture, n05DotPlot, n06G1fG2


def _main_(argv):

   ## RBH thresholds
   # Bestfit score threshold (in the 0-100 range)
   min_sim_RBH = 40
   # length ratio threshold (longest sequence divided by the shortest)
   max_diflen = 1.3
   
   ## NON-RBH thresholds
   # minimum percentage of similarity shared between two homologs (in the 0-1 range)
   min_sim_nonRBH = 0.3 #0.4
   # length ratio threshold (length of the alignment divided by the shortest sequence)
   min_align = 0.5 #0.7
   
   ## MINIMUM ANCHORS
   min_nb_RBH = 2
   min_nb_anch = 2
   

   if len(argv)!=4 and len(argv)!=5:
      print "3/4 arguments are expected:\n\
      -> the clade Name (as 'Yeast' or 'Vertebrate')\n\
      -> 0 (if you want to do all pairwise) / 1 (to choose a subgroup)\n\
      -> delta for RBH genes (an int)\n\
      -> delta for non-RBH genes (OPTIONAL if equal to the previous one) "
      sys.exit(1)

   clade=argv[1]
   Allgenomes=int(argv[2])
   DeltaRBH=argv[3]
   if len(argv)==5:
      DeltaNonRBH=argv[4]
   else:
      DeltaNonRBH=DeltaRBH

   path='../../'+clade
   iniShortNameList=n01Genomes.ListShortName(path)
   iniShortNameList.sort()


   if Allgenomes!=0:    ## Question/Interface
      i=0
      for x in iniShortNameList:
         i+=1
         print i,'\t',x

      questionGenome=raw_input("For which genomes do you want to reconstruct synteny blocks? \n\
   Write their associated numbers as space separated values.\n")
      questionLigne=re.split(" ",questionGenome)
      if len(questionLigne)<=1:
         print "!!! At least 2 genomes must be compared  !!!!"
         sys.exit(1)
      else:
         shortNameList=[]
         for num in questionLigne:
            if int(num)<=len(iniShortNameList):
               shortNameList.append(iniShortNameList[int(num)-1])
            else:
               print "!!! All number has to correspond to a genome  !!!!"
               sys.exit(1)
         shortNameList.sort()
   else: 
      shortNameList=iniShortNameList

   print shortNameList
   ## Creation of the 'Genomes.list' file needed by Opscan
   os.system('mkdir '+path+'/10RBH')
   GenomesList=open(path+'/10RBH/Genomes.list','w')
   for i in shortNameList:
      GenomesList.write(i+'.prt\n')
   GenomesList.close()

   ## Lists of the name, length and centromere position of the chromosomes for each genome
   chromName=[]
   chromNumber=[]
   centroNumber=[]
   for i in shortNameList:
      chromNameI,centroNumberI,chromNumberI=n01Genomes.chromosomesLists(path, i)
      chromName.append(chromNameI)
      centroNumber.append(centroNumberI)
      chromNumber.append(chromNumberI)
   print chromName, centroNumber, chromNumber


   ##################
   ## '../../'+clade+'/10RBH/' contains the RBH 
   ##################

   ## (1) RBH Computation using 'opscan'
   os.system('./go_orthologs '+clade+' '+str(min_sim_RBH)+' '+str(max_diflen)+' '+path+'/10RBH/Genomes.list')


   ##################
   ## '../../'+clade+'/11Blocks/Delta'+Delta+'/' contains the SynChro outputs
   ##################
      
   os.system('mkdir '+path+'/11Blocks')
   ## DELTA 
   os.system('mkdir '+path+'/11Blocks/Delta'+DeltaRBH)
   os.system('mkdir '+path+'/11Blocks/Delta'+DeltaRBH+'/RBHBlocks')    
   os.system('mkdir '+path+'/11Blocks/Delta'+DeltaRBH+'/OrthBlocks')
   os.system('mkdir '+path+'/11Blocks/Delta'+DeltaRBH+'/SyntenyMaps')
   os.system('mkdir '+path+'/11Blocks/Delta'+DeltaRBH+'/Dotplots')
   os.system('mkdir '+path+'/11Blocks/Delta'+DeltaRBH+'/G1fG2')

   ## New Orthologs
   for i in range(len(shortNameList)):
      for j in range(i+1,len(shortNameList)):
         
         shortName1=shortNameList[i]
         shortName2=shortNameList[j]
         order=shortName1<shortName2
         ## first Chromo '1' or '0'
         #o1=n01Genomes.start0(path,shortName1,chromName[i])
         #o2=n01Genomes.start0(path,shortName2,chromName[j])
         if order:
            FirstName,SecondName=shortName1,shortName2
         else:
            FirstName,SecondName=shortName2,shortName1
            #o2,o1=o1,o2
         nameFile=path+'/11Blocks/Delta'+DeltaRBH+'/OrthBlocks/'+FirstName+'.'+SecondName
      
         ## (2) Synteny blocks made of RBH only
         if FirstName+'.'+SecondName+'.rbh.synt' not in os.listdir(path+'/11Blocks/Delta'+DeltaRBH+'/RBHBlocks'):
            print './go_syntorthdist '+DeltaRBH+' '+clade+' '+FirstName+'.'+SecondName
            os.system('./go_syntorthdist '+DeltaRBH+' '+str(min_nb_RBH)+' '+clade+' '+FirstName+'.'+SecondName)
            # syntons start end
            n03Summary.startEnd(path+"/11Blocks/Delta"+DeltaRBH+"/RBHBlocks/",'rbh',clade)

         ## Synteny blocks completed with non-RBH (HOMO at least 30%)
         if (os.listdir(path+'/11Blocks/Delta'+DeltaRBH+'/OrthBlocks')).count(FirstName+'.'+SecondName+'.orth.synt')==0 :      
            # (3) Research for new homologues 55%
            (nameListGenome1,nameListGenome2,codeListGenome1,codeListGenome2,
            recapBlocGene1,recapBlocGene2,recapSimilarityG1,recapSimilarityG2,
            recapGene1Ortho2,recapGene2Ortho1)= n03Summary.recapitulatifList(
            path,shortName1,shortName2,1,DeltaRBH)
            n02Homologs.orthologsNonBDBHFile(path,shortName1,recapBlocGene1,
            shortName2,recapBlocGene2,recapGene1Ortho2,chromNumber[i],chromNumber[j],len(recapGene2Ortho1),DeltaRBH,int(DeltaNonRBH),min_sim_nonRBH,min_align)
            print('orthologsNonBDBH('+shortName1+','+shortName2+') finished')
            n02Homologs.orthologsNonBDBHFile(path,shortName2,recapBlocGene2,
            shortName1,recapBlocGene1,recapGene2Ortho1,chromNumber[j],chromNumber[i],len(recapGene1Ortho2),DeltaRBH,int(DeltaNonRBH),min_sim_nonRBH,min_align)
            print('orthologsNonBDBH('+shortName2+','+shortName1+') finished')

            # (4) Synteny blocks made of RBH and non-RBH pairs
            os.system('awk -v delta='+DeltaRBH+' -v delta2='+DeltaNonRBH+' -v minAnch='+str(min_nb_anch)+' -f get.syntonorth.awk '+nameFile+'.orth.pairs > '+nameFile+'.orth.synt')

            ## START-END files
            n03Summary.startEnd(path+"/11Blocks/Delta"+DeltaRBH+"/OrthBlocks/",'orth',clade)

         ## (5) SyntenyMap picture
         if (os.listdir(path+'/11Blocks/Delta'+DeltaRBH+'/SyntenyMaps')).count(FirstName+'.'+SecondName+'.svg')==0 :      
            newPairOfOrthoDico=n02Homologs.orthologsNonBDBHDico(path,shortName1,shortName2,DeltaRBH)

            (nameListGenome1,nameListGenome2,codeListGenome1,codeListGenome2,
            recapBlocGene1,recapBlocGene2,recapSimilarityG1,recapSimilarityG2,
            recapGene1Ortho2,recapGene2Ortho1)=n03Summary.recapitulatifList(
            path,shortName1,shortName2,2,DeltaRBH)

            shortNameFile=path+'/11Blocks/Delta'+DeltaRBH+'/SyntenyMaps/'+shortName1+'.'+shortName2+'.svg'
            n04SVGPicture.syntenyBlocSVG(shortNameFile,shortName1,chromName[i],
            recapBlocGene1,codeListGenome1,chromNumber[i],centroNumber[i],
            nameListGenome1,recapSimilarityG1,shortName2,chromName[j],recapBlocGene2,
            codeListGenome2,chromNumber[j],centroNumber[j],nameListGenome2,
            recapSimilarityG2,newPairOfOrthoDico,order)
            shortNameFile=path+'/11Blocks/Delta'+DeltaRBH+'/SyntenyMaps/'+shortName2+'.'+shortName1+'.svg'
            n04SVGPicture.syntenyBlocSVG(shortNameFile,shortName2,chromName[j],
            recapBlocGene2,codeListGenome2,chromNumber[j],centroNumber[j],
            nameListGenome2,recapSimilarityG2,shortName1,chromName[i],recapBlocGene1,
            codeListGenome1,chromNumber[i],centroNumber[i],nameListGenome1,
            recapSimilarityG1,newPairOfOrthoDico,not order)
            print('recap('+shortName1+','+shortName2+') finished')
         else:
            print "warning: File "+path+"/11Blocks/Delta"+DeltaRBH+"/SyntenyMaps/"+shortName1+"."+shortName2+".svg exists. Erase it if you want to re-compute it"

         ## (6) Dotplot picture
         if (os.listdir(path+'/11Blocks/Delta'+DeltaRBH+'/Dotplots')).count('dotplot.'+FirstName+'.'+SecondName+'.'+DeltaRBH+'.svg')==0 :
            n05DotPlot.dotplot(path+'/01Genomes/',path+'/11Blocks/Delta'+DeltaRBH+'/',FirstName,SecondName,DeltaRBH)
      
         ## (7) Chromosomal Painting picture
         if (os.listdir(path+'/11Blocks/Delta'+DeltaRBH+'/G1fG2')).count(FirstName+'.'+SecondName+'f'+FirstName+'.'+DeltaRBH+'.svg')==0 :
            n06G1fG2._main_([0,clade,DeltaRBH,FirstName,SecondName])
            n06G1fG2._main_([0,clade,DeltaRBH,SecondName,FirstName])

_main_(sys.argv)
#_main_([0,'Vertebrate','1','1'])
